/*----------------------------------------------------------------------
 *
 * CAMELot/	$Id: macrocell.c,v 2.96 2000/05/25 16:23:47 colombo Exp $
 *
 * File:	macrocell.c
 *
 * Authors:	Stephen P Booth		EPCC
 *		K. Kavoussanakis	EPCC
 *		A Gordon Smith		EPCC
 *		Scott D Telford		EPCC
 *
 * Purpose:     Parallel CA execution framework
 *
 *----------------------------------------------------------------------
 */

static char RCSID[]="@(#)$Id: macrocell.c,v 2.96 2000/05/25 16:23:47 colombo Exp $";

/*---------- Standard headers ----------*/
#include <stdio.h>
#include <stddef.h>
#include <limits.h>
#include <float.h>
#include <string.h>
#include <mpi.h>
#include <math.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <netinet/tcp.h>

/*---------- Local headers ----------*/
/*
 * CPT_INCLUDE_FILE defined on compile line as name of
 * CARPET-generated header (with double quote marks)
 * eg. cc -DCPT_INCLUDE_FILE="cpt_life.h" ...
 *
 */
#include CPT_INCLUDE_FILE
#include "cpt_ccdefs.h"

#ifdef PROFILING		/* This is defined in cpt_ccdefs.h */
#include <mon.h>
#endif
#include "constants.h"
#include "common.h"

/*---------- Local Definitions ----------*/

/*----------------------------------------------------------------------
 timer struct interface
 ----------------------------------------------------------------------*/

#define TITLE_LENGTH 25 /* Length of title member of struct timer */

typedef enum {
        FUNC = 9999,
	VIS,
	PRJ,
	BOUND,
	STEER,
	TOTAL
} tmr_code;

typedef struct {
    double start;
    double stop;
    double sum;
    double best;
    double worst;
    u_char started;
    unsigned long called;
    char title[TITLE_LENGTH];
} timer;

static void init_tmr (timer *, const char *);

static int start_tmr (timer *);

static int stop_tmr (timer *);

static void print_tmr (const timer *, FILE *);

/*----------------------------------------------------------------------
 stats struct interface
 ----------------------------------------------------------------------*/

typedef struct {
    timer func;
    timer vis;
    timer prj;
    timer bound;
    timer steer;
    timer total;
    int rank;
    unsigned long gens;
    int start_gen;
    u_int period;
    u_char work;
    FILE *outfile;
} stats;

static void init_sts (stats *, int, u_char, u_int, const char *);

static void reset_stats (stats *);

static int start_one_timer_sts (stats *, tmr_code, int);

static int stop_one_timer_sts (stats *, tmr_code, int);

static void print_sts (const stats *);

static void close_file_sts (const stats *);

/*----------------------------------------------------------------------
 state_dt struct interface
 ----------------------------------------------------------------------*/

typedef struct {
    MPI_Datatype data;
    int states;
    int state_ind[NumOfStates];
} state_dt;

static void init_state_dt (state_dt *, MPI_Datatype);

static int add_state (state_dt *, int);

/*----------------------------------------------------------------------
 state_dt_list struct interface
 ----------------------------------------------------------------------*/

typedef struct {
    state_dt statetypes[NumOfStates];
    int many;
} state_dt_list;

static void init_state_dt_list (state_dt_list *);

static int add_state_dt (state_dt_list *, int);

/*---------- Global Declarations ----------*/

CptCell *ca1;				/* 1st copy of Cellular Automata */
CptCell *ca2;				/* 2nd copy of Cellular Automata */
int      cpt_dimx, cpt_dimy, cpt_dimz; 	/* Variable local CA dim sizes */
int      cpt_x,    cpt_y,    cpt_z;	/* Current coordinates, X, Y, Z  
					   Must be global for Get[XYZ] */
char 	*out_basename = NULL;		/* Root of periodic save filename */
char 	*out_dirname = NULL;		/* Path of periodic save filename */
int	 save_step = INT_MAX;		/* Period of saves */
int 	 num_gens;			/* Remaining generations to run */
int      cpt_generation = 0;		/* Current generation */
					/* Must be initialised for vis_list */
char	 in_steering = 0;		/* Flag for steering statement */

int      cmtWorldSize, cmtWorldRank;	/* MPI world size and local rank */
int      cmtPrevRank, cmtNextRank;	/* Ranks of neighbour macrocells */
MPI_Comm cmtCommCommand, cmtCommBoundary; /* MPI contexts: data & control */
MPI_Datatype cmtBoundaryType;		/* MPI datatype for boundary data */
int      prot_sockfd = 0, vis_sockfd = 0; /* protocol and visualisation socket 
					   fd, global for comm abstraction
					   independence */

list 	vis_list;			/* List of visualised planes */
plane_list all_planes;			/* List of all planes in the Engine */
int 	alt_intnl_bound[2][NFOLDS];	/* Left/Right internal boundary
					   altered flag */
int 	alt_strip[NFOLDS];		/* Internal cells altered flag */
int	active_strip[NFOLDS];		/* Active Strip flag */
int	manual_folds = 0;		/* Flag to examine (in)active folds */

double 	minmax[NumOfStates][2];		/* Minimum and Maximum value for each
					   state. Mins in [0], Maxes in [1] */
char 	auto_map[NumOfStates];		/* Flags for automatic colour map */
stats	statistics;			/* Statistics for timers */
state_dt_list *dt_list = 0;		/* state_dt_list for AVS files */

#ifdef EVEN_DECOMP
#else
int 	CPT_DIMX[CPT_NPROCS];		/* Actual x data in process */
int 	CPT_F_X[CPT_NPROCS][NFOLDS];	/* Actual x data per strip */
int 	CPT_S_X[CPT_NPROCS][NFOLDS];	/* Total Strip x-sizes */
int	first_small_strip_ind[CPT_NPROCS]; /* Index of first strip of
					      process to have smaller size */
#endif

/*---------- External function prototypes ----------*/
extern void cpt_hook_init (void);
extern void cpt_func (CptCell *, CptCell *);
extern void cpt_set_state (CptCell *, int, void *, int);
extern void cpt_get_state (CptCell *, int, void *, int);
extern void cpt_mpi_type_cell (MPI_Datatype *);
extern void cpt_hook_finalize (void);
extern int cpt_thresh (CptCell *);
extern req_code cpt_steering (CptCell *);

/*---------- External variable declarations ----------*/
extern float         cpt_globpar[NumOfGlobPar];	     /* CARPET parameters */
extern const size_t  cpt_state_size[NumOfStates];    /* Substate bytesizes */
extern MPI_Datatype  cpt_state_mpidt[NumOfStates+1]; /* Substate MPI Dtypes */
#include <rpc/rpc.h>
extern bool_t        (*cpt_state_XDRfn[NumOfStates])();/* Substate XDR function */
extern const int     cpt_determin;		     /* Deterministic flag */
extern const char   *cpt_state_name[NumOfStates];    /* Names of states */

/*---------- Shared function prototypes ---------*/
void cpt_randomize (void);
void cpt_srandom (unsigned int);

/*---------- Local function prototypes ---------*/
static int cmt_read (char *, int);
static int cmt_read_global (char *);
static int cmt_read_all (char *);
static int cmt_write (char *, int);
static int cmt_write_global (char *);
static int cmt_create_fld (char *, char *);
static int cmt_write_fld (char *, char *, int);
static int cmt_write_all (char *, char *);
static int check_fs (char *, char *);
static int cmt_boundary_copy (CptCell *);
static int cmt_boundary_swap (CptCell *);
static void init_boundaries (void); 
static req_code rv (req_code);
static req_code run (void);
static req_code tx_vis_pack (cell *, char);
static void colour_map (const u_char *, u_char *, int, int, double, double);
static req_code serv_save_request (void);
static req_code serv_set_fold (void);
static req_code serv_set_load (void);
static req_code serv_view_state (void);
static req_code serv_set_state (void);
static req_code serv_set_param (void);
static req_code serv_get_param (void);
static req_code serv_terminate (void);
static req_code serv_add_plane (void);
static req_code serv_del_plane (void);
static req_code serv_proj_read (void);
static req_code serv_proj_save (void);
static req_code serv_periodic_save (void);
static req_code serv_set_minmax (void);
static req_code send_ack (req_code);
int cpt_save (char *);
static void check_plane (plane *);
static void bcast_plane (plane *, MPI_Comm);
static void check_pos (int *);
static int get_val_size (const int *);
static void get_x_line (CptCell *, int, u_char *, int);
static void set_x_line (CptCell *, int, u_char *, int);
static void print_all_stats (void);

#ifdef EVEN_DECOMP
static void line2fold (const u_char *, u_char *, size_t);
static void fold2line (const u_char *, u_char *, size_t);
static void get_write_ptr (u_char *, u_char *, int, int, int, int, int, int, int);
static void get_scatter_ptr (u_char **, u_char *, u_char *, int , int);
#else
static void calc_x_sizes (void);
#endif /* EVEN_DECOMP */

/*---------- Main program ----------*/
int main (int argc, char *argv[]) {
  char *c;
  static char in_basename[PATH_MAX];
  int arg;
  int single_load = 0;
  int period;
  int typeSize;
  MPI_Aint typeExt;
  MPI_Datatype cellType;

  size_t size;
  int status;
  req_code request;
  int nodelay = 1, keepalive = 1;
  int serv_tcp_port = 0;	/* Initialisation for batch mode */
  char hostname[100] = "";	/* Initialisation for batch mode */	
  int st;			/* Index for minmax initialisation */

#ifdef EVEN_DECOMP
  if (0 != (DIMX%(CPT_NPROCS*NFOLDS))) {
    fprintf (stderr, "Improper CPT_NPROCS - NFOLDS combination, exiting\n");
    exit (-1);
  }				/* End if (0 != (DIMX%(CPT_NPROCS*NFOLDS))) */
#endif

  MPI_Init (&argc, &argv);

  MPI_Comm_rank (MPI_COMM_WORLD, &cmtWorldRank);
  MPI_Comm_size (MPI_COMM_WORLD, &cmtWorldSize);

  if (cmtWorldSize != CPT_NPROCS) {
      fprintf (stderr, "ERROR: decomposition does not match NPROCS ");
      fprintf (stderr, "       (actual %d, expected %d)\n",
	       cmtWorldSize, CPT_NPROCS);
      MPI_Finalize();
      exit(1);
  }

#ifdef EVEN_DECOMP
#else
/* Calculate dimensions */
  calc_x_sizes ();
#endif

/* malloc dt_list */
  if (!dt_list) {
    if (!(dt_list = (state_dt_list *) malloc (sizeof (state_dt_list)))) {
      fprintf (stderr, "malloc failed, exiting\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!malloc) */
  } else { /* dt_list exists although it should not */
    fprintf (stdout, "Warning, dt_list should not exist!\n");
  }				/* End if (!dt_list) */

/* initialise dt_list */
  init_state_dt_list (dt_list);

  /*--- Cartesian process topology creation for boundary comms */
  period = 1;
  MPI_Cart_create (MPI_COMM_WORLD, 1, &cmtWorldSize, &period, 1,
		   &cmtCommBoundary);
  MPI_Cart_shift (cmtCommBoundary, 0, 1, &cmtPrevRank, &cmtNextRank);
  MPI_Comm_dup (MPI_COMM_WORLD, &cmtCommCommand);

#ifdef DEVELOP
  printf("%d -> [%d] -> %d\n", cmtPrevRank, cmtWorldRank, cmtNextRank);
#endif

  /*--- Call to CARPET generated initialisation routine */
  cpt_hook_init ();

  /*--- Construction of MPI datatype for macrocell boundary data */
  /*--- Adequate for FOLDS as well */

  cpt_mpi_type_cell (&cellType);
  MPI_Type_vector (CPT_Z*CPT_Y, Radius, CPT_X, cellType, &cmtBoundaryType);
  MPI_Type_commit (&cmtBoundaryType);

#ifdef DEVELOP
  if (cmtWorldRank == 0) {
      MPI_Type_size (cellType, &typeSize);
      MPI_Type_extent (cellType, &typeExt);
      printf ("Cell type, size %d, extent %d\n", typeSize, typeExt);
      MPI_Type_size (cmtBoundaryType, &typeSize);
      MPI_Type_extent( cmtBoundaryType, &typeExt );
      printf("Boundary type, size %d, extent %d\n", typeSize, typeExt );
  }
#endif

  /*--- CA data allocation and initialisation [to zero via calloc()] */
  /*--- (enough even in the case of FOLDS) */
  ca1 = (CptCell *) calloc (CPT_X*CPT_Y*CPT_Z, sizeof(CptCell));
  ca2 = (CptCell *) calloc (CPT_X*CPT_Y*CPT_Z, sizeof(CptCell));

  if (!ca1 || !ca2) {
    fprintf (stderr, "ERROR: failed to allocate CA arrays\n");
    MPI_Finalize();
    exit(1);
  }

  /*--- minmax[][] and auto_map[] initialisation */
  for (st = 0; st < NumOfStates; st++) {
    minmax[st][0] = DBL_MAX;
    minmax[st][1] = -DBL_MAX;
    auto_map[st] = 1;
  }				/* End for (st) */

  /*--- Command argument processing loop */
  for (arg=1; (arg < argc) && (argv[arg][0] == '-'); )
  {
#ifdef DEVELOP
    fprintf (stdout, "arg: %i\t%c option\n", arg, argv[arg][1]);
#endif
    switch (argv[arg][1]) {

      case 'l': {
	int stateid;
	if (sscanf (argv[arg]+2, "%d", &stateid) != 1) {
	  fprintf (stderr, "usage: -%c<id>, ignored\n", argv[arg][1]);
	  arg += 1;
	}
	else {
	  single_load = 1;
	  cmt_read (argv[arg+1], stateid);
	  arg += 2;
	}
      }
      break;

      case 'n': {
	if (sscanf (argv[arg]+2, "%d", &num_gens) != 1) {
	  fprintf (stderr, "usage: -%c<gens>, ignored\n", argv[arg][1]);
	}
	arg += 1;
      }
      break;

      case '0':
	/*--- No states loaded -- all set to zero */
	single_load = 1;
	arg += 1;
        break;

      case 'P': {
	if (sscanf (argv[arg]+2, "%i", &serv_tcp_port) != 1) {
	  fprintf (stderr, "usage: -%c<port>; exiting\n", argv[arg][1]);
          MPI_Abort (cmtCommCommand, -1);
	}
	arg += 1;
      }
      break;

      case 'H': {
	if (sscanf (argv[arg]+2, "%s", hostname) != 1) {
	  fprintf (stderr, "usage: -%c<hostname>; exiting\n", argv[arg][1]);
          MPI_Abort (cmtCommCommand, -1);
	}
	arg += 1;
      }
      break;

      case 's': {
        static char out_name[PATH_MAX];
        char *c;
        int cfs;

	if (sscanf (argv[arg]+2, "%d", &save_step) != 1) {
	  fprintf (stderr, "usage: -%c<save_step>, ignored\n", argv[arg][1]);
          save_step = INT_MAX;
	  arg += 1;
	}

	else {
          strncpy (out_name, argv[++arg], sizeof(out_name));

          c = strrchr (out_name, '/');

          if (c != NULL) {
            *c = (char) 0;
            out_dirname = out_name;
            out_basename = c+1;
          }

          else {
            out_dirname = NULL;
            out_basename = out_name;
          }

          c = strrchr (out_basename, '.');
          if (c != NULL) {
            *c = (char) 0;
          }

          if (0 == cmtWorldRank) { /* Only proc 0 calls check_fs() */
            if (OVER_W == (cfs = check_fs (out_dirname, out_basename))) {
              fprintf (stderr, 
                       "Warning: periodic save files will be overwritten!\n");
            } else if (0 > cfs) {
              fprintf (stderr, "Error: check_fs failed!\n");
              MPI_Abort (cmtCommCommand, -1);
            }			/* End if (OVER_W) */
          }			/* End if (0 == cmtWorldRank) */
	  arg += 1;
	}

#ifdef DEVELOP
  fprintf (stdout, "save_step: %i, out_dirname: %s, out_basename: %s\n",
           save_step, out_dirname, out_basename);
#endif
      break;
      }

      case 't': {
        int st_period;
        static char st_fname[PATH_MAX];
        u_char stats_on = (u_char) 0;


	if (sscanf (argv[arg]+2, "%d", &st_period) != 1) {
	  fprintf (stderr, "usage: -%c<st_period>, ignored\n", argv[arg][1]);
	  stats_on = (u_char) 0;
	} else {
	  stats_on = (u_char) 1;
          strncpy (st_fname, argv[++arg], sizeof(st_fname));
        }
	arg += 1;

        /* Initialise statistics */
        init_sts (&statistics, cmtWorldRank, stats_on, st_period, st_fname);

        break;
      }				/* End case 't' */

      case 'x':
      case 'y':
      case 'z':
      case 'X':
      case 'Y':
      case 'Z':
	/*--- Handled before CA array allocation */
	break;

      default:
	fprintf (stderr, "usage: unknown argument '%s', ignored\n", argv[arg]);
	arg += 1;
	break;
    }
  }

  /*--- Command-line argument error, usage info */
  if (argc-arg > 2) {
    fprintf(stderr, "Usage: %s [options] [<in-base>]\n", argv[0]);
    MPI_Abort (cmtCommCommand, -1);
  }

  /*--- Process remaining command-line arguments */
  if (!single_load) {
    /*--- Read entire CA configuration from file, from "basename" */
    strncpy (in_basename, argv[arg++], sizeof(in_basename));
    c = strrchr( in_basename, '.' );
    if (c != NULL) {
      *c = (char) 0;
    }
    if (0 > cmt_read_all (in_basename)) {
      MPI_Abort (cmtCommCommand, -1);
    }                         	/* End if (0 > cmt_read_all) */
  }                         	/* End if (!single_load) */

/* Run in batch mode */
  if ((!serv_tcp_port) || (!strcmp ("", hostname))) { /* Batch mode */
    rv (BATCH);
    run ();
/* Print stats */
    if (!statistics.period) {
      print_all_stats ();
    }
/* Close output file */
    if (0 == cmtWorldRank) {
      close_file_sts (&statistics);
    }				/* End if (0 == cmtWorldRank) */

    MPI_Finalize();
    return 0;
  }				/* End if BATCH */

  if (0 == cmtWorldRank) {
/* Initialise sockets and connect to server */
    if ((prot_sockfd = start_client (serv_tcp_port, hostname)) < 0) {
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if start_client */

    if ((vis_sockfd = start_client (serv_tcp_port, hostname)) < 0) {
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if start_client */

  }				/* End if (0 == cmtWorldRank) */

/* Initialise all_planes plane_list */
  init_plane_list (&all_planes);

/* Start visualisation list */
  init_list (&vis_list);

  do {
    if (0 == cmtWorldRank) {
/* Get request */
      size = sizeof (request);
      if (size != readn (prot_sockfd, (char *) &request, size)) {
        fprintf (stderr, "macrocell: readn error, exiting\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
    }				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get request */
    MPI_Bcast ((int *) &request, 1, MPI_INT, 0, cmtCommCommand);

    status = rv (request);

    if ((status > -1) && ((EVOLVE == request) || (LOOP == request) || (RESUME == request))) {
      if (run () < 0) {
        break;
      } 			/* End if (run () < 0) */
    }				/* End if (status && request) */
  } while (request != EXITCODE); /* End do */

  /*--- CARPET-generated finalisation routine */
  cpt_hook_finalize();

  /*--- Close sockets */
  if (0 == cmtWorldRank) {
    shutdown (prot_sockfd, 2);
    shutdown (vis_sockfd, 2);
  }				/* End if (0 == cmtWorldRank) */

/* Print stats */
  if (!statistics.period) {
    print_all_stats ();
  }
/* Close output file */
  if (0 == cmtWorldRank) {
    close_file_sts (&statistics);
  }				/* End if (0 == cmtWorldRank) */

/* Finalise MPI and exit */
  MPI_Type_free (&cmtBoundaryType);
  MPI_Finalize();
  return 0;

} /* End main */

#ifdef EVEN_DECOMP
#else

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: calc_x_sizes
 * 
 * Calculates the per-process and per-strip sizes
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *
 *
 * Returns: (void)
 *
 *
 *----------------------------------------------------------------------*/

void calc_x_sizes (void) {
  int strip; /* Index */
  int proc; /* Index */
  int usual_CPT_DIMX;
  int usual_CPT_F_X_proc;
  int plus_one;

  usual_CPT_DIMX = DIMX/CPT_NPROCS; /* Calculation taken out */

  for (proc = 0; proc < CPT_NPROCS; proc++) {

/* Calculate number of elements in process
   DIMX/CPT_NPROCS is the usual size. Perhaps some processes have
   an extra one. We decided to exclude proc 0 from this, as it is
   already burdened. This is what the boolean component of the sum,
   DIMX%CPT_NPROCS > CPT_NPROCS-proc-1 calculates. */

    CPT_DIMX[proc] = usual_CPT_DIMX+(DIMX%CPT_NPROCS > CPT_NPROCS-proc-1);

#ifdef DEBUG_CALC_X_SIZES
    if (0 == cmtWorldRank) fprintf (stdout, "rank: %i\t CPT_DIMX: %i\n", proc, CPT_DIMX[proc]);
#endif /* DEBUG */
    usual_CPT_F_X_proc = CPT_DIMX[proc]/NFOLDS; /* Calculation taken out */

    for (strip = 0; strip < NFOLDS; strip ++) {
/* Calculation of CPT_F_X[proc][strip] follows reverse logic than that of
   CPT_DIMX, as there is no point in not burdening the initial strips */

      plus_one = (CPT_DIMX[proc]%NFOLDS > strip); /* Taken out */
      CPT_F_X[proc][strip] = usual_CPT_F_X_proc+plus_one;
      CPT_S_X[proc][strip] = CPT_F_X[proc][strip]+2*Radius;

#ifdef DEBUG_CALC_X_SIZES
      if (0 == cmtWorldRank) fprintf (stdout, "\tstrip: %i\t CPT_F_X[proc][strip]: %i\n", strip, CPT_F_X[proc][strip]); 

#endif /* DEBUG */
    }			/* End for (strip) */

  }			/* End for (proc) */

/* first_small_strip_ind[proc] is the first strip in proc that has size
   smaller than the largest strip. This assumes that the algorithm for 
   assigning strip sizes assigns an extra cell per x-line to the initial 
   strips. Cannot find a good way to calculate it in the previous loop */

  for (proc = 0; proc < CPT_NPROCS; proc++) {
    first_small_strip_ind[proc] = NFOLDS; 	/* Initialisation */

    for (strip = 0; strip < NFOLDS; strip ++) {
      if ((NFOLDS == first_small_strip_ind[proc]) && 
          (CPT_F_X[proc][strip] == CPT_F_X[0][NFOLDS-1])) {
        first_small_strip_ind[proc] = strip;
        break;
      }

    }			/* End for (strip) */

  }			/* End for (proc) */

#ifdef DEBUG_CALC_X_SIZES
      if (0 == cmtWorldRank) {
        for (proc = 0; proc < CPT_NPROCS; proc++) {
          fprintf (stdout, "\tproc: %i, first_small_strip_ind[proc]: %i\n",
                   proc, first_small_strip_ind[proc]);
        }
      }
#endif /* DEBUG */
} /* End calc_x_sizes */

#endif

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_read
 * 
 * Reads data for Cell sub-state from file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   fname	Character string giving filename for CA input
 *	I   stateid	Integer sub-state identifier
 * 
 * Implementation:
 *	Reads the named binary file comprising the values of the given
 *	sub-state for the entire CA. Data is distributed to all
 *	macrocells by the root process (rank == 0), using MPI
 *	communications.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_read (char *fname, int stateid) {
  struct stat statbuf;
  FILE *fin;
  CptCell *cp1, *cp2;
  unsigned char *tmp_data;
  unsigned char *ptr1;
  unsigned char *tmp_data2;
  u_int y, z, size;
  int tmp_data2_size = DIMX; 	/* Standard in this case */

  int i;
  int j;

#ifdef EVEN_DECOMP
#define ARR_SIZE ((NFOLDS > 2) ? NFOLDS : 2)
  int block_lengths[ARR_SIZE];
  MPI_Aint displacements[ARR_SIZE];
  MPI_Datatype types[ARR_SIZE];
  MPI_Aint start_address, address, extent;

  int strip, proc;
  int sendcounts[CPT_NPROCS], disps[CPT_NPROCS];

  MPI_Datatype send_strip_vec, send_fold_type, send_fold_UB_type;
  MPI_Datatype recv_strip_vec, recv_strip_UB_type;

#else /* EVEN_DECOMP */

  const int last = CPT_NPROCS-1;
  int tag;
  register int strip, proc;
  MPI_Datatype send_strip_vec_large, send_strip_vec_small;
  MPI_Datatype recv_strip_vec_large, recv_strip_vec_small;
  MPI_Request request[CPT_NPROCS*NFOLDS];
  MPI_Status status[CPT_NPROCS*NFOLDS];
  int count;

#endif /* EVEN_DECOMP */

  XDR *xdrs;

#ifdef EVEN_DECOMP
    tmp_data = (u_char *) malloc (CPT_DIMX*DIMY*DIMZ*cpt_state_size[stateid]);
#else
    tmp_data = (u_char *) malloc (CPT_DIMX[cmtWorldRank]*DIMY*DIMZ*cpt_state_size[stateid]);
#endif /* EVEN_DECOMP */

    if (!tmp_data) {
      fprintf (stderr, "cmt_read: failed to allocate read scratch space\n");
      MPI_Abort (cmtCommCommand, -1);
    }

/* Preserve initial pointer position */
    ptr1 = tmp_data;

    if (cmtWorldRank == 0) {
      tmp_data2 = (u_char *) malloc (tmp_data2_size*DIMY*DIMZ*cpt_state_size[stateid]);

      if (!tmp_data2) {
        fprintf (stderr, "cmt_read: failed to allocate read scratch space\n");
        MPI_Abort (cmtCommCommand, -1);
      }

	if((stat(fname, &statbuf)==-1) ||
	   ((statbuf.st_mode & S_IFMT)!=S_IFREG) ||
	   !(fin = fopen (fname, "rb"))) {
	  fprintf (stderr, "cmt_read: failed to open file <%s>\n", fname);
          perror ("");
          return (-1);
	} 			/* End if (!fin) */

/* Read file data in tmp_data2 */
#ifdef NO_XDR_READ
        size = fread (tmp_data2, cpt_state_size[stateid], DIMX*DIMY*DIMZ, fin);
/* Check what's read */
        if (size != DIMX*DIMY*DIMZ) {
	  fprintf (stderr, "cmt_read: WARNING, CA larger than data file <%s>\n", fname);
	  fprintf (stdout, "size: %i\n", size);
	}			/* End if (size != DIMX*DIMY*DIMZ) */

#else

        if (!(xdrs = (XDR *) malloc (sizeof (XDR)))) {
          fprintf (stderr, "cmt_read: failed to allocate xdrs memory\n");
          MPI_Abort (cmtCommCommand, -1);
        }			/* End if (!xdrs) */

        xdrstdio_create (xdrs, fin, XDR_DECODE);
        if (!(xdr_array (xdrs, (char **)&tmp_data2, &size, DIMX*DIMY*DIMZ,
              cpt_state_size[stateid], (xdrproc_t) cpt_state_XDRfn[stateid]))) {
          if (size > DIMX*DIMY*DIMZ) {
            fprintf (stderr, "WARNING: CA smaller than data file <%s>\n", fname);
          } else {
            fprintf (stderr, "cmt_read: failed to read from <%s>\n", fname);
            return (-2);
          }			/* End if (size) */

        }			/* End if (!xdr_array) */

#endif

#ifdef NO_XDR_READ
        size = fread (tmp_data2, 1, 1, fin); /* Consume EOF */
        if (!feof (fin)) {
          fprintf (stderr, "WARNING: CA smaller than data file <%s>\n", fname);
        }			/* End if (!feof (fin)) */
#else
#endif

    }				/* End if (cmtWorldRank == 0) */

/* Type strip */
#ifdef EVEN_DECOMP
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X, DIMX, cpt_state_mpidt[stateid],
                     &send_strip_vec);
    MPI_Type_commit (&send_strip_vec);

/* Type fold */
    for (i = 0; i < NFOLDS; i++) {
      types[i] = send_strip_vec;
    }

    displacements[0] = 0;
    MPI_Address (&(tmp_data2[0]), &start_address);

    for (i = 1; i < NFOLDS; i++) {
      MPI_Address (&(tmp_data2[i*CPT_NPROCS*CPT_F_X*cpt_state_size[stateid]]), &address);
      displacements[i] = address-start_address;
    }

    for (i = 0; i < NFOLDS; i++) {
      block_lengths[i] = 1;
    }

    MPI_Type_struct (NFOLDS, block_lengths, displacements, types, &send_fold_type);
    MPI_Type_commit (&send_fold_type);

/* Type fold with fixed extent for scatter */
    types[0] = send_fold_type;    
    types[1] = MPI_UB;

    displacements[0] = 0;
    MPI_Address (&(tmp_data2[0]), &start_address);
    MPI_Address (&(tmp_data2[CPT_F_X*cpt_state_size[stateid]]), &address);
    displacements[1] = address-start_address;

    block_lengths[0] = 1;
    block_lengths[1] = 1;
    
    MPI_Type_struct (2, block_lengths, displacements, types, &send_fold_UB_type);
    MPI_Type_commit (&send_fold_UB_type);

/* Type strip. Different than send_ in stride since buffer is smaller */
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X, CPT_DIMX, cpt_state_mpidt[stateid],
                     &recv_strip_vec);
    MPI_Type_commit (&recv_strip_vec);

/* Type strip with fixed extent for scatter */
    types[0] = recv_strip_vec;
    types[1] = MPI_UB;

    displacements[0] = 0;
    MPI_Address (&(tmp_data[0]), &start_address);
    MPI_Address (&(tmp_data[CPT_F_X*cpt_state_size[stateid]]), &address);
    displacements[1] = address-start_address;
/* The rest are the same as above */

    MPI_Type_struct (2, block_lengths, displacements, types, &recv_strip_UB_type);
    MPI_Type_commit (&recv_strip_UB_type);

    MPI_Scatter (tmp_data2, 1, send_fold_UB_type,
                 tmp_data, NFOLDS, recv_strip_UB_type,
                 0, cmtCommCommand);

#else /* EVEN_DECOMP */

/* For each process role (send-receive) need 2 types of vectors, a 
   small and a large one. The large can be found in CPT_F_X[last][0],
   since it is decided to distribute extraneous cells to the processes
   from last to first and place them in strips from first to last.
   Similarly, the smallest strip will be in proc 0, designated as
   first_small_strip_ind[0]. See calc_x_sizes() for more. */

/* Check above assumption */
    if (CPT_F_X[last][0] < CPT_F_X[0][first_small_strip_ind[0]]) {
      fprintf (stderr, "%s\n", 
               "cmt_read: assumption about large and small strip failed");
      fprintf (stderr, "large: %i, small: %i\n", CPT_F_X[last][0],
               CPT_F_X[0][first_small_strip_ind[0]]);
      MPI_Abort (cmtCommCommand, -1);
    }					/* End if (large < small) */

/* Send types */
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X[last][0], DIMX, 
                     cpt_state_mpidt[stateid], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_vector (DIMY*DIMZ, CPT_F_X[0][first_small_strip_ind[0]], 
                       DIMX, cpt_state_mpidt[stateid], &send_strip_vec_small);
      MPI_Type_commit (&send_strip_vec_small);
    }				/* End if (first_small_strip_ind[0]) */

/* Receive types. These are only important on the receiver side */
    MPI_Type_vector(DIMY*DIMZ, CPT_F_X[cmtWorldRank][0], CPT_DIMX[cmtWorldRank],
                     cpt_state_mpidt[stateid], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_vector (DIMY*DIMZ, 
                    CPT_F_X[cmtWorldRank][first_small_strip_ind[cmtWorldRank]], 
                       CPT_DIMX[cmtWorldRank], cpt_state_mpidt[stateid], 
                       &recv_strip_vec_small);
      MPI_Type_commit (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */

/* Send data */
    if (0 == cmtWorldRank) {
/* First run strip then run processor, so as to traverse tmp_data2 linearly.
   tmp_data2 contains the data in physical order. Going down the x-axis
   one meets first strip 0 of process 1 and then strip 1 of process 0 */

      i = 0;		/* Index to position in tmp_data2[] */
      tag = 0;		/* Tag for comm and request[] index */

      for (strip = 0; strip < NFOLDS; strip++) {
        int advance;	/* Bytes to advance arrays (calc taken out) */

        for (proc = 0; proc < CPT_NPROCS; proc++) {
          if (strip < first_small_strip_ind[proc]) {
            MPI_Isend (&tmp_data2[i], 1, send_strip_vec_large, proc, 
                       tag, cmtCommCommand, &request[tag]);
#ifdef DEBUG_CMT_READ
          fprintf (stdout, "Sent LARGE proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG */
          } else {
            MPI_Isend (&tmp_data2[i], 1, send_strip_vec_small, proc, 
                       tag, cmtCommCommand, &request[tag]);
#ifdef DEBUG_CMT_READ
          fprintf (stdout, "Sent SMALL proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG */
          }				/* End if strip */
          i += CPT_F_X[proc][strip]*cpt_state_size[stateid];
          tag++;

        }				/* End for (proc) */
      }					/* End for (strip) */
    }				/* End if (0 == cmtWorldRank) */

/* Receive data */
    i = 0;		/* Index to position in tmp_data[] */
    for (strip = 0; strip < NFOLDS; strip++) {
      tag = strip*CPT_NPROCS + cmtWorldRank;
      if (strip < first_small_strip_ind[cmtWorldRank]) {
        MPI_Recv (&tmp_data[i], 1, recv_strip_vec_large, 0, 
                  tag, cmtCommCommand, &status[cmtWorldRank]);
#ifdef DEBUG_CMT_READ
      fprintf (stdout, "Recv LARGE proc %i, strip %i, tag %i, count %i\n", cmtWorldRank, strip, tag, count);
#endif /* DEBUG */
      } else {
        MPI_Recv (&tmp_data[i], 1, recv_strip_vec_small, 0, 
                  tag, cmtCommCommand, &status[cmtWorldRank]);
#ifdef DEBUG_CMT_READ
      fprintf (stdout, "Recv SMALL proc %i, strip %i, tag %i, count %i\n", cmtWorldRank, strip, tag, count);
#endif /* DEBUG */
      }				/* End if strip */
      i += CPT_F_X[cmtWorldRank][strip]*cpt_state_size[stateid];

    }				/* End for (strip) */

    if (0 == cmtWorldRank) {
      if (MPI_SUCCESS != MPI_Waitall (CPT_NPROCS*NFOLDS, request, status)) {
        fprintf (stderr, "cmt_read: MPI_Waitall failed\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (MPI_SUCCESS != MPI_Waitall) */
    }				/* End if (0 == cmtWorldRank) */

#endif /* EVEN_DECOMP */

    cp1 = CA_REF (ca1, Radius, Radius, Radius);
    cp2 = CA_REF (ca2, Radius, Radius, Radius);

/* Copy to cps introducing haloes */
    for (z = 0; z < DIMZ; z++) {

      for (y = 0; y < DIMY; y++) {
        float *f_ptr;
#ifdef EVEN_DECOMP
        set_x_line (cp1, stateid, tmp_data, CPT_DIMX);
        set_x_line (cp2, stateid, tmp_data, CPT_DIMX);
#ifdef DEBUG_CMT_READ
        f_ptr = tmp_data;
        for (strip = 0; strip < NFOLDS; strip ++) {
          for (j = 0; j < CPT_NPROCS; j++) {
            MPI_Barrier (cmtCommCommand);
            if (j == cmtWorldRank) {
              for (i = 0; i < CPT_F_X; i ++) {
                fprintf (stdout, "%#2.1f  ", f_ptr[i]);
              }
              f_ptr += CPT_F_X;
              fflush (stdout);
            }
          }
        }
        MPI_Barrier (cmtCommCommand);
        if (0 == cmtWorldRank) fprintf (stdout, "\n");
#endif /* DEBUG */
        tmp_data += CPT_DIMX*cpt_state_size[stateid];
#else
        set_x_line (cp1, stateid, tmp_data, CPT_DIMX[cmtWorldRank]);
        set_x_line (cp2, stateid, tmp_data, CPT_DIMX[cmtWorldRank]);

#ifdef DEBUG_CMT_READ
        f_ptr = tmp_data;
        for (strip = 0; strip < NFOLDS; strip ++) {
          int j;
          for (j = 0; j < CPT_NPROCS; j++) {
            MPI_Barrier (cmtCommCommand);
            if (j == cmtWorldRank) {
              for (i = 0; i < CPT_F_X[cmtWorldRank][strip]; i ++) {
                fprintf (stdout, "%#2.1f  ", f_ptr[i]);
              }
              f_ptr += CPT_F_X[cmtWorldRank][strip];
              fflush (stdout);
            }
          }
        }
        MPI_Barrier (cmtCommCommand);
        if (0 == cmtWorldRank) fprintf (stdout, "\n");
#endif /* DEBUG */
        tmp_data += CPT_DIMX[cmtWorldRank]*cpt_state_size[stateid];
#endif /* EVEN_DECOMP */


        cp1 += CPT_X;
        cp2 += CPT_X;

      }					/* End for (y) */

      cp1 += (2*Radius)*CPT_X;
      cp2 += (2*Radius)*CPT_X;
    }					/* End for (z) */

/* Return pointer to original position */
    tmp_data = ptr1;

    free (tmp_data);

#ifdef EVEN_DECOMP
    MPI_Type_free (&send_strip_vec);
    MPI_Type_free (&send_fold_type);
    MPI_Type_free (&send_fold_UB_type);
    MPI_Type_free (&recv_strip_vec);
    MPI_Type_free (&recv_strip_UB_type);
#else 
    MPI_Type_free (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_free (&send_strip_vec_small);
    }				/* End if (first_small_strip_ind[0]) */
    MPI_Type_free (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_free (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */
#endif /* EVEN_DECOMP */

    if (0 == cmtWorldRank) {
        free (tmp_data2);
#ifdef NO_XDR_READ
#else
        xdr_destroy (xdrs);
        free (xdrs);
#endif
/* Close file. If you close it before xdr_destroy() you get a core */
        if (fclose (fin)) {
          perror ("cmt_read");
          return (-3);
        }			/* End if (fclose (fin)) */

    }				/* End if (0 == cmtWorldRank) */

    return 0;

} /* End cmt_read */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_read_global
 * 
 * Reads global CA configuration file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   fname	Character string giving filename for CA input
 * 
 * Implementation:
 *	The root process (rank == 0) reads a binary file containing
 *	all global CA information for the current generation. The data
 *	is distributed to all macrocell processes using MPI broadcast.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_read_global (char *fname) {
#define GLOBSIZE 8
  struct stat statbuf;
  FILE *fin;
  static int itmp[GLOBSIZE];
  unsigned int size;

  XDR *xdrs;

  if (cmtWorldRank == 0) {
    if ((stat(fname, &statbuf)==-1) ||
        ((statbuf.st_mode & S_IFMT)!=S_IFREG) ||
        !(fin = fopen (fname, "rb"))) {
      fprintf (stderr, "cmt_read_global: failed to open parameter input file <%s>\n", fname);
      perror ("");
      return (-1);
    }

#ifdef NO_XDR_READ
	size = fread (itmp, sizeof(itmp[0]), GLOBSIZE, fin);
    if (size != GLOBSIZE) {
      fprintf (stderr, "cmt_read_global: parameter file read failed <%s>\n",
              fname);
      fprintf (stderr, "Data read: %i, expected: %i\n", size, GLOBSIZE);
      return (-1);
    }
#else
    if (!(xdrs = (XDR *) malloc (sizeof (XDR)))) {
      fprintf (stderr, "cmt_read_global: failed to allocate xdrs memory\n");
      MPI_Abort (cmtCommCommand, -1);
    } 				/* End if (!xdrs) */
    xdrstdio_create (xdrs, fin, XDR_DECODE);
    if (!(xdr_vector (xdrs, (char *)itmp, GLOBSIZE, sizeof(int), (xdrproc_t) xdr_int))) {
      fprintf (stderr, "cmt_read_global: failed to read from <%s>\n", fname);
      MPI_Abort (cmtCommCommand, -1);
    } 				/* End if (!xdr_vector) */
          
#endif

#ifdef DEBUG
        fprintf (stdout, "Dimension:\t");
        fprintf (stdout, "%i\n", itmp [0]);
        fprintf (stdout, "Expected DIMX:\t");
        fprintf (stdout, "%i\n", itmp [1]);
        fprintf (stdout, "Expected DIMY:\t");
        fprintf (stdout, "%i\n", itmp [2]);
        fprintf (stdout, "Expected DIMZ:\t");
        fprintf (stdout, "%i\n", itmp [3]);
        fprintf (stdout, "cpt_generation:\t");
        fprintf (stdout, "%i\n", itmp [4]);
        fprintf (stdout, "NumOfStates:\t");
        fprintf (stdout, "%i\n", itmp [5]);
        fprintf (stdout, "Folds:\t\t");
        fprintf (stdout, "%i\n", itmp [6]);
        fprintf (stdout, "NumOfGlobPar:\t");
        fprintf (stdout, "%i\n", itmp [7]);
#endif /* DEBUG */

#ifdef NO_XDR_READ
	size = fread (cpt_globpar, sizeof(cpt_globpar[0]), NumOfGlobPar, fin);
    if (size != NumOfGlobPar) {
      fprintf (stderr, "cmt_read_global: parameter file read failed <%s>\n",
               fname);
      fprintf (stderr, "Global Parameters read: %i, expected: %i\n", size, NumOfGlobPar);
       return (-1);
    }

#else
    if (!(xdr_vector (xdrs, (char *)cpt_globpar, NumOfGlobPar,
          sizeof(float), (xdrproc_t) xdr_float))) {
      fprintf (stderr, "cmt_read_global: failed to read from <%s>\n", fname);
      MPI_Abort (cmtCommCommand, -1);
    } 				/* End if (!xdr_vector) */
          
#endif
  }					/* End if (cmtWorldRank == 0) */

  MPI_Bcast (itmp, GLOBSIZE, MPI_INT, 0, cmtCommCommand);

  MPI_Bcast (cpt_globpar, sizeof(cpt_globpar)/sizeof(float), MPI_FLOAT,
             0, cmtCommCommand);

  if (itmp[0] != Dimension) {
    fprintf (stderr, 
             "cmt_read_global: dimension of CA read: %i, expected: %i\n",
             itmp[0], Dimension);
    return (-1);
  }

  if (itmp[1] != DIMX) {
    fprintf (stderr, 
             "cmt_read_global: X-size of CA read: %i, expected: %i\n",
              itmp[1], DIMX);
  }

  if (itmp[2] != DIMY) {
    fprintf (stderr, 
             "cmt_read_global: Y-size of CA read: %i, expected: %i\n",
              itmp[2], DIMY);
  }

  if (itmp[3] != DIMZ) {
    fprintf (stderr, 
             "cmt_read_global: Z-size of CA read: %i, expected: %i\n",
              itmp[3], DIMZ);
  }

  if (0 > (cpt_generation = itmp[4])) {
    fprintf (stderr, "cmt_read_global: invalid cpt_generation input: %i\n",
             itmp[4]);
    cpt_generation = 0;
  }

  if (itmp[5] != NumOfStates) {
    fprintf (stderr, 
             "cmt_read_global: Number of States read: %i, expected: %i\n",
              itmp[5], NumOfStates);
    return (-1);
  }

  if (itmp[6] != NFOLDS) {
    if (0 == cmtWorldRank) {
      fprintf (stderr, 
               "cmt_read_global: Number of Folds read: %i, expected: %i\n",
                itmp[6], NFOLDS);
      fprintf (stderr, "\t\tSetting Number of Folds to %i\n", NFOLDS);
    }				/* End if (0 == cmtWorldRank) */

    itmp[6] = NFOLDS;
  }

  if (itmp[7] != NumOfGlobPar) {
    fprintf (stderr, 
             "cmt_read_global: Number of Parameters read: %i, expected: %i\n",
              itmp[7], NumOfGlobPar);
    return (-1);
  }

  if (0 == cmtWorldRank) {
#ifdef NO_XDR_READ
#else
    xdr_destroy (xdrs);
    free (xdrs);
#endif
/* Close file. If you close it before xdr_destroy() you get a core */
    if (fclose (fin)) {
      perror ("cmt_read_global");
      MPI_Abort (cmtCommCommand, -1);
    }
  }				/* End if (0 == cmtWorldRank) */

  return 0;

} /* End cmt_read_global */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_read_all
 * 
 * Reads entire CA configuration from file.
 * 
 * Parameters:
 *	I   bname	Character string giving "basename" of CA input
 * 
 * Implementation:
 *	Reads the entire CA configuration. Global file read by
 *	cmt_read_global(); Per-substate files each read by
 *	cmt_read().
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_read_all (char *bname) {
    int i, ret;
    static char fname[PATH_MAX];
    static char suffix[8];

    sprintf (fname, "%s.cpj", bname);

    if (0 > cmt_read_global (fname)) {
      return (-1);
    }                         /* End if (0 > cmt_read_global) */

    strncpy (fname, bname, sizeof(fname));

    for (i = 0; i < NumOfStates; i++) {
      sprintf (fname, "%s%c%c%c.cmt", bname,
               '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));

      if (0 > (ret = cmt_read (fname, i))) {
        return (ret);
      }				/* End if (0 > cmt_read) */
    }

    return 0;
}

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_write
 * 
 * Writes data for Cell sub-state to file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   fname	Character string giving filename for CA output
 *	I   stateid	Integer sub-state identifier
 * 
 * Implementation:
 *	Writes a file comprising the values of the given sub-state
 *      for the entice CA. Data is collected by the root process
 *      (rank == 0), using MPI communications, and written in binary
 *      format.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_write (char *fname, int stateid) {
    FILE *fout;
    CptCell *cp;
    unsigned char *tmp_data;
    unsigned char *tmp_data2;
    u_int y, z, size;
    int i;

    u_char *ptr1;

#ifdef EVEN_DECOMP
    MPI_Datatype send_strip_vec, send_strip_UB_type;
    MPI_Datatype recv_strip_vec, recv_fold_type, recv_fold_UB_type;
#define ARR_SIZE ((NFOLDS > 2) ? NFOLDS : 2)
    int block_lengths[ARR_SIZE];
    MPI_Aint displacements[ARR_SIZE];
    MPI_Datatype types[ARR_SIZE];
    MPI_Aint start_address, address, extent;

#else /* EVEN_DECOMP */

  const int last = CPT_NPROCS-1;
  int tag;
  register int strip, proc;
  MPI_Datatype send_strip_vec_large, send_strip_vec_small;
  MPI_Datatype recv_strip_vec_large, recv_strip_vec_small;
  MPI_Request request[CPT_NPROCS*NFOLDS];
  MPI_Status status[CPT_NPROCS*NFOLDS];
  int count;

#endif /* EVEN_DECOMP */

    XDR *xdrs;

#ifdef DEBUG_CMT_WRITE
  float *f_ptr;
#endif /* DEBUG */

#ifdef EVEN_DECOMP
    tmp_data = (u_char *) malloc (CPT_DIMX*DIMY*DIMZ*cpt_state_size[stateid]);

#else
    tmp_data = (u_char *) malloc (CPT_DIMX[cmtWorldRank]*DIMY*DIMZ*cpt_state_size[stateid]);
#endif /* EVEN_DECOMP */

    if (!tmp_data) {
      fprintf (stderr, "cmt_write: failed to allocate read scratch space\n");
      MPI_Abort (cmtCommCommand, -1);
    }

    if (cmtWorldRank == 0) {

      tmp_data2 = (u_char *) malloc (DIMX*DIMY*DIMZ*cpt_state_size[stateid]);

      if (!tmp_data2) {
        fprintf (stderr, "cmt_write: failed to allocate scratch space\n");
        MPI_Abort (cmtCommCommand, -1);
      }

      fout = fopen (fname, "wb");
      if (!fout) {
        fprintf (stderr, "cmt_write: failed to open file <%s>\n", fname);
        perror ("");
        MPI_Abort (cmtCommCommand, -1);
      }

#ifdef NO_XDR_WRITE
#else
      if (!(xdrs = (XDR *) malloc (sizeof(XDR)))) {
        fprintf (stderr, "cmt_write: failed to allocate xdrs memory\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (!xdrs) */

      xdrstdio_create (xdrs, fout, XDR_ENCODE);

#endif

    }				/* End if (cmtWorldRank == 0) */

    if (cpt_generation%2 == 0)
        cp = CA_REF (ca1, Radius, Radius, Radius);
    else
        cp = CA_REF (ca2, Radius, Radius, Radius);

/* Preserve initial pointer position */
    ptr1 = tmp_data;

    for (z = 0; z < DIMZ; z++) {

      for (y = 0; y < DIMY; y++) {
/* Fill tmp_data with one line */
#ifdef EVEN_DECOMP
        get_x_line (cp, stateid, tmp_data, CPT_DIMX);
        tmp_data += CPT_DIMX*cpt_state_size[stateid];
#else
        get_x_line (cp, stateid, tmp_data, CPT_DIMX[cmtWorldRank]);
        tmp_data += CPT_DIMX[cmtWorldRank]*cpt_state_size[stateid];
#endif
        cp += CPT_X;

      }					/* End for (y) */

      cp += (2*Radius)*CPT_X;
    }					/* End for (z) */

/* Return pointer to original position */
    tmp_data = ptr1;

#ifdef EVEN_DECOMP
/* Type strip */
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X, CPT_DIMX, cpt_state_mpidt[stateid],
                     &send_strip_vec);
    MPI_Type_commit (&send_strip_vec);

/* Type strip with fixed extent for gather */
    types[0] = send_strip_vec;    
    types[1] = MPI_UB;

    displacements[0] = 0;
    MPI_Address (&(tmp_data[0]), &start_address);
    MPI_Address (&(tmp_data[CPT_F_X*cpt_state_size[stateid]]), &address);
    displacements[1] = address-start_address;

    block_lengths[0] = 1;
    block_lengths[1] = 1;
    
    MPI_Type_struct (2, block_lengths, displacements, types, &send_strip_UB_type);
    MPI_Type_commit (&send_strip_UB_type);

/* Type strip. Different than send_ in stride since buffer is bigger */
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X, DIMX, cpt_state_mpidt[stateid],
                     &recv_strip_vec);
    MPI_Type_commit (&recv_strip_vec);

/* Type fold */
    for (i = 0; i < NFOLDS; i++) {
      types[i] = recv_strip_vec;
    }

    displacements[0] = 0;
    MPI_Address (&(tmp_data2[0]), &start_address);

    for (i = 1; i < NFOLDS; i++) {
      MPI_Address (&(tmp_data2[i*CPT_NPROCS*CPT_F_X*cpt_state_size[stateid]]), &address);
      displacements[i] = address-start_address;
    }

    for (i = 0; i < NFOLDS; i++) {
      block_lengths[i] = 1;
    }

    MPI_Type_struct (NFOLDS, block_lengths, displacements, types, &recv_fold_type);
    MPI_Type_commit (&recv_fold_type);

/* Type fold with fixed extent for gather */
    types[0] = recv_fold_type;    
    types[1] = MPI_UB;

    displacements[0] = 0;
    MPI_Address (&(tmp_data2[0]), &start_address);
    MPI_Address (&(tmp_data2[CPT_F_X*cpt_state_size[stateid]]), &address);
    displacements[1] = address-start_address;

    block_lengths[0] = 1;
    block_lengths[1] = 1;
    
    MPI_Type_struct (2, block_lengths, displacements, types, &recv_fold_UB_type);
    MPI_Type_commit (&recv_fold_UB_type);

    MPI_Gather (tmp_data, NFOLDS, send_strip_UB_type,
                tmp_data2, 1, recv_fold_UB_type, 0, cmtCommCommand); 

#else /* EVEN_DECOMP */

/* Send types. These are only important on the receiver side */
    MPI_Type_vector (DIMY*DIMZ,CPT_F_X[cmtWorldRank][0], CPT_DIMX[cmtWorldRank],
                     cpt_state_mpidt[stateid], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_vector (DIMY*DIMZ, 
                     CPT_F_X[cmtWorldRank][first_small_strip_ind[cmtWorldRank]],
                       CPT_DIMX[cmtWorldRank], cpt_state_mpidt[stateid], 
                       &send_strip_vec_small);
      MPI_Type_commit (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */

/* For each process role (send-receive) need 2 types of vectors, a 
   small and a large one. The large can be found in CPT_F_X[last][0],
   since it is decided to distribute extraneous cells to the processes
   from last to first and place them in strips from first to last.
   Similarly, the smallest strip will be in proc 0, designated as
   first_small_strip_ind[0]. See calc_x_sizes() for more. */

/* Check above assumption */
    if (CPT_F_X[last][0] < CPT_F_X[0][first_small_strip_ind[0]]) {
      fprintf (stderr, "%s\n", 
               "cmt_write: assumption about large and small strip failed");
      fprintf (stderr, "large: %i, small: %i\n", CPT_F_X[last][0],
               CPT_F_X[0][first_small_strip_ind[0]]);
      MPI_Abort (cmtCommCommand, -1);
    }					/* End if (large < small) */

/* Receive types. */
    MPI_Type_vector (DIMY*DIMZ, CPT_F_X[last][0], DIMX,
                     cpt_state_mpidt[stateid], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_vector (DIMY*DIMZ, CPT_F_X[0][first_small_strip_ind[0]],
                       DIMX, cpt_state_mpidt[stateid], &recv_strip_vec_small);
      MPI_Type_commit (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */

/* Receive data */
    if (0 == cmtWorldRank) {
/* First run strip then run processor, so as to traverse tmp_data2 linearly.
   tmp_data2 contains the data in physical order. Going down the x-axis
   one meets first strip 0 of process 1 and then strip 1 of process 0 */

      i = 0;		/* Index to position in tmp_data2[] */
      tag = 0;		/* Tag for comm and request[] index */

      for (strip = 0; strip < NFOLDS; strip++) {
        int advance;	/* Bytes to advance arrays (calc taken out) */

        for (proc = 0; proc < CPT_NPROCS; proc++) {
          if (strip < first_small_strip_ind[proc]) {
#ifdef DEBUG_CMT_WRITE
            fprintf (stdout, "Recv LARGE proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_CMT_WRITE */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_large, proc, 
                       tag, cmtCommCommand, &request[tag]);
          } else {
#ifdef DEBUG_CMT_WRITE
            fprintf (stdout, "Recv SMALL proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_CMT_WRITE */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_small, proc, 
                       tag, cmtCommCommand, &request[tag]);
          }				/* End if strip */
          i += CPT_F_X[proc][strip]*cpt_state_size[stateid];
          tag++;

        }				/* End for (proc) */
      }					/* End for (strip) */
    }				/* End if (0 == cmtWorldRank) */

/* Send data */
    i = 0;		/* Index to position in tmp_data[] */
    for (strip = 0; strip < NFOLDS; strip++) {
      tag = strip*CPT_NPROCS + cmtWorldRank;
      if (strip < first_small_strip_ind[cmtWorldRank]) {
#ifdef DEBUG_CMT_WRITE
        fprintf (stdout, "proc %i, Sent LARGE strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_CMT_WRITE */
        MPI_Send (&tmp_data[i], 1, send_strip_vec_large, 0, 
                  tag, cmtCommCommand);
      } else {
#ifdef DEBUG_CMT_WRITE
          fprintf (stdout, "proc %i, Sent SMALL strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_CMT_WRITE */
        MPI_Send (&tmp_data[i], 1, send_strip_vec_small, 0, 
                  tag, cmtCommCommand);
      }				/* End if strip */
      i += CPT_F_X[cmtWorldRank][strip]*cpt_state_size[stateid];

    }				/* End for (strip) */

    if (0 == cmtWorldRank) {
      if (MPI_SUCCESS != MPI_Waitall (CPT_NPROCS*NFOLDS, request, status)) {
        fprintf (stderr, "cmt_write: MPI_Waitall failed\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End  (MPI_Waitall) */
    }				/* End if (0 == cmtWorldRank) */

#endif /* EVEN_DECOMP */

/* Write data to file */
    if (0 == cmtWorldRank) {
      int i;          /* DEBUG */

/* write tmp_data2 */

      size = DIMX*DIMY*DIMZ;
#ifdef NO_XDR_WRITE
      if (size != fwrite (tmp_data2, cpt_state_size[stateid], size, fout)) {
        fprintf (stderr, "cmt_write: failed to write to <%s>\n", fname);
         MPI_Abort (cmtCommCommand, -1);
      }				/* End if fwrite */
#else 

      if (!(xdr_array (xdrs, (char **)&tmp_data2, &size, size, 
            cpt_state_size[stateid], (xdrproc_t) cpt_state_XDRfn[stateid]))) {
        fprintf (stderr, "cmt_write: failed to write to <%s>\n", fname);
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (!xdr_array) */
#endif

    }				/* End if (0 == cmtWorldRank) */

#ifdef EVEN_DECOMP
    MPI_Type_free (&send_strip_vec);
    MPI_Type_free (&send_strip_UB_type);
    MPI_Type_free (&recv_strip_vec);
    MPI_Type_free (&recv_fold_type);
    MPI_Type_free (&recv_fold_UB_type);
#else
    MPI_Type_free (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_free (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */
    MPI_Type_free (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_free (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */
#endif /* EVEN_DECOMP */
    free (tmp_data);
    if (0 == cmtWorldRank) {
      free (tmp_data2);
#ifdef NO_XDR_WRITE
#else
      xdr_destroy (xdrs);
      free (xdrs);
#endif
      if (fclose (fout)) {
        perror ("cmt_write");
        MPI_Abort (cmtCommCommand, -1);
      }
    }				/* End if (0 == cmtWorldRank) */

    return 0;

} /* End cmt_write */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_write_global
 * 
 * Writes global CA configuration file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   fname	Character string giving filename for CA output
 * 
 * Implementation:
 *	The root process, rank == 0, writes a binary file containing
 *      all global CA information for the current generation.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_write_global (char *fname) {
  FILE *fout;
#define GLOBSIZE 8
  static int itmp[GLOBSIZE];
  int i; 
  unsigned int size;

  XDR *xdrs;

  if (0 == cmtWorldRank) {
    if (!(fout = fopen (fname, "wb"))) {
      fprintf (stderr, "cmt_write_global: failed to open parameter output file <%s>\n", fname);
      perror ("");
      return (-1);
    }				/* End if (!fout) */

#ifdef NO_XDR_WRITE
#else
/* malloc XDR instance */
    if (!(xdrs = (XDR *) malloc (sizeof(XDR)))) {
      fprintf (stderr, "cmt_write_global: failed to allocate xdrs memory\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!xdrs) */
/* Create xdr stream */
    xdrstdio_create (xdrs, fout, XDR_ENCODE);

#endif
    i = 0;
    itmp[i++] = Dimension;
    itmp[i++] = DIMX;
    itmp[i++] = DIMY;
    itmp[i++] = DIMZ;
    itmp[i++] = cpt_generation;
    itmp[i++] = NumOfStates;
    itmp[i++] = NFOLDS;
    itmp[i++] = NumOfGlobPar;

    if (i > GLOBSIZE) {
      fprintf (stderr, "cmt_write_global: more data than allocated space\n");
      return (-1);
    }

#ifdef NO_XDR_WRITE
    size = fwrite (itmp, sizeof(itmp[0]), i, fout);
    if (size != i) {
      fprintf (stderr, "cmt_write_global: parameter file write failed <%s>\n",
	       fname);
      return (-1);
    }

#else
    if (!(xdr_vector (xdrs, (char *)itmp, i, sizeof(int), (xdrproc_t) xdr_int))) {
      fprintf (stderr, "cmt_write_global: failed to write to <%s>\n", fname);
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!xdr_vector) */
#endif

#ifdef NO_XDR_WRITE
    size = fwrite (cpt_globpar, sizeof(cpt_globpar[0]), NumOfGlobPar, fout);
    if (size != NumOfGlobPar) {
      fprintf (stderr, "cmt_write_global: parameter file write failed <%s>\n",
	       fname);
      return (-1);
    }

#else
    if (!(xdr_vector (xdrs, (char *)cpt_globpar, NumOfGlobPar, sizeof(float), (xdrproc_t) xdr_float))) {
      fprintf (stderr, "cmt_write_global: failed to write to <%s>\n", fname);
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!xdr_vector) */
#endif

#ifdef NO_XDR_WRITE
#else
    xdr_destroy (xdrs);
    free (xdrs);
#endif

    if (fclose (fout)) {
      perror ("cmt_write_global");
      MPI_Abort (cmtCommCommand, -1);
    }

  }				/* End if (0 == cmtWorldRank) */
  return 0;

} /* End cmt_write_global */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_create_fld
 * 
 * Creates AVS field file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   dname	Character string giving "dirname" of CA output
 *	I   bname	Character string giving "basename" of CA output
 * 
 * Implementation:
 *	If dt_list is NULL, it is malloced and initialised. Then the
 *	states of the CA are entered in the state_dt_list::statetypes member
 *	according to their datatype. Finally, fld files are written 
 *	one for each datatype.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_create_fld (char *dname, char *bname) {
  int i, j;
  static char *data;
  static char *stub;
  static char fname[PATH_MAX];
  FILE *fout;

/* Check dt_list */
  if (!dt_list) {
    fprintf (stderr, "cmt_create_fld: uninitialised dt_list, exiting\n");
    return (-1);
  } 				/* End if (!dt_list) */

/* Create files */
  for (i = 0; i < dt_list->many; i++) {
    
    if (MPI_UNSIGNED_CHAR == dt_list->statetypes[i].data) {
      data = "unsigned char";
      stub = "u_char";
    }				/* End if u_char */

    else if (MPI_CHAR == dt_list->statetypes[i].data) {
      data = "char";
      stub = "char";
    }				/* End if char */

    else if (MPI_UNSIGNED_SHORT == dt_list->statetypes[i].data) {
      data = "unsigned short";
      stub = "u_short";
    }				/* End if u_short */

    else if (MPI_SHORT == dt_list->statetypes[i].data) {
      data = "short";
      stub = "short";
    }				/* End if short */

    else if (MPI_UNSIGNED == dt_list->statetypes[i].data) {
      data = "unsigned int";
      stub = "u_int";
    }				/* End if u_int */

    else if (MPI_INT == dt_list->statetypes[i].data) {
      data = "integer";
      stub = "int";
    }				/* End if int */

    else if (MPI_FLOAT == dt_list->statetypes[i].data) {
      data = "float";
      stub = "float";
    }				/* End if float */

    else if (MPI_DOUBLE == dt_list->statetypes[i].data) {
      data = "double";
      stub = "double";
    }				/* End if double */

    else {
      fprintf (stderr, "cmt_create_fld: unknown datatype %i, exiting\n", dt_list->statetypes[i].data);
      MPI_Abort (cmtCommCommand, -1);
    }

/* Create fname */
    sprintf (fname, "%s/%s_%s.fld", (dname ? dname : "."), bname, stub);

/* Open the file */
    if (!(fout = fopen (fname, "w"))) {
      fprintf (stderr, "cmt_create_fld: failed to open AVS field output file <%s>\n", fname);
      perror ("");
      return (-1);
    }				/* End if (!fout) */

    fprintf (fout, "# AVS field file\n");
    fprintf (fout, "# CAMELot generated\n");
    fprintf (fout, "nstep = %i\n", (num_gens+1)/save_step); /* num_gens decremented */
    fprintf (fout, "ndim = %i\n", Dimension);
    fprintf (fout, "dim1 = %i\n", DIMX);
    fprintf (fout, "dim2 = %i\n", DIMY);
    fprintf (fout, "dim3 = %i\n", DIMZ);
    fprintf (fout, "nspace = 3\n");
    fprintf (fout, "veclen = %i\n", dt_list->statetypes[i].states);
    fprintf (fout, "data = %s\n", data);
    fprintf (fout, "field = uniform\n");
    fprintf (fout, "label = ");
    for (j = 0; j < dt_list->statetypes[i].states; j++) {
/* A nasty hack is included here: ENEA reported an incompatibility with
   AVS, the characters [ and ] are not allowed in the label field.
   The convention used is that these are replaced with underscores 
   and, in order to avoid matching names of arrays with variables,
   scalar variable names are postfixed with an underscore.

   The issue should be fixed in the parser (easy where cpt_state_name[]
   is created) if there is a subsequent release of the software.
   The code in this for loop should only contain
      fprintf (fout, "%s ", 
               cpt_state_name[dt_list->statetypes[i].state_ind[j]]);
*/
      char tmp_buf[PATH_MAX+2]; /* One for _, one for \0 */
      char *tmp_ptr;

      strncpy (tmp_buf, cpt_state_name[dt_list->statetypes[i].state_ind[j]], 
               PATH_MAX);
      if (NULL == (tmp_ptr = strchr(tmp_buf, '['))) {
        tmp_buf[strlen (tmp_buf)] = '_'; /* Append _ to scalar statename */
      } else {
        *tmp_ptr = '_';
        tmp_ptr = strchr(tmp_buf, ']');
        *tmp_ptr = '_';
      }
      
      fprintf (fout, "%s ", tmp_buf);
    }				/* End for (j) */
    fprintf (fout, "\n");

/* Close file */
    if (fclose (fout)) {
      perror ("cmt_create_fld");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (fclose) */

  }				/* End for (i) */

  return 0;

} /* End cmt_create_fld */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_write_fld
 * 
 * Updates fld CA configuration file (at root process, rank == 0).
 * 
 * Parameters:
 *	I   dname	Character string giving "dirname" of CA output
 *	I   bname	Character string giving "basename" of CA output
 * 
 * Implementation:
 *	The root process, rank == 0, appends information about the
 *	current iteration in the already existing field files.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_write_fld (char *dname, char *bname, int time) {
  FILE *fout;
  static char *stub;
  static char fname[PATH_MAX];
  static char cmtfname[PATH_MAX];
  int i, j;

  
  for (i = 0; i < dt_list->many; i++) {
    
    if (MPI_UNSIGNED_CHAR == dt_list->statetypes[i].data) {
      stub = "u_char";
    }				/* End if u_char */

    else if (MPI_CHAR == dt_list->statetypes[i].data) {
      stub = "char";
    }				/* End if char */

    else if (MPI_UNSIGNED_SHORT == dt_list->statetypes[i].data) {
      stub = "u_short";
    }				/* End if u_short */

    else if (MPI_SHORT == dt_list->statetypes[i].data) {
      stub = "short";
    }				/* End if short */

    else if (MPI_UNSIGNED == dt_list->statetypes[i].data) {
      stub = "u_int";
    }				/* End if u_int */

    else if (MPI_INT == dt_list->statetypes[i].data) {
      stub = "int";
    }				/* End if int */

    else if (MPI_FLOAT == dt_list->statetypes[i].data) {
      stub = "float";
    }				/* End if float */

    else if (MPI_DOUBLE == dt_list->statetypes[i].data) {
      stub = "double";
    }				/* End if double */

    else {
      fprintf (stderr, "cmt_write_fld: unknown datatype %i, exiting\n", dt_list->statetypes[i].data);
      MPI_Abort (cmtCommCommand, -1);
    }

/* Create fname */
    sprintf (fname, "%s/%s_%s.fld", (dname ? dname : "."), bname, stub);

/* Open file */
    if (!(fout = fopen (fname, "a"))) {
      fprintf (stderr, "cmt_write_fld: failed to open AVS field output file <%s>\n", fname);
      perror ("");
      return (-1);
    }

    fprintf (fout, "\ntime value = %i\n", time+1);

    for (j = 0; j < dt_list->statetypes[i].states; j++) {
      int stateid = dt_list->statetypes[i].state_ind[j]; 
/* Create cmtfname */
      sprintf (cmtfname, "%s/%c%c%c%s%c%c%c.cmt", (dname ? dname : "."),
               '0'+(time/100), 
               '0'+((time%100)/10), '0'+((time%100)%10), bname,
               '0'+(stateid/100), '0'+((stateid%100)/10), '0'+((stateid%100)%10));

      fprintf (fout, "variable %i file=%s filetype=binary\n", j+1, cmtfname);

    }				/* End for (j) */

/* Write EOT after the end of each round (ENEA request) */
    fprintf (fout, "EOT\n");

/* Close file */
    if (fclose (fout)) {
      perror ("cmt_write_fld");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (fclose) */

  }				/* End for (i) */

  return 0;

} /* End cmt_write_fld */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_write_all
 * 
 * Writes entire CA configuration to file.
 * 
 * Parameters:
 *	I   dname	Character string giving "dirname" of CA output
 *	I   bname	Character string giving "basename" of CA output
 * 
 * Implementation:
 *	Writes the entire CA configuration, with sequence number
 *      as prefix. Global file written by cmt_write_global();
 *      Per- substate files each written by cmt_write().
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_write_all (char *dname, char *bname) {
    static int lastIndex = 0;
    static char fname[PATH_MAX];
    int i;


/* Start PRJ timer */
    start_one_timer_sts (&statistics, PRJ, 0);

    if (cmtWorldRank == 0) {
      if (dname != NULL)
        sprintf (fname, "%s/%c%c%c%s.cpj", dname,
		 '0'+(lastIndex/100),
		 '0'+((lastIndex%100)/10), '0'+((lastIndex%100)%10), bname);
      else
        sprintf (fname, "%c%c%c%s.cpj",
		 '0'+(lastIndex/100),
		 '0'+(lastIndex/10), '0'+(lastIndex%10), bname);

      if (0 > cmt_write_global (fname)) {
        fprintf (stderr, "cmt_write_all: cmt_write_global failed\n");
        return (-1);
      }					/* End if (bad cmt_write_global) */
    }					/* End if (cmtWorldRank == 0) */

    for (i=0; i < NumOfStates; i++) {
      if (cmtWorldRank == 0) {
        if (dname != NULL)
	  sprintf (fname, "%s/%c%c%c%s%c%c%c.cmt", dname,
		   '0'+(lastIndex/100),
	           '0'+((lastIndex%100)/10), '0'+((lastIndex%100)%10), bname,
	           '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
	else
	  sprintf (fname, "%c%c%c%s%c%c%c.cmt",
		   '0'+(lastIndex/100),
	           '0'+((lastIndex%100)/10), '0'+((lastIndex%100)%10), bname,
	           '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
      }					/* End if (cmtWorldRank == 0) */

      if (0 > cmt_write (fname, i)) {
        fprintf (stderr, "cmt_write_all: cmt_write failed\n");
        return (-1);
      }					/* End if (bad cmt_write) */

    }					/* End for (i) */

    if (cmtWorldRank == 0) {

/* Create field files */
      if (!lastIndex) {
        if (0 > cmt_create_fld (dname, bname)) {
          fprintf (stderr, "cmt_write_all: cmt_create_fld failed\n");
          return (-1);
        }				/* End if (bad cmt_create_fld) */
      }					/* End if (!dt_list) */

/* Update field files */
      if (0 > cmt_write_fld (dname, bname, lastIndex)) {
        fprintf (stderr, "cmt_write_all: cmt_write_fld failed\n");
        return (-1);
      }					/* End if (bad cmt_create_fld) */

    }					/* End if (cmtWorldRank == 0) */

    lastIndex += 1;

/* Stop PRJ timer */
    stop_one_timer_sts (&statistics, PRJ, 0);

    return 0;

} /* End cmt_write_all */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: check_fs
 * 
 * Checks for existing files named as the ones to be written
 * 
 * Parameters:
 *	I   dname	Character string giving "dirname" of CA output
 *	I   bname	Character string giving "basename" of CA output
 * 
 * Implementation:
 *	Loop over prefixes (save index) and internally over suffixes 
 *	(stateid) to check for possible cpj and cmt files. Finally 
 *	check for fld files.
 *
 * Returns: (int)
 *	(OVER_W) if file exists
 *	(0)   	 if safe
 *	(-ve) 	 on failure.
 *
 *----------------------------------------------------------------------*/

static int check_fs (char *dname, char *bname) {
  char fname[PATH_MAX];
  char pref[3+1];
  static char *stub;
  struct stat statbuf;
  register int i, j;

  for (j = 0; j < 1000; j++) {
/* Calculate prefix */
    sprintf (pref, "%c%c%c",
             '0'+(j/100), '0'+((j%100)/10), '0'+((j%100)%10));
/* Check for cpj */
    sprintf (fname, "%s/%s%s.cpj",
             (dname ? dname : "."), pref, bname);
    if (!stat (fname, &statbuf)) {
#ifdef DEBUG
      fprintf (stdout, "check_fs: %s found\n", fname);
#endif /* DEBUG */
      return (OVER_W);
    } else if (ENOENT != errno) {
      perror ("check_fs");
      return (-1);
#ifdef DEBUG
    } else {
      fprintf (stdout, "%s not found\n", fname);
#endif /* DEBUG */
    }

/* Check for cmt */
    for (i = 0; i < NumOfStates; i++) {
      sprintf (fname, "%s/%s%s%c%c%c.cmt", 
               (dname ? dname : "."), pref, bname,
               '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
      if (!stat (fname, &statbuf)) {
#ifdef DEBUG
        fprintf (stdout, "check_fs: %s found\n", fname);
#endif /* DEBUG */
        return (OVER_W);
      } else if (ENOENT != errno) {
        perror ("check_fs");
        return (-1);
#ifdef DEBUG
      } else {
        fprintf (stdout, "%s not found\n", fname);
#endif /* DEBUG */
      }
    }				/* End for (i) */
  }				/* End for (j) */

/* Check dt_list */
  if (!dt_list) {
    fprintf (stderr, "check_fs: uninitialised dt_list, exiting\n");
    return (-1);
  }				/* End if (!dt_list) */

/* Check for fld */
  for (i = 0; i < dt_list->many; i++) {
    
    if (MPI_UNSIGNED_CHAR == dt_list->statetypes[i].data) {
      stub = "u_char";
    }				/* End if u_char */

    else if (MPI_CHAR == dt_list->statetypes[i].data) {
      stub = "char";
    }				/* End if char */

    else if (MPI_UNSIGNED_SHORT == dt_list->statetypes[i].data) {
      stub = "u_short";
    }				/* End if u_short */

    else if (MPI_SHORT == dt_list->statetypes[i].data) {
      stub = "short";
    }				/* End if short */

    else if (MPI_UNSIGNED == dt_list->statetypes[i].data) {
      stub = "u_int";
    }				/* End if u_int */

    else if (MPI_INT == dt_list->statetypes[i].data) {
      stub = "integer";
    }				/* End if int */

    else if (MPI_FLOAT == dt_list->statetypes[i].data) {
      stub = "float";
    }				/* End if float */

    else if (MPI_DOUBLE == dt_list->statetypes[i].data) {
      stub = "double";
    }				/* End if double */

    else {
      fprintf (stderr, "check_fs: unknown datatype %i, exiting\n", dt_list->statetypes[i].data);
      MPI_Abort (cmtCommCommand, -1);
    }

/* Create fname */
    sprintf (fname, "%s/%s_%s.fld", (dname ? dname : "."), bname, stub);

/* Check */
    if (!stat (fname, &statbuf)) {
#ifdef DEBUG
      fprintf (stdout, "check_fs: %s found\n", fname);
#endif /* DEBUG */
      return (OVER_W);
    } else if (ENOENT != errno) {
      perror ("check_fs");
      return (-1);
#ifdef DEBUG
    } else {
      fprintf (stdout, "%s not found\n", fname);
#endif /* DEBUG */
    }

  }				/* End for (i) */

  return (0);

} /* End check_fs */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_boundary_copy
 * 
 * Performs memory-memory copying to implement cyclic boundary
 *   condition on dimensions not distributed across processes.
 * 
 * Parameters:
 *	IO  ca		Ref to CA copy to read/write
 * 
 * Implementation:
 *	Uses memcpy (loops) for local data transfer on each
 *	non-distributed dimension.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_boundary_copy (CptCell *ca) {
  register int tz;

  memcpy (CA_REF(ca, 0, 0, 0),
	  CA_REF(ca, DIMZ, 0, 0),
	  Radius * CPT_Y * CPT_X * sizeof(CptCell));
  memcpy (CA_REF(ca, CPT_Z-Radius, 0, 0),
	  CA_REF(ca, Radius, 0, 0),
	  Radius * CPT_Y * CPT_X * sizeof(CptCell));

  for (tz=0; tz < CPT_Z; tz++) {
      memcpy (CA_REF(ca, tz, 0, 0),
	      CA_REF(ca, tz, DIMY, 0),
	      Radius * CPT_X * sizeof(CptCell));
      memcpy (CA_REF(ca, tz, CPT_Y-Radius, 0),
	      CA_REF(ca, tz, Radius, 0),
	      Radius * CPT_X * sizeof(CptCell));
  }

  return 0;

} /* End cmt_boundary_copy */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: cmt_boundary_swap
 * 
 * Performs MPI communications to exchange boundary data between macrocells.
 * 
 * Parameters:
 *	IO  ca		Ref to CA copy to read/write
 * 
 * Implementation:
 *	Determines if the previous and next cell boundaries have changed
 *	in which case active_strip is set for the strip.
 *	Uses previously set up boundary datatype, immediate
 *      receives from macrocell neighbours, sends to macrocell
 *      neighbours, then blocks for completion of receives.
 *      Tags used to differentiate transfer directions.
 *	Discerns between manual_folds/cpt_determin set/unset.
 *
 * Returns: (int)
 *	(0)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static int cmt_boundary_swap (CptCell *ca) {
  register int i;
  MPI_Request r_req[2*NFOLDS];
  MPI_Request s_req[2*NFOLDS];
  MPI_Status stats[2*NFOLDS];
  int prev_active[NFOLDS];
  int next_active[NFOLDS];

#ifdef EVEN_DECOMP
#else
  CptCell *cp;
  int tag;
#endif


  if (!manual_folds && cpt_determin) {
    for (i = 0; i < NFOLDS; i++) {
      MPI_Irecv (&prev_active[i], 1, MPI_INT, 
	         cmtPrevRank, 2*i, cmtCommBoundary, &r_req[2*i]);
      MPI_Irecv (&next_active[i], 1, MPI_INT, 
  	         cmtNextRank, 2*i+1, cmtCommBoundary, &r_req[2*i+1]);
    }				/* End for */

    for (i = 0; i < NFOLDS; i++) {
      if (cmtWorldSize-1 == cmtWorldRank) {
        MPI_Isend (&alt_intnl_bound[1][(i+NFOLDS-1)%NFOLDS], 1, MPI_INT, 
                  cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
      } else {
          MPI_Isend (&alt_intnl_bound[1][i], 1, MPI_INT, 
                    cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
      }

      if (0 == cmtWorldRank) {
        MPI_Isend (&alt_intnl_bound[0][(i+1)%NFOLDS], 1, MPI_INT, 
                  cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
      } else {
        MPI_Isend (&alt_intnl_bound[0][i], 1, MPI_INT, 
                  cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
      }
    }				/* End for */

    MPI_Waitall (2*NFOLDS, r_req, stats);
    MPI_Waitall (2*NFOLDS, s_req, stats);

#ifdef DEVELOP
    for (i = 0; i < cmtWorldSize; i++) {
      int j;
      MPI_Barrier (cmtCommCommand);

      if (cmtWorldRank == i) {

        for (j = 0; j < NFOLDS; j++) {
          fprintf (stdout, "%i: prev_active[%i] = %i\t", 
                   cmtWorldRank, j*cmtWorldSize+cmtWorldRank, prev_active[j]);
          fprintf (stdout, "next_active[%i] = %i\n", 
                   j*cmtWorldSize+cmtWorldRank, next_active[j]);
        } 			/* End for */

      } 			/* End if (cmtWorldRank == i) */

    } 				/* End for */
#endif /* DEVELOP */

#ifdef EVEN_DECOMP
/* --- Issue Receives */
    for (i = 0; i < NFOLDS; i++) {

      MPI_Irecv (CA_REF(ca, 0, 0, i*CPT_S_X), 1, cmtBoundaryType,
                 cmtPrevRank, 2*i, cmtCommBoundary, &r_req[2*i]);
      if (prev_active[i]) {
        active_strip[i] = 1;

#ifdef DEVELOP
        fprintf (stdout, "%i: expect to recv from prev, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i);
        fprintf (stdout, "%i: active_strip[%i] = 1\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank);
#endif /* DEVELOP */

      }				/* End if (prev_active[i]) */

      MPI_Irecv (CA_REF(ca, 0, 0, (i+1)*CPT_S_X-Radius), 1, cmtBoundaryType,
                 cmtNextRank, 2*i+1, cmtCommBoundary, &r_req[2*i+1]);
      if (next_active[i]) {
        active_strip[i] = 1;

#ifdef DEVELOP
        fprintf (stdout, "%i: expect to recv from next, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i+1);
        fprintf (stdout, "%i: active_strip[%i] = 1\n", cmtWorldRank, i*cmtWorldSize+cmtWorldRank);
#endif /* DEVELOP */

      }				/* End if (next_active[i]) */

    }				/* End for */

/* --- Issue Sends */
    for (i = 0; i < NFOLDS; i++) {

      if (cmtWorldSize-1 == cmtWorldRank) {

        MPI_Isend (CA_REF(ca, 0, 0, CPT_F_X+((i+NFOLDS-1)%NFOLDS)*CPT_S_X), 1, 
                  cmtBoundaryType, cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
#ifdef DEVELOP
        fprintf (stdout, "%i: Strip: %i sent data to next, tag: %i\n", 
                 cmtWorldRank, ((i+NFOLDS-1)%NFOLDS)*cmtWorldSize+cmtWorldRank, 2*i);
#endif /* DEVELOP */

      } 			/* End if (cmtWorldSize-1 == cmtWorldRank) */

      else {			/* cmtWorldSize-1 != cmtWorldRank */

        MPI_Isend (CA_REF(ca, 0, 0, CPT_F_X+i*CPT_S_X), 1, 
                  cmtBoundaryType, cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
#ifdef DEVELOP
        fprintf (stdout, "%i: Strip: %i sent data to next, tag: %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i);
#endif /* DEVELOP */

      } 			/* End if (cmtWorldSize-1 != cmtWorldRank) */

      if (0 == cmtWorldRank) {

        MPI_Isend (CA_REF(ca, 0, 0, ((i+1)%NFOLDS)*CPT_S_X+Radius), 1, 
                    cmtBoundaryType, cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
#ifdef DEVELOP
        fprintf (stdout, "%i: Strip: %i sent data to prev, tag: %i\n", 
                 cmtWorldRank, ((i+1)%NFOLDS)*cmtWorldSize+cmtWorldRank, 2*i+1);
#endif /* DEVELOP */

      } 			/* End if (0 == cmtWorldRank) */

      else {			/* 0 != cmtWorldRank */

        MPI_Isend (CA_REF(ca, 0, 0, i*CPT_S_X+Radius), 1, 
                  cmtBoundaryType, cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
#ifdef DEVELOP
        fprintf (stdout, "%i: Strip: %i sent data to prev, tag: %i\n", cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i+1);
#endif /* DEVELOP */

      } 			/* End if (0 != cmtWorldRank) */

    }				/* End for */

    MPI_Waitall (2*NFOLDS, r_req, stats);
    MPI_Waitall (2*NFOLDS, s_req, stats);

  }				/* End if (!manual_folds && cpt_determin) */

  else {			/* manual_folds or !cpt_determin*/

    for (i = 0; i < NFOLDS; i++) {
      MPI_Irecv (CA_REF(ca, 0, 0, i*CPT_S_X), 1, cmtBoundaryType,
	         cmtPrevRank, 2*i, cmtCommBoundary, &r_req[2*i]);
      MPI_Irecv (CA_REF(ca, 0, 0, (i+1)*CPT_S_X-Radius), 1, cmtBoundaryType,
	         cmtNextRank, 2*i+1, cmtCommBoundary, &r_req[2*i+1]);
    }				/* End for */

    for (i = 0; i < NFOLDS; i++) {
      if (cmtWorldSize-1 == cmtWorldRank) {
        MPI_Isend (CA_REF(ca, 0, 0, CPT_F_X+((i+NFOLDS-1)%NFOLDS)*CPT_S_X), 1, 
                  cmtBoundaryType, cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
      } else {
        MPI_Isend (CA_REF(ca, 0, 0, CPT_F_X+i*CPT_S_X), 1, 
                  cmtBoundaryType, cmtNextRank, 2*i, cmtCommBoundary, &s_req[2*i]);
      }				/* End if */

      if (0 == cmtWorldRank) {
        MPI_Isend (CA_REF(ca, 0, 0, ((i+1)%NFOLDS)*CPT_S_X+Radius), 1, 
                  cmtBoundaryType, cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
      } else {
        MPI_Isend (CA_REF(ca, 0, 0, i*CPT_S_X+Radius), 1, 
                  cmtBoundaryType, cmtPrevRank, 2*i+1, cmtCommBoundary, &s_req[2*i+1]);
      }				/* End if */
    }				/* End for */

    MPI_Waitall (2*NFOLDS, r_req, stats);
    MPI_Waitall (2*NFOLDS, s_req, stats);

  }				/* End manual_folds */

#else /* EVEN_DECOMP */

/* Initialise cp */
    cp = CA_REF (ca, 0, 0, 0);
/* --- Issue Receives */
    for (i = 0; i < NFOLDS; i++) {

      MPI_Irecv (cp, 1, cmtBoundaryType,
                 cmtPrevRank, 2*i, cmtCommBoundary, &r_req[2*i]);
/* Advance cp */
      cp += CPT_S_X[cmtWorldRank][i];

/* Update active_strip[] */
      if (prev_active[i]) {
        active_strip[i] = 1;

#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: expect to recv from prev, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

      }				/* End if (prev_active[i]) */

      MPI_Irecv (cp-Radius, 1, cmtBoundaryType,
                 cmtNextRank, 2*i+1, cmtCommBoundary, &r_req[2*i+1]);
      if (next_active[i]) {
        active_strip[i] = 1;

#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: expect to recv from next, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i+1);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

      }				/* End if (next_active[i]) */

    }				/* End for */

#ifdef DEBUG_CMT_BOUNDARY_SWAP
    MPI_Barrier (cmtCommBoundary);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

/* Initialise cp */
      cp = CA_REF (ca, 0, 0, 0);

/* --- Issue Sends */
    for (i = 0; i < NFOLDS; i++) {

      if (0 == cmtWorldRank) {
        tag = 2*((i+NFOLDS-1)%NFOLDS)+1;
      } else {
        tag = 2*i+1;
      }				/* End if (0 == cmtWorldRank) */

      MPI_Isend (cp+Radius, 1, cmtBoundaryType, 
                 cmtPrevRank, tag, cmtCommBoundary, &s_req[tag]);
#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: Strip: %i sent data to prev, tag: %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, tag);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

      if (cmtWorldSize-1 == cmtWorldRank) {
        tag = 2*((i+1)%NFOLDS);
      } else {
        tag = 2*i;
      }				/* End if (cmtWorldSize-1 == cmtWorldRank) */

      MPI_Isend (cp+CPT_F_X[cmtWorldRank][i], 1, cmtBoundaryType, 
                 cmtNextRank, tag, cmtCommBoundary, &s_req[tag]);
#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: Strip: %i sent data to next, tag: %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, tag);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

/* Advance cp */
      cp += CPT_S_X[cmtWorldRank][i];

    }				/* End for */

    MPI_Waitall (2*NFOLDS, r_req, stats);
    MPI_Waitall (2*NFOLDS, s_req, stats);

  }				/* End if (!manual_folds && cpt_determin) */

  else {			/* manual_folds or !cpt_determin*/

/* Initialise cp */
    cp = CA_REF (ca, 0, 0, 0);
/* --- Issue Receives */
    for (i = 0; i < NFOLDS; i++) {

      MPI_Irecv (cp, 1, cmtBoundaryType,
                 cmtPrevRank, 2*i, cmtCommBoundary, &r_req[2*i]);
/* Advance cp */
      cp += CPT_S_X[cmtWorldRank][i];

#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: expect to recv from prev, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

      MPI_Irecv (cp-Radius, 1, cmtBoundaryType,
                 cmtNextRank, 2*i+1, cmtCommBoundary, &r_req[2*i+1]);

#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: expect to recv from next, fold: %i, tag %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, 2*i+1);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

    }				/* End for */

#ifdef DEBUG_CMT_BOUNDARY_SWAP
    MPI_Barrier (cmtCommBoundary);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

/* Initialise cp */
      cp = CA_REF (ca, 0, 0, 0);

/* --- Issue Sends */
    for (i = 0; i < NFOLDS; i++) {

      if (0 == cmtWorldRank) {
        tag = 2*((i+NFOLDS-1)%NFOLDS)+1;
      } else {
        tag = 2*i+1;
      }				/* End if (0 == cmtWorldRank) */

      MPI_Isend (cp+Radius, 1, cmtBoundaryType, 
                 cmtPrevRank, tag, cmtCommBoundary, &s_req[tag]);
#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: Strip: %i sent data to prev, tag: %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, tag);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

      if (cmtWorldSize-1 == cmtWorldRank) {
        tag = 2*((i+1)%NFOLDS);
      } else {
        tag = 2*i;
      }				/* End if (cmtWorldSize-1 == cmtWorldRank) */

      MPI_Isend (cp+CPT_F_X[cmtWorldRank][i], 1, cmtBoundaryType, 
                 cmtNextRank, tag, cmtCommBoundary, &s_req[tag]);
#ifdef DEBUG_CMT_BOUNDARY_SWAP
        fprintf (stdout, "%i: Strip: %i sent data to next, tag: %i\n", 
                 cmtWorldRank, i*cmtWorldSize+cmtWorldRank, tag);
#endif /* DEBUG_CMT_BOUNDARY_SWAP */

/* Advance cp */
      cp += CPT_S_X[cmtWorldRank][i];

    }				/* End for */

    MPI_Waitall (2*NFOLDS, r_req, stats);
    MPI_Waitall (2*NFOLDS, s_req, stats);
  }				/* End manual_folds */
#endif /* EVEN_DECOMP */
  return 0;
}

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: init_boundaries
 *
 * Sets active_strip[] and alt_intnl_bound[][]for all strips before executing 
 * boundary copy/swap on both copies of CA Engine, then cancels 
 * alt_intnl_bound[][] and alt_strip[].
 * Must only be used when the system is brought to a new state.
 *
 * Parameters:
 *      void
 *
 * Implementation:
 * 	trivial
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void init_boundaries (void) {
  int strip;

  if (!manual_folds && cpt_determin) {

    for (strip = 0; strip < NFOLDS; strip++) {
      active_strip[strip] = 1;
      alt_intnl_bound[0][strip] = 1;
      alt_intnl_bound[1][strip] = 1;
    }				/* End for (strip) */

  }				/* End if (!manual_folds && cpt_determin) */

/* Start timer BOUND */
  start_one_timer_sts (&statistics, BOUND, 0);
  cmt_boundary_copy (ca1);
  cmt_boundary_swap (ca1);
/* Stop timer BOUND */
  stop_one_timer_sts (&statistics, BOUND, 0);
/* Start timer BOUND */
  start_one_timer_sts (&statistics, BOUND, 0);
  cmt_boundary_copy (ca2);
  cmt_boundary_swap (ca2);
/* Stop timer BOUND */
  stop_one_timer_sts (&statistics, BOUND, 0);

  if (!manual_folds && cpt_determin) {
/* Zero alt_intnl_bound and alt_strip */
    memset ((char *) alt_intnl_bound, 0, 2*NFOLDS*sizeof (int));
    memset ((char *) alt_strip, 0, NFOLDS*sizeof (int));
  }				/* End if (!manual_folds && cpt_determin) */

} /* End init_boundaries */


#ifdef EVEN_DECOMP
/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: line2fold
 * 
 * Get data destined for linear, one fold interpretation and rearrange them 
 * so that per PE data are grouped together
 * 
 * Parameters:
 *	I    source	U_char ptr containing the original data
 *	O    target	U_char ptr containing the original data
 *	I    e_size	Size_t size of actual element data
 * 
 * Implementation:
 *	Traverse the target array linearly and put in the data
 * 	for each processor by accessing source, the original data array,
 *	in CPT_N_PROCS*CPT_F_X*e_size strides.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void line2fold (const u_char *source, u_char *target, size_t e_size) {
  register size_t halo = Radius*e_size;		/* halo */
  register size_t strip_size = CPT_F_X*e_size;	/* size of actual strip data */
  register int proc; /* index to the processor whose data are accessed 
			by tmp_data */
  register int fold; /* index to the fold portion of the data accessed 
			by tmp_data */

  for (proc = 0; proc < cmtWorldSize; proc++) {

    for (fold = 0; fold < NFOLDS; fold++) {

      memcpy (&(target[fold*strip_size]), 
              &(source[fold*CPT_NPROCS*strip_size]), strip_size);
    }			/* End for (fold) */

    target += NFOLDS*strip_size;
    source += strip_size;
  }			/* End for (proc) */

} /* End line2fold */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: fold2line
 *
 * Change folded data to normal external representation.
 * 
 * Parameters:
 *	I    source	U_char ptr containing the original data
 *	O    target	U_char ptr containing the final data
 *	I    e_size	Size_t size of actual element data
 * 
 * Implementation: 
 *	Traverse the source array linearly and write the data
 * 	for each fold by accessing the target data array
 *	in CPT_N_PROCS*CPT_F_X*e_size strides.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void fold2line (const u_char *source, u_char *target, size_t e_size) {
  register size_t strip_size = CPT_F_X*e_size; /* size of actual strip data */
  register int proc; /* index to the processor whose data are accessed 
			by target */
  register int fold; /* index to the fold portion of the data accessed 
			by target */

  for (proc = 0; proc < cmtWorldSize; proc++) {

    for (fold = 0; fold < NFOLDS; fold++) {

      memcpy (&(target[fold*CPT_NPROCS*strip_size]), 
              &(source[fold*strip_size]), strip_size);
    }			/* End for (fold) */

    source += NFOLDS*strip_size;
    target += strip_size;
  }			/* End for (proc) */

} /* End fold2line */

#endif /* EVEN_DECOMP */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: rv
 * 
 * Calls protocol function to serve the request, then transmits the 
 * acknowledgement to the GUI (where applicable).
 * 
 * Parameters:
 *	I   request	Req_code of received request
 * 
 * Implementation:
 * 	switch statement calling appropriate function. proc 0 returns 
 *  	the result of the ack, the rest the status of the serv function
 * 	called. The two values may differ if writen in send_ack () fails,
 *	but in that case send_ack will MPI_Abort.
 *
 * Returns: (req_code)
 *	(ve)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code rv (req_code request) {
  static int started = 0;	/* Flag denoting if EVOLVE or LOOP have 
				   been called */
  static int paused = 0;	/* Flag denoting if PAUSE has been called */
  static int init_gen = 0;	/* Flag denoting whether cpt_generation
				   has been initialised and must be
				   preserved */
  size_t size;
  req_code status;

  switch (request) {
    case SAVE_REQUEST: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_save_request ();
      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case SAVE_REQUEST: */

    case SET_FOLD: {
#ifdef DEVELOP
      fprintf (stdout, "Request: %i \n", request);
#endif /* DEVELOP */
      status = serv_set_fold ();
      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case SET_FOLD: */

    case SET_LOAD: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_set_load ();

/*--- Initialise boundaries, unset alt_intnl_bound[][] and set alt_strip[] */
      init_boundaries ();

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case SET_LOAD: */

    case VIEW_STATE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_view_state ();
      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case VIEW_STATE: */

    case SET_STATE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_set_state ();

/*--- Initialise boundaries, unset alt_intnl_bound[][] and set alt_strip[] */
      init_boundaries ();

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End SET_STATE: */

    case SET_PARAM: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }                         /* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_set_param ();
      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End SET_PARAM: */

    case GET_PARAM: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }                         /* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_get_param ();
      return (status);
    }				/* End GET_PARAM: */

    case EVOLVE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }                         /* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      if (started && !paused) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine running, request ignored\n");
        }			/* End if (0 == cmtWorldRank) */
        return (IGNORED);
      }				/* End if (started) */

      if (0 == cmtWorldRank) {
/* Get num_gens */
        size = sizeof (num_gens);
        if (size != readn (prot_sockfd, (char *) &num_gens, size)) {
          fprintf (stderr, "rv: readn error, exiting\n");
          MPI_Abort (cmtCommCommand, -1);
        }			/* End if readn */
      }				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get num_gens */
      MPI_Bcast ((int *) &num_gens, 1, MPI_INT, 0, cmtCommCommand);

      set_gen (&vis_list, cpt_generation);

/*--- Update local objects */
      started = 1;
      paused = 0;
      init_gen = 0;

/*--- Update cyclic and macrocell boundaries */
      init_boundaries ();

#ifdef DEVELOP
      fprintf (stdout, "Traversing: ");
      traverse (vis_list, '#');
#endif /* DEVELOP */

      return (EVOLVE);
    }				/* End case EVOLVE: */

    case BATCH: {
      if (started || paused || prot_sockfd) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine running, request ignored\n");
        }			/* End if (0 == cmtWorldRank) */
        return (IGNORED);
      }				/* End if (started) */

#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }
#endif /* DEVELOP */

/*--- Update global objects */

/*--- Update local objects */
      started = 1;
      paused = 0;
      init_gen = 0;

/*--- Update cyclic and macrocell boundaries */
      init_boundaries ();

      return (BATCH);
    }				/* End case BATCH: */

    case LOOP: {
      if (started && !paused) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine running, request ignored\n");
        }			/* End if (0 == cmtWorldRank) */
        return (IGNORED);
      }				/* End if (started) */

/*--- Update global objects */
      set_gen (&vis_list, cpt_generation);

      num_gens = INFINITE;

/*--- Update local objects */
      started = 1;
      paused = 0;
      init_gen = 0;

/*--- Initial cyclic and macrocell boundaries */
      init_boundaries ();
      return (LOOP);
    }                         /* End case LOOP: */
    
    case RESUME: {
      if (!paused) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine not paused, request ignored\n");
        }                       /* End if (0 == cmtWorldRank) */
        return (IGNORED);
      }                         /* End if (!paused) */
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      } 			/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */

/*--- Update local objects */
      paused = 0;
      init_gen = 0;
      set_gen (&vis_list, cpt_generation);

/* Do not alter cpt_generation, restore num_gens to initial value
   (run decrements it) */
#ifdef BUG
      if (INFINITE != num_gens) {
        num_gens += cpt_generation;
      } 			/* End if (INFINITE != num_gens) */
#endif
      return (RESUME);
    }				/* End case RESUME: */

    case TERMINATE: {
      if (!started) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine not started yet, request ignored\n");
          return (send_ack (IGNORED));
        }			/* End if (0 == cmtWorldRank) */
        return (IGNORED);
      }				/* End if (!started) */
      started = 0;
      paused = 0;
      status = serv_terminate ();
      if (0 == cmtWorldRank) {
#ifdef DEVELOP
        fprintf (stdout, "Request: %i\n", request);
#endif /* DEVELOP */

        return (send_ack (status));
      } 			/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case TERMINATE: */

    case PAUSE: {
      if (!started || paused) {
        if (0 == cmtWorldRank) {
          fprintf (stderr, "CA Engine not started yet, request ignored\n");
        }                       /* End if (0 == cmtWorldRank) */
        return (IGNORED);	/* No Ack in this case */
      }				/* End if (!started) */
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      paused = 1;
      return (request);
    }				/* End case PAUSE: */

    case ADD_PLANE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */

      status = serv_add_plane ();

      return (status);
    }				/* End case ADD_PLANE: */

    case DEL_PLANE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */

      status = serv_del_plane ();

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case DEL_PLANE:  */

    case READ_PROJECT: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_proj_read ();

      init_gen = (status > 0);

#ifdef DEBUG
      fprintf (stdout, "init_gen: %i \n", init_gen);
#endif /* DEBUG */

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case READ_PROJECT:  */

    case SAVE_PROJECT: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_proj_save ();

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case SAVE_PROJECT:  */

    case PERIODIC_SAVE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i \n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      status = serv_periodic_save ();

      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case PERIODIC_SAVE:  */

    case SET_MINMAX: {
#ifdef DEVELOP
      fprintf (stdout, "Request: %i \n", request);
#endif /* DEVELOP */
      status = serv_set_minmax ();
      if (0 == cmtWorldRank) {
        return (send_ack (status));
      }				/* End if (0 == cmtWorldRank) */
      return (status);
    }				/* End case SET_MINMAX: */

    case FINISHED: {
#ifdef DEVELOP
      fprintf (stdout, "Request: %i\n", request);
#endif /* DEVELOP */
      if (!started) {
        fprintf (stderr, "CA Engine not started yet, request ignored\n");
        return (IGNORED);       /* No Ack in this case */
      }                         /* End if (!started) */

#ifdef PROFILING
      monitor ((int (*)()) 0, (int (*)()) 0, (WORD *) 0, 0, 0);
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }				/* End if (0 == cmtWorldRank) */
#endif

      started = 0;

      return (FINISHED);
    }                           /* End case FINISHED: */

    case EXITCODE: {
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "Request: %i\n", request);
      }                       /* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */
      return (EXITCODE);
    }                           /* End case EXITCODE: */

    default: {
      if (0 == cmtWorldRank) {
        fprintf (stderr, "Request %i unknown, ignored\n", request);
      }
      return (IGNORED);		/* No Ack in this case */
    }				/* End default */
  }				/* End switch */
/* NOTREACHED */
  return (request);

} /* End rv */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: run
 * 
 * The main processing loop of the CA Engine. Contains socket routine
 * to check for incoming messages from GUI.
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Cut off from main() and augmented to contain socket polling
 *	for incoming GUI messages, as well as visualisation. Request 
 *	is broadcast and handled by rv(). Infinite loops are handled 
 *	(although for (;;) seems finite), by not decreasing num_gens.
 *	Processing loop (3 nested for()) calls cpt_func and updates 
 *	boundaries. In the end rv (FINISHED) is called to update 
 *	'started' static variable in rv().
 *
 * Returns: (req_code)
 *	(ve)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code run (void) {
#ifdef EVEN_DECOMP
  register CptCell *cp1, *cp2;
  int 	pos;
  int 	maxfdp1; 
  int 	isset;
  req_code request;
  struct timeval timeout;
  fd_set fdvar;
  size_t size;
  cell 	*vis_ptr;
  register int z_offset = 1;	/* cpt_z must start with 1 */
  register int y_offset = 1;	/* cpt_y must start with 1 */
  register int x_offset = cmtWorldRank*CPT_F_X+1; /* likewise but x is divided
						     if folds and processors */
  register int my_x;		/* Index */
  register int foldsize = DIMX/NFOLDS; /* One off calculation */
  register int z_max = DIMZ+1;	/* for loop condition calculation */
  register int y_max = DIMY+1;	/* for loop condition calculation */
  int 	strip;				/* Current strip accessed */
  const int l_int_bound = Radius; 	/* Left internal boundary */
  const int r_int_bound = CPT_F_X-1-Radius; 	/* Right internal boundary */
  register int cell_size = sizeof (CptCell);
  req_code steer_res;			/* Result of steering function */
#else /* EVEN_DECOMP */
  register CptCell *cp1, *cp2;
  int 	pos;
  int 	maxfdp1; 
  int 	isset;
  int i;
  req_code request;
  struct timeval timeout;
  fd_set fdvar;
  size_t size;
  cell 	*vis_ptr;
  register int z_offset = 1;	/* cpt_z must start with 1 */
  register int y_offset = 1;	/* cpt_y must start with 1 */
  register int x_offset; 	/* likewise but x need caclulation */
  register int my_x;		/* Index */
  register int z_max = DIMZ+1;	/* for loop condition calculation */
  register int y_max = DIMY+1;	/* for loop condition calculation */
  int 	strip;				/* Current strip accessed */
  const int l_int_bound = Radius; 	/* Left internal boundary */
  int r_int_bound; 			/* Right internal boundary */
  register int cell_size = sizeof (CptCell);
  req_code steer_res;			/* Result of steering function */
#endif /* EVEN_DECOMP */

/* Start TOTAL timer */
  start_one_timer_sts (&statistics, TOTAL, cpt_generation);

  if (0 == cmtWorldRank) {
#ifdef DEVELOP
    fprintf (stdout, "run cpt_gen: %i,", cpt_generation);
    fprintf (stdout, " num_gens: %i)\n", num_gens);
#endif /* DEVELOP */

/* Initialise timeout struct. Polling requires members set to 0 */
    timeout.tv_sec = (long) 0;
    timeout.tv_usec = (long) 0;

    maxfdp1 = prot_sockfd + 1; 	/* Used with select */
  }				/* End if (0 == cmtWorldRank) */

  for (pos = 1; num_gens > 0;) {

    do {
      request = NULL_CODE; 		/* Initialise every time! */

      if ((0 == cmtWorldRank) && prot_sockfd) {
        FD_ZERO (&fdvar);
        FD_SET (prot_sockfd, &fdvar);

/* Poll */
        if (select (maxfdp1, &fdvar,(fd_set *) 0, (fd_set *) 0, &timeout) < 0) {
          fprintf (stderr, "run: select failed. Exiting...\n");
          MPI_Abort (cmtCommCommand, -1);
        }			/* End if select */

        if ((isset = FD_ISSET (prot_sockfd, &fdvar)) < 0) {
          fprintf (stderr, "run: FD_ISSET error. Exiting...\n");
          MPI_Abort (cmtCommCommand, -1);
        } 			/* End FD_ISSET error */

        else if (0 != isset) { 	/* Message waiting in socket */
/* Get request */
          size = sizeof (request);
          if (size != readn (prot_sockfd, (char *) &request, size)) {
            fprintf (stderr, "run: readn error, exiting\n");
            MPI_Abort (cmtCommCommand, -1);
          }			/* End if readn */

        } 			/* End if (0 != isset) */

      }				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get request */
      MPI_Bcast ((int *) &request, 1, MPI_INT, 0, cmtCommCommand);
      if (0 != request) {
        stop_one_timer_sts (&statistics, TOTAL, cpt_generation);
        rv (request);
        start_one_timer_sts (&statistics, TOTAL, cpt_generation);
      }				/* End if (0 != request) */ 

/* Process all messages pending in socket except if a flow control one occurs */
    } while ((0 != request) && (TERMINATE != request) 
         && (EXITCODE != request) && (PAUSE != request));
				/* End do */

    if ((TERMINATE == request) || (EXITCODE == request) || (PAUSE == request)) {
       stop_one_timer_sts (&statistics, TOTAL, cpt_generation);
       return (request);
    }				/* End if stop */

    /*--- Generation tracking on stdout */
#ifdef DEVELOP
    if ((cmtWorldRank == 0) && (cpt_generation % 1) == 0) {
      fprintf(stdout, "%5d ", cpt_generation );
      fflush (stdout);
      pos += 6;

      if (pos > 60) {
	pos = 1;
	printf( "\n" );
      }

    }				/* End if (cmtWorldRank && 10) */
#endif /* DEVELOP */

    /*--- Set up CA pointers, toggle between CA copies */
    if (0 == cpt_generation % 2) {
      cp1 = CA_REF (ca1, Radius, Radius, 0);
      cp2 = CA_REF (ca2, Radius, Radius, 0);
    }

    else {
      cp1 = CA_REF (ca2, Radius, Radius, 0);
      cp2 = CA_REF (ca1, Radius, Radius, 0);
    }

#ifdef DEVELOP
        for (strip = 0; strip < NFOLDS; strip++) {
          if (!active_strip[strip]) {
            fprintf (stdout, "Strip %i inactive\t", strip*cmtWorldSize+cmtWorldRank);
          }			/* End if */
          else {
            fprintf (stdout, "Strip %i active\t", strip*cmtWorldSize+cmtWorldRank);
          }			/* End else */
        }			/* End for */

#endif /* DEVELOP */

/* Zero alt_intnl_bound and alt_strip */
    memset ((char *) alt_intnl_bound, 0, 2*NFOLDS*sizeof (int));
    memset ((char *) alt_strip, 0, NFOLDS*sizeof (int));

#ifdef EVEN_DECOMP
#else
/* x_offset calculation. This is calculated by estimating the distance
   of the first element of the processor from the beginning of the data
   assuming that the data are folded but contain no haloes */
    x_offset = 1;
    for (i = 0; i < cmtWorldRank; i++) {
      x_offset += CPT_F_X[i][0];
    }				/* End for (i) */

#ifdef DEBUG_RUN
        for (i = 0; i < CPT_NPROCS; i++) {
          MPI_Barrier (cmtCommCommand);
          if (i == cmtWorldRank) {
            fprintf (stdout, "%i: x_offset: %i\n", cmtWorldRank, x_offset);
          }
        }				/* End for (i) */
#endif /* DEBUG_RUN */
#endif /* EVEN_DECOMP */

    /*--- CA Evolution loop */
    /* Start Update Function Timer */
    start_one_timer_sts (&statistics, FUNC, 0);

    for (cpt_z = z_offset; cpt_z < z_max; cpt_z++) {

      for (cpt_y = y_offset; cpt_y < y_max; cpt_y++) {

#ifdef EVEN_DECOMP 
#else 
        cpt_x = x_offset;
#endif /* EVEN_DECOMP */

        for (strip = 0; strip < NFOLDS; strip++) {
          cp1 += Radius;
          cp2 += Radius;

          if (!cpt_determin || active_strip[strip]) {
#ifdef EVEN_DECOMP 
          cpt_x = x_offset + foldsize*strip;
#else
/* Right internal boundary calculation */
          r_int_bound = CPT_F_X[cmtWorldRank][strip]-1-Radius;
#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP 
	    for (my_x = 0; my_x < CPT_F_X; my_x++, cpt_x++) {
#else
	    for (my_x = 0; my_x < CPT_F_X[cmtWorldRank][strip];my_x++,cpt_x++) {
#ifdef DEBUG_RUN
              for (i = 0; i < CPT_NPROCS; i++) {
                if (i == cmtWorldRank) {
                  fprintf (stdout, "%i: cpt_x: %i | ", cmtWorldRank, cpt_x);
                  fflush (stdout);
                }
        }				/* End for (i) */
#endif /* DEBUG_RUN */
#endif /* EVEN_DECOMP */
	  /*--- CARPET-defined transition function. Alters cp2 */
              cpt_func (cp1, cp2);

/* Check if function altered the cell */
              if (!manual_folds && cpt_determin) {

                if ((my_x < l_int_bound) && (!alt_intnl_bound[0][strip])) {
                  alt_intnl_bound[0][strip] = (memcmp (cp1, cp2, cell_size) && !cpt_thresh (cp2));
#ifdef DEVELOP
                  if (alt_intnl_bound[0][strip])
                    fprintf (stdout, "%i: my_x: %i, strip: %i, cpt_y: %i, cpt_z: %i: l_bound changed\n", cmtWorldRank, my_x, strip*cmtWorldSize+cmtWorldRank, cpt_y, cpt_z);
#endif /* DEVELOP */
                }			/* End if (l_int_bound unchanged) */

                else if ((my_x > r_int_bound) && (!alt_intnl_bound[1][strip])){
                  alt_intnl_bound[1][strip] = (memcmp (cp1, cp2, cell_size) && !cpt_thresh (cp2));
#ifdef DEVELOP
                  if (alt_intnl_bound[1][strip])
                    fprintf (stdout, "%i: my_x: %i, strip: %i, cpt_y: %i, cpt_z: %i: r_bound changed\n", cmtWorldRank, my_x, strip*cmtWorldSize+cmtWorldRank, cpt_y, cpt_z);
#endif /* DEVELOP */
                }			/* End if (r_int_bound unchanged) */

                else if ((!alt_strip[strip]) && (my_x >= l_int_bound) &&
                         (my_x <= r_int_bound)) {
                  alt_strip[strip] = (memcmp (cp1, cp2, cell_size) && !cpt_thresh (cp2));
#ifdef DEVELOP
                  if (alt_strip[strip])
                    fprintf (stdout, "%i: my_x: %i, strip: %i, cpt_y: %i, cpt_z: %i: alt_strip changed\n", cmtWorldRank, my_x, strip*cmtWorldSize+cmtWorldRank, cpt_y, cpt_z);
#endif /* DEVELOP */
                }			/* End if (strip unchanged) */

              }			/* End if (!manual_folds && cpt_determin) */

	  /*--- Increment CA pointers */
	      cp1++;
	      cp2++;
            }			/* End for (my_x) */
#ifdef DEBUG_RUN
  MPI_Barrier (cmtCommCommand);
  if (0 == cmtWorldRank) fprintf (stdout, "\n");
#endif /* DEBUG_RUN */

          }			/* End if (active_strip) */

          else {		/* Inactive strip */
#ifdef EVEN_DECOMP
            cp1 += CPT_F_X;
            cp2 += CPT_F_X;
#else
            cp1 += CPT_F_X[cmtWorldRank][strip];
            cp2 += CPT_F_X[cmtWorldRank][strip];
#endif /* EVEN_DECOMP */
          }			/* End Inactive strip */

          cp1 += Radius;
          cp2 += Radius;

#ifdef EVEN_DECOMP
#else
/* cpt_x adjustment for the next strip. The physical position in the
   folded, non-haloed data (i.e. cpt_x) must advance by the total 
   length of the *same* indexed strip in the *next* processes, plus
   the length of the *next* index strip in the *previous* processes. */
          for (i = 0; i < cmtWorldRank; i++) {
            if (NFOLDS == strip+1) {
              break;
            }
            cpt_x += CPT_F_X[i][strip+1];
          }			/* End for (i) */
          for (i = cmtWorldRank+1; i < CPT_NPROCS; i++) {
            cpt_x += CPT_F_X[i][strip];
          }			/* End for (i) */
#endif /* EVEN_DECOMP */
        }			/* End for (strip) */

      }				/* End for (cpt_y) */

      /*--- Skip halo */
      cp1 += (2*Radius)*CPT_X;
      cp2 += (2*Radius)*CPT_X;
    }				/* End for (cpt_z) */

    /* Stop Update Function Timer */
    stop_one_timer_sts (&statistics, FUNC, 0);

/* Update active_strip */
    for (strip = 0; strip < NFOLDS; strip++) {

      if (active_strip[strip] && (!manual_folds && cpt_determin)) {
        active_strip[strip] = (alt_intnl_bound[0][strip] || 
                               alt_intnl_bound[1][strip] ||
                               alt_strip[strip]);
#ifdef DEVELOP
        if (!active_strip[strip]) {
          fprintf (stdout, "%i: active_strip[%i] = 0\n", cmtWorldRank, strip*cmtWorldSize+cmtWorldRank);
        }
#endif /* DEVELOP */
      }		/* End if (active_strip && !manual_folds && cpt_determin) */

    }				/* End for strip */

    /*--- Update cyclic and Macrocell boundaries, copy back to read-only CA */
    if (cpt_generation % 2 == 0) {
/* Start timer BOUND */
      start_one_timer_sts (&statistics, BOUND, 0);
      cmt_boundary_copy (ca2);
      cmt_boundary_swap (ca2);
/* Stop timer BOUND */
      stop_one_timer_sts (&statistics, BOUND, 0);

/* This memcpy is essential in case cells are updated in some steps only */
      memcpy (ca1, ca2, CPT_X*CPT_Y*CPT_Z*cell_size);
    }				/* End if (cpt_generation % 2 == 0) */

    else {
/* Start timer BOUND */
      start_one_timer_sts (&statistics, BOUND, 0);
      cmt_boundary_copy (ca1);
      cmt_boundary_swap (ca1);
/* Stop timer BOUND */
      stop_one_timer_sts (&statistics, BOUND, 0);

      memcpy (ca2, ca1, CPT_X*CPT_Y*CPT_Z*cell_size);
    }				/* End if (cpt_generation % 2 != 0) */

    /*--- Execute steering function. Copy chosen irrelevant due to memcpy */
    in_steering = 1;
    start_one_timer_sts (&statistics, STEER, 0);
    /*--- Synchronise processes to avoid race conditions */
    MPI_Barrier (cmtCommCommand);
    steer_res = cpt_steering (ca1);
    stop_one_timer_sts (&statistics, STEER, 0);
    in_steering = 0;

    /*--- Increment generation, decrement num_gens */
    cpt_generation++;
    if (INFINITE != num_gens) {
      num_gens--;
    }				/* End if (INFINITE != num_gens) */

    while ((0 != (vis_ptr = first (&vis_list))) 
        && (vis_ptr->generation <= cpt_generation)) {
      tx_vis_pack (vis_ptr, 0); /* No need to force visualisation */
      vis_ptr->generation = cpt_generation + vis_ptr->data->vis_step;
      reorder (&vis_list, vis_ptr);
#ifdef DEVELOP
      if (0 == cmtWorldRank) {
        fprintf (stdout, "vis_list\t");
        traverse (vis_list, '#');
      }
#endif

    }				/* End while (vis_ptr->generation) */

    if ((0 == cmtWorldRank) && vis_sockfd) {
      int tx = GEN_NO;
/* Writen GEN_NO */
      size = sizeof (tx);
      if (size != writen (vis_sockfd, (char *) &tx, size)) {
        fprintf (stderr, "run: writen failed!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End bad writen */

/* Writen cpt_generation */
      size = sizeof (cpt_generation);
      if (size != writen (vis_sockfd, (char *) &cpt_generation, size)) {
        fprintf (stderr, "run: writen failed!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End bad writen */

    }				/* End if (0 == cmtWorldRank) */

    /*--- Write CA configuration, if on "save step" */
    if (cpt_generation % save_step == 0) {

      if (0 > cmt_write_all (out_dirname, out_basename)) {
        fprintf (stderr, "run: cmt_write_all failed!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (bad cmt_write_all) */

    }				/* End if (cpt_generation % save_step == 0) */

    /*--- Write statistics, if on period */
    if (statistics.period && (0 == cpt_generation % statistics.period)) {
      print_all_stats ();
    }				/* End if (0 == cpt_generation % period) */

    if (FINISHED == steer_res) {
      stop_one_timer_sts (&statistics, TOTAL, cpt_generation);
      rv (FINISHED);
      return (FINISHED);
    }				/* End if (FINISHED) */

  }				/* End for cpt_generation */
  

  /*--- Write final CA configuration, if not already on "save step" */
  if ((cpt_generation % save_step != 0) && (save_step != INT_MAX)) {
    if (0 > cmt_write_all (out_dirname, out_basename)) {
      fprintf (stderr, "run: cmt_write_all failed!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (bad cmt_write_all) */

  } 				/* End if (cpt_generation % save_step != 0) */

    /*--- Write final statistics, if not on period */
    if (statistics.period && (0 != cpt_generation % statistics.period)) {
      print_all_stats ();
    }				/* End if (0 != cpt_generation % period) */

  stop_one_timer_sts (&statistics, TOTAL, cpt_generation);
  rv (FINISHED);
  return (FINISHED);

} /* End run */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: tx_vis_pack
 * 
 * Called by run() to periodically visualise planes. Visualisation data
 * are mapped to colours in parallel (when possible).
 * 
 * Parameters:
 *	I   c_ptr	Cell ptr containing the data of the plane to be 
 *			visualised
 * 
 * Implementation:
 *	Checks if plane has already been visualised due to belonging to 
 *	a class. Else it writes VIS_PACK, the ID, and number of elements 
 *	of the plane and the plane substate data to the GUI. In any case, 
 *	it updates the generation pointer of the cell and accordingly 
 * 	reorders the visualisation list. The substate visualisation 
 * 	algorithm is taken from serv_view_state().
 *
 * Returns: (req_code)
 *	(VIS_PACK)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code tx_vis_pack (cell *c_ptr, char force) {
  register size_t size;
  const register size_t uc_size = sizeof (u_char);
  req_code myrequest = VIS_PACK;
  int ID, valsize;
  CptCell *cp;
  CptCell *ca = ((0 == cpt_generation%2) ? ca1 : ca2);
  unsigned char *tmp_data;
  unsigned char *tmp_data2;
#ifdef EVEN_DECOMP
  unsigned char *tmp_data3;
#endif /* EVEN_DECOMP */
  unsigned char *col_data;
  register int y, z;
  int tmp_data2_size;		/* Size of tmp_data2 */
#ifdef EVEN_DECOMP
  int tmp_data3_size;		/* Size of tmp_data3 */
#endif /* EVEN_DECOMP */
  int substate;			/* Substate to be visualised */

  register int my_x; 
  int my_y, my_z; 		/* Real data dims of volume affected */
  int y_start, y_end, z_start, z_end; /* Loop limits */
  int z_disp, y_disp, x_disp;	/* Displacements for CA_REF macro */
  int *pos;			/* Plane definition alias */
  int work;			/* Flag to denote cells idle due to
				   data decomposition */
  MPI_Comm cmtCommWork;		/* Communicator for working cells */
  int work_rank;		/* Rank in cmtCommWork */
  int work_size;		/* Size of cmtCommWork */

  double min, max; 		/* Local min and max of substate */

  u_char *ptr1;
  u_char *ptr2;

#ifdef EVEN_DECOMP
  MPI_Datatype send_vec;
  MPI_Datatype recv_vec;
  MPI_Aint displ[2];
  int blocklengths[2];
  MPI_Datatype types[2];
#else

  const int last = CPT_NPROCS-1;
  int tag;
  int i;			/* Index to position in tmp_data2[] */
  register int strip, proc;
  MPI_Datatype send_strip_vec_large, send_strip_vec_small;
  MPI_Datatype recv_strip_vec_large, recv_strip_vec_small;
  MPI_Request request[CPT_NPROCS*NFOLDS];
  MPI_Status status[CPT_NPROCS*NFOLDS];
  int count;

#endif /* EVEN_DECOMP */

/* Start VIS timer */
  start_one_timer_sts (&statistics, VIS, 0);
#ifdef DEBUG
  if (0 == cmtWorldRank) {
    fprintf (stdout, "%i: gen: %i, ID: %i", cmtWorldRank, cpt_generation, c_ptr->data->ID);
    fprintf (stdout, " last_vis: %i\n", c_ptr->data->class_ptr->last_vis);
  }				/* End if (0 == cmtWorldRank) */
#endif

/* Not visualising if (!force) and already visualised */
  if ((!force) && (c_ptr->data->class_ptr->last_vis == cpt_generation)) {

#ifdef DEVELOP
    if (0 == cmtWorldRank) {
      fprintf (stdout, "!!!!!!! %i: Plane %i already visualised\n", cpt_generation, c_ptr->data->ID);
    }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */

    work = 0;			/* Initialised so as not to free erroneously */
    goto exit;

  } 

  else {
    if (c_ptr->data->class_ptr) { /* Does not exist if plane is IGNORED */
      c_ptr->data->class_ptr->last_vis = cpt_generation;
    }				/* End if (class_ptr) */
  }				/* End if (last_vis) */

/* Make alias for c_ptr->data->pt.pos */
  pos = c_ptr->data->pt.pos;

/* Define my_x, my_y, my_z, [y|z]_start, [y|z]_end */
  if (0 == pos[0]) { 		/* x free */
#ifdef EVEN_DECOMP
    my_x = CPT_DIMX;
#else
    my_x = CPT_DIMX[cmtWorldRank];
#endif /* EVEN_DECOMP */

  } else {			/* pos[0] != 0 */
    my_x = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[0]) */

  if (0 == pos[1]) { 		/* y free */
    my_y = DIMY;
  } else {			/* pos[1] != 0 */
    my_y = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[1]) */
  y_start = 0;
  y_end = y_start+my_y;

  if (0 == pos[2]) { 		/* z free */
    my_z = DIMZ;
  } else {			/* pos[2] != 0 */
    my_z = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[2]) */
  z_start = 0;
  z_end = z_start+my_z;

#ifdef EVEN_DECOMP 
/* A subtle point: if (1 == my_x) only one PE has work to do. The one 
   whose cmtWorldRank satisfies:
   cmtWorldRank == ((pos[0]-1)/CPT_F_X)%CPT_NPROCS
   (taking into account that we start from x=1).
   The division gives us the number of strip, the modulo operation
   maps it to a fold. */

    work = 1;

    if (1 == my_x) {
      work = (cmtWorldRank == (((pos[0]-1)/CPT_F_X)%CPT_NPROCS));
    }				/* End if my_x */
#else /* EVEN_DECOMP */
/* A subtle point: if (1 == my_x) only one PE has work to do. 
   Calculate this PE and find the x_disp at the same time */

  work = 1;		/* Initialisation */
  x_disp = Radius;	/* The others done below */

  if (1 == my_x) {
    int top = 1;
    int found = 0;
    strip = 0;
    while (strip < NFOLDS) {

      proc = 0;
      while (proc < CPT_NPROCS) {
        top += CPT_F_X[proc][strip];
        if (pos[0] < top) {
          found = 1;
        }			/* End if (pos[0] < top) */
        if (found) break;
        proc++;

      }				/* End while (proc) */
      if (found) {
        x_disp += pos[0]-(top-CPT_F_X[proc][strip]);
        /* The distance from the current strip start */
        break;
      }				/* End if (found) */

      x_disp += CPT_S_X[cmtWorldRank][strip];
      strip++;	/* This *must* be the last command of the loop! */

    }				/* End while (strip) */

    if (proc != cmtWorldRank) {
      work = 0;
    }				/* End if (proc != cmtWorldRank) */

#ifdef DEBUG_TX_VIS_PACK
    for (z=0; z < CPT_NPROCS; z++) {
      MPI_Barrier (cmtCommCommand);
      if (z == cmtWorldRank) {
        fprintf (stdout, "rank: %i ", cmtWorldRank);
        for (y=0; y < 3; y++) {
          fprintf (stdout, "pos[%i]: %i\t", y, pos[y]);
        }
        fprintf (stdout, "\n");
        fprintf (stdout, "proc: %i, strip: %i, x_disp: %i ", proc, strip, x_disp);
        fprintf (stdout, "work: %i\n", work);
      } 			/* End if (z == work_rank) */
    }				/* End for */

#endif /* DEBUG_TX_VIS_PACK */

  }				/* End if my_x */

#endif /* EVEN_DECOMP */

/* Define new communicator cmtCommWork */
/* tx_vis_pack() needs this for Allreduce, unlike serv_view_state() */
    if ((0 == cmtWorldRank) || work) {
      MPI_Comm_split (cmtCommCommand, 0, 0, &cmtCommWork);
      MPI_Comm_size (cmtCommWork, &work_size);
      MPI_Comm_rank (cmtCommWork, &work_rank);
      if ((0 == cmtWorldRank) && (0 != work_rank)) {
        fprintf (stderr, "Error! 0 in World is not 0 in cmtCommWork!\n");
        MPI_Abort (cmtCommCommand, -1);
      }
    } else {
      MPI_Comm_split (cmtCommCommand, MPI_UNDEFINED, 0, &cmtCommWork);
      goto exit;
    }				/* End cmtCommWork */

/* Calculate displacements */
#ifdef EVEN_DECOMP 
  z_disp = y_disp = x_disp = Radius;
#else
/* x_disp has already been taken care of */
  z_disp = y_disp = Radius;
#endif /* EVEN_DECOMP */
  if (0 != pos[2]) {
    z_disp += pos[2]-1;
  }				/* End if (pos[2]) */

  if (0 != pos[1]) {
    y_disp += pos[1]-1;
  }				/* End if (pos[1]) */

#ifdef EVEN_DECOMP
  if (0 != pos[0]) {
/* (pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X moves us to the right strip
   (pos[0]-1)%CPT_F_X moves us to the right column in the strip. */
    x_disp += (((pos[0]-1)%CPT_F_X)+((pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X));
  }				/* End if (pos[0]) */
#else
/* x_disp already done */
#endif /* EVEN_DECOMP */

/* Get substate */
  substate = c_ptr->data->substate;

/* Allocate memory for per-PE status data. my_x is the number of elements
   required */
  tmp_data = (unsigned char *) malloc (my_x*my_y*my_z*cpt_state_size[substate]);

  if (!tmp_data) {
    fprintf (stderr, "tx_vis_pack: failed to allocate per-PE scratch space\n");
    MPI_Abort (cmtCommCommand, -1);
  }				/* End if (!tmp_data) */

/* Allocate memory for colour mapped data. The array has the same number of
   elements as tmp_data but the size of the elements is uc_size */
  col_data = (unsigned char *) malloc (my_x*my_y*my_z*uc_size);

  if (!col_data) {
    fprintf (stderr, "tx_vis_pack: failed to allocate colour map data space\n");
    MPI_Abort (cmtCommCommand, -1);
  }				/* End if (!col_data) */

/* Allocate memory for temporary storage of received state data.
   Need enough space either to hold DIMX, or just 1 element from the 
   working process.
   Important: tx_vis_pack only sends chars */
      if (0 == cmtWorldRank) {
/* Calculate tmp_data2_size */
#ifdef EVEN_DECOMP
        if (1 < work_size) {
          tmp_data2_size = 2; 		/* One for proc 0 as well */
        } else {
          tmp_data2_size = 1;
        }

#else /* EVEN_DECOMP */
        tmp_data2_size = 1;
#endif /* EVEN_DECOMP */

        if (0 == pos[0]) {
          tmp_data2_size = DIMX;
        }				/* End if (0 == pos[0]) */

        tmp_data2 = (u_char *) malloc (tmp_data2_size*my_y*my_z*uc_size);

        if (!tmp_data2) {
          fprintf (stderr, "tx_vis_pack: failed to allocate read space\n");
          MPI_Abort (cmtCommCommand, -1);
        }				/* End if (!tmp_data2) */

#ifdef EVEN_DECOMP
/* Calculate tmp_data3_size */
        if (DIMX == tmp_data2_size) { /* All cells work */
          tmp_data3_size = tmp_data2_size;
        } 				/* End if (DIMX == tmp_data2_size) */

        else {			/* One cell works */
          tmp_data3_size = my_x;
        }				/* End if tmp_data2_size */

        tmp_data3 = (u_char *) malloc (tmp_data3_size*my_y*my_z*uc_size);
    
        if (!tmp_data3) {
          fprintf (stderr, "tx_vis_pack: malloc failed, exiting\n");
          MPI_Abort (cmtCommCommand, -1);
        }				/* End if (!tmp_data3) */
#else /* EVEN_DECOMP */
#endif /* EVEN_DECOMP */

      }					/* End if (0 == cmtWorldRank) */

#ifdef DEBUG_TX_VIS_PACK
    for (z=0; z < work_size; z++) {
      MPI_Barrier (cmtCommWork);
      if (z == work_rank) {
        fprintf (stdout, "\nwork_rank: %i\n", work_rank);
        fprintf (stdout, "\nworld_rank: %i\n", cmtWorldRank);
#if 1
        for (y=0; y < 3; y++) {
          fprintf (stdout, "pos[%i]: %i\t", y, pos[y]);
        }
        fprintf (stdout, "\n");
        fprintf (stdout, "my_x: %i, my_y: %i, my_z: %i\n",my_x, my_y, my_z);
        fprintf (stdout, "y_start: %i, y_end: %i, z_start: %i, z_end: %i\n",y_start, y_end,  z_start, z_end);
        fprintf (stdout, "x_disp %i, y_disp %i, z_disp %i\n",x_disp, y_disp, z_disp);
#endif
        fprintf (stdout, "tmp_data2_size: %i\n", tmp_data2_size);
        fprintf (stdout, "work: %i\n", work);
        fprintf (stdout, "work_size: %i\n", work_size);
      } 			/* End if (z == work_rank) */
    }				/* End for */
#endif

/* --- minmax[substate][] calculation */
  if (auto_map[substate] && cpt_generation) {
/* Calculate local min, max */
    min = minmax[substate][0]; 
    max = minmax[substate][1];
    if (work) {
      for (z = z_start; z < z_end; z++) {

        for (y = y_start; y < y_end; y++) {
          cp = CA_REF (ca, z_disp+z, y_disp+y, x_disp);

/* Fill tmp_data with one line of length my_x*size */
          get_x_line (cp, substate, tmp_data, my_x);

          if (MPI_UNSIGNED_CHAR == cpt_state_mpidt[substate]) {
            u_char *ptr;
            int i;	/* Index */

            ptr = tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if u_char */

          else if (MPI_CHAR == cpt_state_mpidt[substate]) {
            char *ptr;
            int i;	/* Index */

            ptr = (char *) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if char */

          else if (MPI_SHORT == cpt_state_mpidt[substate]) {
            short *ptr;
            int i;	/* Index */

            ptr = (short *) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
              max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if short */

          else if (MPI_UNSIGNED_SHORT == cpt_state_mpidt[substate]) {
            u_short *ptr;
            int i;	/* Index */

            ptr = (u_short*) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if u_short */

          else if (MPI_INT == cpt_state_mpidt[substate]) {
            int *ptr;
            int i;	/* Index */

            ptr = (int *) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if int */

          else if (MPI_UNSIGNED == cpt_state_mpidt[substate]) {
            u_int *ptr;
            int i;	/* Index */
    
            ptr = (u_int*) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if u_int */

          else if (MPI_FLOAT == cpt_state_mpidt[substate]) {
            float *ptr;
            int i;	/* Index */
  
            ptr = (float *) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if float */

          else if (MPI_DOUBLE == cpt_state_mpidt[substate]) {
            double *ptr;
            int i;	/* Index */

            ptr = (double *) tmp_data;
            for (i = 0; i < my_x; i++) {
              if (ptr[i] < min) {
                min = ptr[i];
              }		/* End if min */
              if (ptr[i] > max) {
                max = ptr[i];
              }		/* End if max */
            }		/* End for i */
          }		/* End if double */

          else {
            fprintf (stderr, "tx_vis_pack: Incorrect cpt_state_mpidt, exiting\n");
            MPI_Abort (cmtCommCommand, -1);
          }			/* End else */

        }			/* End for(y) */

      }			/* End for(z) */

    }			/* End if (work) */

/* Get minmax[substate] to all cells */
    MPI_Allreduce (&min, &minmax[substate][0], 1, MPI_DOUBLE, MPI_MIN, cmtCommWork);
    MPI_Allreduce (&max, &minmax[substate][1], 1, MPI_DOUBLE, MPI_MAX, cmtCommWork);

#ifdef DEBUG
    if (0 == cmtWorldRank) {
      fprintf (stdout, "%i, min: %f, max: %f\n", cmtWorldRank, minmax[substate][0], minmax[substate][1]);
    }
#endif

  }			/* End if (auto_map[substate]) */

/* End minmax[substate][] calculation */

/* We have got minmax[substate][] now, colour_map and visualise */
/* Preserve initial pointer positions */
  ptr1 = tmp_data;
  ptr2 = col_data;

/* Colour map the data on each process */
  for (z = z_start; z < z_end; z++) {

    for (y = y_start; y < y_end; y++) {

      cp = CA_REF (ca, z_disp+z, y_disp+y, x_disp);

/* my_x is the number of elements involved in each PE */

      if (work) {
/* Fill tmp_data with one line of length my_x*size */
        get_x_line (cp, substate, tmp_data, my_x);
        colour_map (tmp_data, col_data, substate, my_x, minmax[substate][0], minmax[substate][1]);
        tmp_data += my_x*cpt_state_size[substate];
        col_data += my_x*uc_size;
      }			/* End if (work) */

    }				/* End for(y) */

  }				/* End for(z) */

/* Return pointers to original position */
  tmp_data = ptr1;
  col_data = ptr2;

#ifdef DEBUG
  if (0 == work_rank) {
    fprintf (stdout, "tmp_data: %i, col_data: %i\n", tmp_data, col_data);
    fprintf (stdout, "tmp_data2: %i, tmp_data3: %i\n", tmp_data2, tmp_data3);
  }

  for (z = z_start; z < z_end; z++) {

    for (y = y_start; y < y_end; y++) {
      int w, i;

      for (w=0; w < work_size; w++) {
        MPI_Barrier (cmtCommWork);
        if (w == work_rank) {
          if (0 == cmtWorldRank) {
            fprintf (stdout, "\n");
          } else {
            fprintf (stdout, "| ");
          }
          for (i = 0; i < my_x; i++) {
            fprintf (stdout, "%i ", col_data[i]);
          }
          col_data += my_x*uc_size;
          fflush (stdout);
        }
      } 

    }				/* End for(y) */

  }				/* End for(z) */

  col_data = ptr2;
#endif /* DEBUG */

#ifdef EVEN_DECOMP
  MPI_Type_vector (my_y*my_z, my_x, my_x*work_size, 
                   MPI_UNSIGNED_CHAR, &send_vec);
  MPI_Type_commit (&send_vec);

  displ[0] = 0;
  displ[1] = my_x;
  blocklengths[0] = 1;
  blocklengths[1] = 1;
  types[0] = send_vec;
  types[1] = MPI_UB;
  MPI_Type_struct (2, blocklengths, displ, types, &recv_vec);
  MPI_Type_commit (&recv_vec);

  MPI_Gather (col_data, my_x*my_y*my_z, MPI_UNSIGNED_CHAR,
              tmp_data2, 1, recv_vec, 0, cmtCommWork);

#else /* EVEN_DECOMP */

/* If (1 == my_x) do it as before (careful with the vector names)
   else do it as in cmt_write (careful with y and z) */
  if (1 != my_x) {

/* For each process role (send-receive) need 2 types of vectors, a 
   small and a large one. The large can be found in CPT_F_X[last][0],
   since it is decided to distribute extraneous cells to the processes
   from last to first and place them in strips from first to last.
   Similarly, the smallest strip will be in proc 0, designated as
   first_small_strip_ind[0]. See calc_x_sizes() for more. */

/* Check above assumption */
    if (CPT_F_X[last][0] < CPT_F_X[0][first_small_strip_ind[0]]) {
      fprintf (stderr, "%s\n", 
               "cmt_write: assumption about large and small strip failed");
      fprintf (stderr, "large: %i, small: %i\n", CPT_F_X[last][0],
               CPT_F_X[0][first_small_strip_ind[0]]);
      MPI_Abort (cmtCommCommand, -1);
    }					/* End if (large < small) */

/* Send types. */
    MPI_Type_vector (my_y*my_z,CPT_F_X[cmtWorldRank][0], CPT_DIMX[cmtWorldRank],
                     MPI_UNSIGNED_CHAR, &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_vector (my_y*my_z, 
                     CPT_F_X[cmtWorldRank][first_small_strip_ind[cmtWorldRank]],
                       CPT_DIMX[cmtWorldRank], MPI_UNSIGNED_CHAR, 
                       &send_strip_vec_small);
      MPI_Type_commit (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */

/* Receive types. */
    MPI_Type_vector (my_y*my_z, CPT_F_X[last][0], DIMX,
                     MPI_UNSIGNED_CHAR, &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_vector (my_y*my_z, CPT_F_X[0][first_small_strip_ind[0]],
                       DIMX, MPI_UNSIGNED_CHAR, &recv_strip_vec_small);
      MPI_Type_commit (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */

/* Receive data */
    if (0 == cmtWorldRank) {
/* First run strip then run processor, so as to traverse tmp_data2 linearly.
   tmp_data2 contains the data in physical order. Going down the x-axis
   one meets first strip 0 of process 1 and then strip 1 of process 0 */

      i = 0;		/* Index to position in tmp_data2[] */
      tag = 0;		/* Tag for comm and request[] index */

      for (strip = 0; strip < NFOLDS; strip++) {
        int advance;	/* Bytes to advance arrays (calc taken out) */

        for (proc = 0; proc < CPT_NPROCS; proc++) {
          if (strip < first_small_strip_ind[proc]) {
#ifdef DEBUG_TX_VIS_PACK
            fprintf (stdout, "Recv LARGE proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_TX_VIS_PACK */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_large, proc, 
                       tag, cmtCommCommand, &request[tag]);
          } else {
#ifdef DEBUG_TX_VIS_PACK
            fprintf (stdout, "Recv SMALL proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_TX_VIS_PACK */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_small, proc, 
                       tag, cmtCommCommand, &request[tag]);
          }				/* End if strip */
          i += CPT_F_X[proc][strip]*uc_size;
          tag++;

        }		/* End for (proc) */
      }			/* End for (strip) */
    }			/* End if (0 == cmtWorldRank) */

/* Send data */
    if (work) {

      i = 0;		/* Index to position in tmp_data[] */
      for (strip = 0; strip < NFOLDS; strip++) {
        tag = strip*CPT_NPROCS + cmtWorldRank;
        if (strip < first_small_strip_ind[cmtWorldRank]) {
#ifdef DEBUG_TX_VIS_PACK
          fprintf (stdout, "proc %i, Sent LARGE strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_TX_VIS_PACK */
          MPI_Send (&col_data[i], 1, send_strip_vec_large, 0, 
                    tag, cmtCommCommand);
        } else {
#ifdef DEBUG_TX_VIS_PACK
            fprintf (stdout, "proc %i, Sent SMALL strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_TX_VIS_PACK */
          MPI_Send (&col_data[i], 1, send_strip_vec_small, 0, 
                    tag, cmtCommCommand);
        }			/* End if strip */
        i += CPT_F_X[cmtWorldRank][strip]*uc_size;

      }				/* End for (strip) */
    }				/* End if (work) */

    if (0 == cmtWorldRank) {
      if (MPI_SUCCESS != MPI_Waitall (CPT_NPROCS*NFOLDS, request, status)) {
        fprintf (stderr, "tx_vis_pack: MPI_Waitall failed\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End  (MPI_Waitall) */
    }				/* End if (0 == cmtWorldRank) */

  }				/* End if (1 != my_x) */

  else { /* (1 == my_x) */

/* Send types. */
    MPI_Type_vector (my_y*my_z, 1 , my_x,
                     MPI_UNSIGNED_CHAR, &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);

/* Receive types. */
    MPI_Type_vector (my_y*my_z, 1, tmp_data2_size,
                     MPI_UNSIGNED_CHAR, &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);

/* Receive data */
    if (0 == cmtWorldRank) {

#ifdef DEBUG_TX_VIS_PACK
      fprintf (stdout, "Recv LARGE proc %i, strip %i, tag 0\n", proc, strip);
#endif /* DEBUG_TX_VIS_PACK */

      MPI_Irecv (tmp_data2, 1, recv_strip_vec_large, proc, 
                 0, cmtCommCommand, &request[0]);
    }				/* End if (0 == cmtWorldRank) */

/* Send data */
    if (work) {

#ifdef DEBUG_TX_VIS_PACK
      fprintf (stdout, "proc %i, Sent LARGE strip %i, tag 0\n", cmtWorldRank, strip);
#endif /* DEBUG_TX_VIS_PACK */
      MPI_Send (col_data, 1, send_strip_vec_large, 0, 
                0, cmtCommCommand);

    }				/* End if (work) */

    if (0 == cmtWorldRank) {
      MPI_Waitall (1, request, status);
    }				/* End if (0 == cmtWorldRank) */

  }				/* End else */

#endif /* EVEN_DECOMP */

#ifdef DEBUG
      if (0 == cmtWorldRank) {
        fprintf (stdout, "\n\n");
        ptr1 = tmp_data2;
        for (z = z_start; z < z_end; z++) {

          for (y = y_start; y < y_end; y++) {
          int i;
            for (i = 0; i<my_x*work_size; i++) {
              fprintf (stdout, "%i ", tmp_data2[i]);
            }
            tmp_data2 += my_x*work_size*uc_size;
            fprintf (stdout, "\n");
          }				/* End for(y) */

        }				/* End for(z) */
        tmp_data2 = ptr1;
      }				/* End if (0 == cmtWorldRank) */
#endif /* DEBUG */

/* Write data to socket */
  if (0 == cmtWorldRank) {
    int i; 		/* DEBUG */

/* Write VIS_PACK */
    size = sizeof (myrequest);
    if (size != writen (vis_sockfd, (char *) &myrequest, size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

/* Write ID */
    ID = c_ptr->data->ID;
    size = sizeof (ID);
    if (size != writen (vis_sockfd, (char *) &ID, size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

/* Write valsize */
    valsize = get_val_size (pos);
    size = sizeof (valsize);
    if (size != writen (vis_sockfd, (char *) &valsize, size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

/* Write minmax */
    size = 2*sizeof (double);
    if (size != writen (vis_sockfd, (char *) minmax[substate], size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */
#ifdef EVEN_DECOMP
/* Writen tmp_data3 */
    get_write_ptr (tmp_data2, tmp_data3, tmp_data3_size, my_x, my_y, my_z,
                   uc_size, work, work_size); 

    size = tmp_data3_size*my_y*my_z*uc_size;
    if (size != writen (vis_sockfd, (char *) tmp_data3, size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */
#else /* EVEN_DECOMP */
/* Writen tmp_data2 */
    size = tmp_data2_size*my_y*my_z*uc_size;
    if (size != writen (vis_sockfd, (char *) tmp_data2, size)) {
      fprintf (stderr, "tx_vis_pack: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

#endif /* EVEN_DECOMP */
  }				/* End if (0 == cmtWorldRank) */

  free (tmp_data);
  free (col_data);
#ifdef EVEN_DECOMP
  MPI_Type_free (&send_vec);
  MPI_Type_free (&recv_vec);
#else /* EVEN_DECOMP */
  MPI_Type_free (&send_strip_vec_large);
  if ((NFOLDS != first_small_strip_ind[cmtWorldRank]) && (1 != my_x)) {
    MPI_Type_free (&send_strip_vec_small);
  }			/* End if (first_small_strip_ind[cmtWorldRank]) */
  MPI_Type_free (&recv_strip_vec_large);
  if ((NFOLDS != first_small_strip_ind[0]) && (1 != my_x)) {
    MPI_Type_free (&recv_strip_vec_small);
  }			/* End if (first_small_strip_ind[0]) */
#endif /* EVEN_DECOMP */
  if (0 == cmtWorldRank) {
    free (tmp_data2);
#ifdef EVEN_DECOMP
    free (tmp_data3);
#endif /* EVEN_DECOMP */
  }				/* End if (0 == cmtWorldRank) */

exit:
  if (work) {
    MPI_Comm_free (&cmtCommWork);
  }				/* End if (work) */

/* Stop VIS timer */
  stop_one_timer_sts (&statistics, VIS, 0);
  return (VIS_PACK);
  
} /* End tx_vis_pack */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: colour_map
 * 
 * Maps orig_data to mapped_data so as to graphically represent chars, 
 * ints, floats or doubles using u_chars (colours).
 * 
 * Parameters:
 *	I orig_data    u_char ptr with original data
 * 	O mapped_data  u_char ptr with mapped data
 *	I stateid      int stateid of state represented by orig_data
 *	I no_data      int number of actual elements in [orig|mapped]_data
 *	I gmin         double minimum value of elements in CA
 *	I gmax         double maximum value of elements in CA
 * 
 * Implementation:
 *	The algorithm for mapping the data: map gmin to 1, gmax to 255,
 *	uniformly map the intermediate data to the intermediate colour
 *	values. 
 *	Maps data to 1 if (gmax == gmin).
 *	Four cases for the possible types of the actual data.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void colour_map (const u_char *orig_data, u_char *mapped_data,
 			int stateid, int no_data, double gmin, double gmax) {
  register int i;
  double width = (((DBL_MAX == gmin) || (0 == gmax-gmin)) ?
		  0.0 : 254.0/(gmax-gmin));

  if (MPI_UNSIGNED_CHAR == cpt_state_mpidt[stateid]) {
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) orig_data[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) orig_data[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) orig_data[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if u_char */

  else if (MPI_CHAR == cpt_state_mpidt[stateid]) {
    char *c_ptr = (char *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) c_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) c_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) c_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if char */

  else if (MPI_UNSIGNED_SHORT == cpt_state_mpidt[stateid]) {
    u_short *us_ptr = (u_short *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) us_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) us_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) us_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if u_short */

  else if (MPI_SHORT == cpt_state_mpidt[stateid]) {
    short *s_ptr = (short *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) s_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) s_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) s_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if short */

  else if (MPI_UNSIGNED == cpt_state_mpidt[stateid]) {
    u_int *ui_ptr = (u_int *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) ui_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) ui_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) ui_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if u_int */

  else if (MPI_INT == cpt_state_mpidt[stateid]) {
    int *i_ptr = (int *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) i_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) i_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) i_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if int */

  else if (MPI_FLOAT == cpt_state_mpidt[stateid]) {
    float *f_ptr = (float *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) f_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) f_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) f_ptr[i]-gmin)*width+1.0);
      }				/* End if */
    } 				/* End for */
  }				/* End if float */

  else if (MPI_DOUBLE == cpt_state_mpidt[stateid]) {
    double *d_ptr = (double *) orig_data;
    for (i = 0; i < no_data; i++) {
      if (gmin > (double) d_ptr[i]) {
        mapped_data[i] = (u_char) 1;
      } else if (gmax < (double) d_ptr[i]) {
        mapped_data[i] = (u_char) 255;
      } else {
        mapped_data[i] = (u_char) (((double) d_ptr[i]-gmin)*width+1.0);
      }				/* End if */
      mapped_data[i] = (u_char) (((double) d_ptr[i]-gmin)*width+1.0);
    } 				/* End for */
  }				/* End if double */

  else {
    fprintf (stderr, "colour_map: unknown cpt_state_mpidt[stateid] %i, exiting\n", cpt_state_mpidt[stateid]);
    MPI_Abort (cmtCommCommand, -1);
  }

} /* End colour_map */ 

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_save_request
 * 
 * Process 0 reads in stateid and fname and saves the substate data 
 * to fname.
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 * 	Proc 0 readns stateid, strlen (fname), fname then broadcasts
 *	the stateid and all processes call cmt_write()
 *
 * Returns: (req_code)
 *	(SAVE_REQUEST)   on success.
 *	(ERROR_CODE) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_save_request (void) {
    FILE *fout;
    int length;
    int stateid; 
    char *fname;
    int size;

    if (0 == cmtWorldRank) {
/* Read stateid */
      size = sizeof (int);
      if (size != readn (prot_sockfd, (char *) &stateid, size)) {
        fprintf (stderr, "serv_save_request: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
#ifdef DEBUG
      fprintf (stdout, "serv_save_request: Got stateid: %i\n", stateid);
#endif

/* Read filename length as read by strlen () on the other side */
      if (size != readn (prot_sockfd, (char *) &length, size)) {
        fprintf (stderr, "serv_save_request: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if readn */
#ifdef DEBUG
      fprintf (stdout, "serv_save_request: Got length: %i\n", length);
#endif

/* malloc fname to fit terminating '\0' */
      if (0== (fname = (char *) malloc ((length+1)*sizeof (char)))) {
        fprintf (stderr, "serv_save_request: malloc error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if malloc */

/* Read filename */
      if (length != readn (prot_sockfd, fname, length)) {
        fprintf (stderr, "serv_save_request: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if readn */
/*  Terminate fname */
      fname [length] = (char) 0;

#ifdef DEVELOP
      fprintf (stdout, "serv_save_request fname: %s\n", fname);
#endif /* DEVELOP */

    } 				/* End if (0 == cmtWorldRank) */

/* All the cells execute MPI_Bcast to get stateid */
      MPI_Bcast ((int *) &stateid, 1, MPI_INT, 0, cmtCommCommand);

    if (0 == cmt_write (fname, stateid)) {
      return (SAVE_REQUEST);
    } else {
      return (ERROR_CODE);
    }

} /* End serv_save_request */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_set_fold
 * 
 * Manually sets the set of active folds.
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads the range from the socket, checks it for compatibility with
 *	Engine configuration, updates manual_folds and active_strip[].
 *
 * Returns: (req_code)
 *	(SET_FOLD)	on success.
 *	(IGNORED)	if folds unacceptable
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_set_fold (void) {
    register const size_t size = 2*sizeof (int);
    int folds[2];
    int fold;

    if (0 == cmtWorldRank) {
/* Read folds */
      if (size != readn (prot_sockfd, (char *) folds, size)) {
        fprintf (stderr, "serv_set_fold: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
#ifdef DEVELOP
      fprintf (stdout, "serv_set_fold: Got folds: %i-%i\n", folds[0], folds[1]);
#endif
    }				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get folds */
    MPI_Bcast ((int *) folds, 2, MPI_INT, 0, cmtCommCommand);

/* Check folds values */
    if ((folds[0] > folds[1]) || (folds[0] < 0) || (folds[1] > NFOLDS-1)) {
#ifdef DEVELOP
      fprintf (stderr, "serv_set_fold: bad folds definition, ignored\n");
#endif
      return (IGNORED);
    }				/* End if bad folds */

/* Set manual_folds */
    if ((0 == folds[0]) && (NFOLDS-1 == folds[1])) {
      manual_folds = 0;
#ifdef DEVELOP
      fprintf (stdout, "serv_set_fold: manual folds unset\n");
#endif
    }				/* End if all folds */

    else {
      manual_folds = 1;
#ifdef DEVELOP
      fprintf (stdout, "serv_set_fold: manual folds set\n");
#endif
    }				/* End if all folds */

/* Set active_strip */
    for (fold = 0; fold < NFOLDS; fold++) {

      if ((fold < folds[0]) || (fold > folds[1])) {
        active_strip[fold] = 0;
      }				/* End if (fold outside active range) */

      else {
        active_strip[fold] = 1;
      }				/* End if (fold in active range) */

#ifdef DEVELOP
     fprintf (stdout, "%i: active_strip[%i]: %i\n", cmtWorldRank, fold, active_strip[fold]);
#endif
    }				/* End for */

    return (SET_FOLD);

} /* End serv_set_fold */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_set_load
 * 
 * Does exactly what cmt_read() does except that it reads the stateid and
 * filename from the socket.
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads from the socket the stateid and name of the binary file 
 *	comprising the values of the sub-state for the entire CA. Calls
 *	cmt_read() for the rest.
 *
 * Returns: (req_code)
 *	(SET_LOAD)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_set_load () {
    FILE *fin;
    int length;
    int stateid; 
    char *fname;
    int size;

    if (0 == cmtWorldRank) {
/* Read stateid */
      size = sizeof (stateid);
      if (size != readn (prot_sockfd, (char *) &stateid, size)) {
        fprintf (stderr, "serv_set_load: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
#ifdef DEBUG
      fprintf (stdout, "serv_set_load: Got stateid: %i\n", stateid);
#endif

/* Read filename length as read by strlen () on the other side */
      if (size != readn (prot_sockfd, (char *) &length, size)) {
        fprintf (stderr, "serv_set_load: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if readn */

/* malloc fname to fit terminating '\0' */
      if (0== (fname = (char *) malloc ((length+1)*sizeof (char)))) {
        fprintf (stderr, "serv_set_load: malloc error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if malloc */

/* Read filename */
      if (length != readn (prot_sockfd, fname, length)) {
        fprintf (stderr, "serv_set_load: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }                         /* End if readn */
/*  Terminate fname */
      fname [length] = (char) 0;
    } 				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get stateid */
    MPI_Bcast ((int *) &stateid, 1, MPI_INT, 0, cmtCommCommand);

    if (0 == cmt_read (fname, stateid)) {
      return (SET_LOAD);
    } else {
      return (ERROR_CODE);
    }

} /* End serv_set_load */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_view_state
 * 
 * Writes to the GUI the substate data for a plane. 
 * Adapted from cmt_write()
 * 
 * Parameters:
 * 
 * Implementation:
 *	Reads the position of the point, checks and corrects it; reads 
 *	substate. Then writes the current generation, calculates the 
 *	dimensions of the Automaton that are concerned and writes the 
 *	substate data in x-dimension doses. 
 *	Assumes that the entities are of full size.
 *
 * Returns: (req_code)
 * 	(VIEW_STATE) 	on success
 *	(-ve)		on failure
 *
 *----------------------------------------------------------------------*/

static req_code serv_view_state () {
  CptCell *cp;
  CptCell *ca = ((0 == cpt_generation%2) ? ca1 : ca2);
  unsigned char *tmp_data;
  unsigned char *tmp_data2;
  register int y, z;
  register size_t size;
  int substate;
  int tmp_data2_size;		/* Size of tmp_data2 */

  int my_x, my_y, my_z; 	/* Real data dims of CA affected */
  int y_start, y_end, z_start, z_end; /* Loop limits */
  int z_disp, y_disp, x_disp;	/* Displacements for CA_REF macro */
  int pos[3];			/* Plane definition */
  int work;			/* Flag to denote cells idle due to
				   data decomposition */
#ifdef EVEN_DECOMP
  unsigned char *tmp_data3;
  unsigned char *write_ptr;
  int tmp_data3_size;		/* Size of tmp_data3 */
  MPI_Comm cmtCommWork;		/* Communicator for working cells */
  int work_rank;		/* Rank in cmtCommWork */
  int work_size;		/* Size of cmtCommWork */
#else /* EVEN_DECOMP */
  int i;
#endif /* EVEN_DECOMP */

  u_char *ptr1;

#ifdef EVEN_DECOMP
  MPI_Datatype send_vec;
  MPI_Datatype recv_vec;
  MPI_Aint displ[2];
  int blocklengths[2];
  MPI_Datatype types[2];

#else /* EVEN_DECOMP */

  const int last = CPT_NPROCS-1;
  int tag;
  register int strip, proc;
  MPI_Datatype send_strip_vec_large, send_strip_vec_small;
  MPI_Datatype recv_strip_vec_large, recv_strip_vec_small;
  MPI_Request request[CPT_NPROCS*NFOLDS];
  MPI_Status status[CPT_NPROCS*NFOLDS];
  int count;

#endif /* EVEN_DECOMP */

  if (0 == cmtWorldRank) {
/* Read pos */
    size = sizeof (pos);
    if (size != readn (prot_sockfd, (char *) pos, size)) {
      fprintf (stderr, "serv_view_state: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */

/* Safeguards for pos */
    check_pos (pos);

/* Read substate */
    size = sizeof (substate);
    if (size != readn (prot_sockfd, (char *) &substate, size)) {
      fprintf (stderr, "serv_view_state: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */

/* Check substate */
    if ((0 > substate) || (NumOfStates-1 < substate)) {
      fprintf (stderr, "serv_view_state: bad substate\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (substate) */

  }				/* End if (0 == cmtWorldRank) */

/* Execute MPI_Bcast to communicate pos and substate */
  MPI_Bcast (pos, 3, MPI_INT, 0, cmtCommCommand);
  MPI_Bcast (&substate, 1, MPI_INT, 0, cmtCommCommand);
 
/* Define my_x, my_y, my_z, [y|z]_start, [y|z]_end */
  if (0 == pos[0]) { 		/* x free */
#ifdef EVEN_DECOMP
    my_x = CPT_DIMX;
#else
    my_x = CPT_DIMX[cmtWorldRank];
#endif /* EVEN_DECOMP */
  } else {			/* pos[0] != 0 */
    my_x = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[0]) */

  if (0 == pos[1]) { 		/* y free */
    my_y = DIMY;
  } else {			/* pos[1] != 0 */
    my_y = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[1]) */
  y_start = 0;
  y_end = y_start+my_y;		/* Both cases */

  if (0 == pos[2]) { 		/* z free */
    my_z = DIMZ;
  } else {			/* pos[2] != 0 */
    my_z = 1;			/* Assumes full extent of line/plane/cube */
  }				/* End if (pos[2]) */
  z_start = 0;
  z_end = z_start+my_z;	/* Both cases */

#ifdef EVEN_DECOMP 
/* A subtle point: if (1 == my_x) only one PE has work to do. The one 
   whose cmtWorldRank satisfies:
   cmtWorldRank == ((pos[0]-1)/CPT_F_X)%CPT_NPROCS
   (taking into account that we start from x=1).
   The division gives us the number of strip, the modulo operation
   maps it to a fold. */

    work = 1;

    if (1 == my_x) {
      work = (cmtWorldRank == (((pos[0]-1)/CPT_F_X)%CPT_NPROCS));
#ifdef DEVELOP
      if (work) fprintf (stdout, "%i: work!\n", cmtWorldRank);
#endif /* DEVELOP */
    }				/* End if my_x */
#else /* EVEN_DECOMP */
/* A subtle point: if (1 == my_x) only one PE has work to do. 
   Calculate this PE and find the x_disp at the same time */

  work = 1;		/* Initialisation */
  x_disp = Radius;	/* The others done below */

  if (1 == my_x) {
    int top = 1;
    int found = 0;
    strip = 0;
    while (strip < NFOLDS) {

      proc = 0;
      while (proc < CPT_NPROCS) {
        top += CPT_F_X[proc][strip];
        if (pos[0] < top) {
          found = 1;
        }			/* End if (pos[0] < top) */
        if (found) break;
        proc++;

      }				/* End while (proc) */
      if (found) {
        x_disp += pos[0]-(top-CPT_F_X[proc][strip]); 
        /* The distance from the current strip start */
        break;
      }				/* End if (found) */
      x_disp += CPT_S_X[cmtWorldRank][strip];
      strip++;	/* This *must* be the last command of the loop! */

    }				/* End while (strip) */

    if (proc != cmtWorldRank) {
      work = 0;
    }				/* End if (proc != cmtWorldRank) */

#ifdef DEBUG_SERV_VIEW_STATE
    for (z=0; z < CPT_NPROCS; z++) {
      MPI_Barrier (cmtCommCommand);
      if (z == cmtWorldRank) {
        fprintf (stdout, "rank: %i ", cmtWorldRank);
        for (y=0; y < 3; y++) {
          fprintf (stdout, "pos[%i]: %i\t", y, pos[y]);
        }
        fprintf (stdout, "\n");
        fprintf (stdout, "proc: %i, strip: %i, x_disp: %i ", proc, strip, x_disp);
        fprintf (stdout, "work: %i\n", work);
      } 			/* End if (z == work_rank) */
    }				/* End for */

#endif /* DEBUG_SERV_VIEW_STATE */

  }				/* End if my_x */

#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP
/* Define new communicator cmtCommWork */
  if ((0 == cmtWorldRank) || work) {
    MPI_Comm_split (cmtCommCommand, 0, 0, &cmtCommWork);
    MPI_Comm_size (cmtCommWork, &work_size);
    MPI_Comm_rank (cmtCommWork, &work_rank);
  } else {
    MPI_Comm_split (cmtCommCommand, MPI_UNDEFINED, 0, &cmtCommWork);
    goto exit;
  }				/* End cmtCommWork */

#else /* EVEN_DECOMP */

  if ((0 != cmtWorldRank) && !work) {
    goto exit;
  }				/* End if ((0 != cmtWorldRank) && !work) */
#endif /* EVEN_DECOMP */


/* Allocate memory for per-PE status data. my_x is the number of elements
   required for each line */
  tmp_data = (unsigned char *) malloc (my_x*my_y*my_z*cpt_state_size[substate]);

  if (!tmp_data) {
    fprintf (stderr, "serv_view_state: failed to allocate scratch space\n");
    MPI_Abort (cmtCommCommand, -1);
  }				/* End if (!tmp_data) */

/* Allocate memory for temporary storage of received state data.
   Need enough space either to hold DIMX, or just 1 element from the
   working process */
  if (0 == cmtWorldRank) {
/* Calculate tmp_data2_size */
#ifdef EVEN_DECOMP
    if (1 < work_size) {
      tmp_data2_size = 2; 	/* One for proc 0 as well */
    } else {
      tmp_data2_size = 1;
    }

#else /* EVEN_DECOMP */
    tmp_data2_size = 1;
#endif /* EVEN_DECOMP */

    if (0 == pos[0]) {
      tmp_data2_size = DIMX;
    }				/* End if (0 == pos[0]) */

    tmp_data2 = (u_char *) malloc (tmp_data2_size*my_y*my_z*cpt_state_size[substate]);

    if (!tmp_data2) {
      fprintf (stderr, "serv_view_state: failed to allocate read space\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data2) */

#ifdef EVEN_DECOMP
/* Calculate tmp_data3_size */
    if (DIMX == tmp_data2_size) { /* All cells work */
      tmp_data3_size = tmp_data2_size;
    } 				/* End if (DIMX == tmp_data2_size) */

    else {			/* One cell works */
      tmp_data3_size = my_x;
    }				/* End if tmp_data2_size */

    tmp_data3 = (u_char *) malloc (tmp_data3_size*my_y*my_z*cpt_state_size[substate]);
    
    if (!tmp_data3) {
      fprintf (stderr, "serv_view_state: malloc failed, exiting\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data3) */
#else /* EVEN_DECOMP */
#endif /* EVEN_DECOMP */

  }				/* End if (0 == cmtWorldRank) */

/* Calculate displacements */
#ifdef EVEN_DECOMP 
  z_disp = y_disp = x_disp = Radius;
#else
/* x_disp has already been taken care of */
  z_disp = y_disp = Radius;
#endif /* EVEN_DECOMP */
  if (0 != pos[2]) {
    z_disp += pos[2]-1;
  }				/* End if (pos[2]) */

  if (0 != pos[1]) {
    y_disp += pos[1]-1;
  }				/* End if (pos[1]) */

#ifdef EVEN_DECOMP
  if (0 != pos[0]) {
/* (pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X moves us to the right strip
   (pos[0]-1)%CPT_F_X moves us to the right column in the strip. */
    x_disp += (((pos[0]-1)%CPT_F_X)+((pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X));
  }				/* End if (pos[0]) */
#else
/* x_disp already done */
#endif /* EVEN_DECOMP */

  ptr1 = tmp_data;

  for (z = z_start; z < z_end; z++) {

    for (y = y_start; y < y_end; y++) {

      cp = CA_REF (ca, z_disp+z, y_disp+y, x_disp);

/* my_x is the number of elements involved in each PE */

      if (work) {
/* Fill tmp_data with one line of length my_x*size */
        get_x_line (cp, substate, tmp_data, my_x);
        tmp_data += my_x*cpt_state_size[substate];
      }			/* End if (work) */

    }				/* End for(y) */

  }				/* End for(z) */

/* Return pointer to original position */
  tmp_data = ptr1;

#ifdef EVEN_DECOMP
  MPI_Type_vector (my_y*my_z, my_x, my_x*work_size, 
                   cpt_state_mpidt[substate], &send_vec);
  MPI_Type_commit (&send_vec);

  displ[0] = 0;
  displ[1] = my_x;
  blocklengths[0] = 1;
  blocklengths[1] = 1;
  types[0] = send_vec;
  types[1] = MPI_UB;
  MPI_Type_struct (2, blocklengths, displ, types, &recv_vec);
  MPI_Type_commit (&recv_vec);

  MPI_Gather (tmp_data, my_x*my_y*my_z, cpt_state_mpidt[substate],
              tmp_data2, 1, recv_vec, 0, cmtCommWork);

#else /* EVEN_DECOMP */

/* If (1 == my_x) do it as before (careful with the vector names)
   else do it as in cmt_write (careful with y and z) */
  if (1 != my_x) {

/* For each process role (send-receive) need 2 types of vectors, a 
   small and a large one. The large can be found in CPT_F_X[last][0],
   since it is decided to distribute extraneous cells to the processes
   from last to first and place them in strips from first to last.
   Similarly, the smallest strip will be in proc 0, designated as
   first_small_strip_ind[0]. See calc_x_sizes() for more. */

/* Check above assumption */
    if (CPT_F_X[last][0] < CPT_F_X[0][first_small_strip_ind[0]]) {
      fprintf (stderr, "%s\n", 
              "serv_view_state: assumption about large and small strip failed");
      fprintf (stderr, "large: %i, small: %i\n", CPT_F_X[last][0],
               CPT_F_X[0][first_small_strip_ind[0]]);
      MPI_Abort (cmtCommCommand, -1);
    }					/* End if (large < small) */

/* Send types. */
    MPI_Type_vector (my_y*my_z,CPT_F_X[cmtWorldRank][0], CPT_DIMX[cmtWorldRank],
                     cpt_state_mpidt[substate], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_vector (my_y*my_z, 
                     CPT_F_X[cmtWorldRank][first_small_strip_ind[cmtWorldRank]],
                       CPT_DIMX[cmtWorldRank], cpt_state_mpidt[substate], 
                       &send_strip_vec_small);
      MPI_Type_commit (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */

/* Receive types. */
    MPI_Type_vector (my_y*my_z, CPT_F_X[last][0], DIMX,
                     cpt_state_mpidt[substate], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_vector (my_y*my_z, CPT_F_X[0][first_small_strip_ind[0]],
                       DIMX, cpt_state_mpidt[substate], &recv_strip_vec_small);
      MPI_Type_commit (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */

/* Receive data */
    if (0 == cmtWorldRank) {
/* First run strip then run processor, so as to traverse tmp_data2 linearly.
   tmp_data2 contains the data in physical order. Going down the x-axis
   one meets first strip 0 of process 1 and then strip 1 of process 0 */

      i = 0;		/* Index to position in tmp_data2[] */
      tag = 0;		/* Tag for comm and request[] index */

      for (strip = 0; strip < NFOLDS; strip++) {
        int advance;	/* Bytes to advance arrays (calc taken out) */

        for (proc = 0; proc < CPT_NPROCS; proc++) {
          if (strip < first_small_strip_ind[proc]) {
#ifdef DEBUG_SERV_VIEW_STATE
            fprintf (stdout, "Recv LARGE proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_SERV_VIEW_STATE */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_large, proc, 
                       tag, cmtCommCommand, &request[tag]);
          } else {
#ifdef DEBUG_SERV_VIEW_STATE
            fprintf (stdout, "Recv SMALL proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_SERV_VIEW_STATE */
            MPI_Irecv (&tmp_data2[i], 1, recv_strip_vec_small, proc, 
                       tag, cmtCommCommand, &request[tag]);
          }				/* End if strip */
          i += CPT_F_X[proc][strip]*cpt_state_size[substate];
          tag++;

        }				/* End for (proc) */
      }					/* End for (strip) */
    }				/* End if (0 == cmtWorldRank) */

/* Send data */
    if (work) {

      i = 0;		/* Index to position in tmp_data[] */
      for (strip = 0; strip < NFOLDS; strip++) {
        tag = strip*CPT_NPROCS + cmtWorldRank;
        if (strip < first_small_strip_ind[cmtWorldRank]) {
#ifdef DEBUG_SERV_VIEW_STATE
          fprintf (stdout, "proc %i, Sent LARGE strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_SERV_VIEW_STATE */
          MPI_Send (&tmp_data[i], 1, send_strip_vec_large, 0, 
                    tag, cmtCommCommand);
        } else {
#ifdef DEBUG_SERV_VIEW_STATE
            fprintf (stdout, "proc %i, Sent SMALL strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_SERV_VIEW_STATE */
          MPI_Send (&tmp_data[i], 1, send_strip_vec_small, 0, 
                    tag, cmtCommCommand);
        }			/* End if strip */
        i += CPT_F_X[cmtWorldRank][strip]*cpt_state_size[substate];

      }				/* End for (strip) */
    }				/* End if (work) */

    if (0 == cmtWorldRank) {
      MPI_Waitall (CPT_NPROCS*NFOLDS, request, status);
    }				/* End if (0 == cmtWorldRank) */

  }				/* End if (1 != my_x) */

  else { /* (1 == my_x) */

/* Send types. */
    MPI_Type_vector (my_y*my_z, 1 , my_x,
                     cpt_state_mpidt[substate], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);

/* Receive types. */
    MPI_Type_vector (my_y*my_z, 1, tmp_data2_size,
                     cpt_state_mpidt[substate], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);

/* Receive data */
    if (0 == cmtWorldRank) {

#ifdef DEBUG_SERV_VIEW_STATE
      fprintf (stdout, "Recv LARGE proc %i, strip %i, tag 0\n", proc, strip);
#endif /* DEBUG_SERV_VIEW_STATE */

      MPI_Irecv (tmp_data2, 1, recv_strip_vec_large, proc, 
                 0, cmtCommCommand, &request[0]);
    }				/* End if (0 == cmtWorldRank) */

/* Send data */
    if (work) {

#ifdef DEBUG_SERV_VIEW_STATE
      fprintf (stdout, "proc %i, Sent LARGE strip %i, tag 0\n", cmtWorldRank, strip);
#endif /* DEBUG_SERV_VIEW_STATE */
      MPI_Send (tmp_data, 1, send_strip_vec_large, 0, 
                0, cmtCommCommand);

    }				/* End if (work) */

    if (0 == cmtWorldRank) {
      MPI_Waitall (1, request, status);
    }				/* End if (0 == cmtWorldRank) */

  }				/* End else */

#endif /* EVEN_DECOMP */

/* Write data to socket */
  if (0 == cmtWorldRank) {
    int i; 			/* DEBUG */

/* Writen cpt_generation */
    size = sizeof (cpt_generation);
    if (size != writen (prot_sockfd, (char *) &cpt_generation, size)) {
      fprintf (stderr, "serv_view_state: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

#ifdef EVEN_DECOMP
    get_write_ptr (tmp_data2, tmp_data3, tmp_data3_size, my_x, my_y, my_z,
                   cpt_state_size[substate], work, work_size); 
/* Writen tmp_data3 */
    size = tmp_data3_size*my_y*my_z*cpt_state_size[substate];
    if (size != writen (prot_sockfd, (char *) tmp_data3, size)) {
      fprintf (stderr, "serv_view_state: writen error!\n");
      MPI_Abort (cmtCommWork, -1);
    }				/* End if writen */

#else /* EVEN_DECOMP */

/* Writen tmp_data2 */
    size = tmp_data2_size*my_y*my_z*cpt_state_size[substate];
    if (size != writen (prot_sockfd, (char *) tmp_data2, size)) {
      fprintf (stderr, "serv_view_state: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */

#endif /* EVEN_DECOMP */
  }				/* End if (0 == cmtWorldRank) */

  free (tmp_data);
#ifdef EVEN_DECOMP
  MPI_Type_free (&send_vec);
  MPI_Type_free (&recv_vec);
#else
    MPI_Type_free (&send_strip_vec_large);
    if ((NFOLDS != first_small_strip_ind[cmtWorldRank]) && (1 != my_x)) {
      MPI_Type_free (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */
    MPI_Type_free (&recv_strip_vec_large);
    if ((NFOLDS != first_small_strip_ind[0]) && (1 != my_x)) {
      MPI_Type_free (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */
#endif /* EVEN_DECOMP */
  if (0 == cmtWorldRank) {
    free (tmp_data2);
#ifdef EVEN_DECOMP
    free (tmp_data3);
#else
#endif /* EVEN_DECOMP */
  }				/* End if (0 == cmtWorldRank) */

exit:
#ifdef EVEN_DECOMP
  if (work) {
    MPI_Comm_free (&cmtCommWork);
  }
#endif
  return (VIEW_STATE);

} /* End serv_view_state */
 
/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_set_state
 * 
 * Reads the substate data from the socket and sets them in the CA Engine.
 * Adapted from cmt_read().
 * 
 * Parameters:
 * 
 * Implementation:
 *	Reads and checks the point defining the plane and the stateid.
 *	Calculates the dimensions of the Automaton that are concerned 
 *	and reads and sets the substate data in x-dimension doses.
 *
 * Returns: (req_code)
 *	(SET_STATE)	on success
 *	(-ve)		else
 *
 *----------------------------------------------------------------------*/

static req_code serv_set_state (void) {
    CptCell *cp1, *cp2;
    unsigned char *tmp_data;
    unsigned char *tmp_data2;
    register int y, z;
    register size_t size;
    int tmp_data2_size;		/* Size of tmp_data2 */

    int my_x, my_y, my_z; 	/* Real data dims of volume affected */
    int y_start, y_end, z_start, z_end; /* Loop limits */
    int z_disp, y_disp, x_disp;	/* Displacements for CA_REF macro */
    int pos[3];			/* Plane definition */
    int stateid; 
    int work; 			/* Flag for PE (not) idle due to layout
				   of entity in pos[] */
#ifdef EVEN_DECOMP
    unsigned char *tmp_data3;
    unsigned char *scatter_ptr;
#else 
    const int last = CPT_NPROCS-1;
    unsigned char *ptr1;	/* Used for ptr arithmetic */
    int advance;
    int i, tag;
    register int strip, proc;
    MPI_Datatype send_strip_vec_large, send_strip_vec_small;
    MPI_Datatype recv_strip_vec_large, recv_strip_vec_small;
    MPI_Request request[CPT_NPROCS*NFOLDS];
    MPI_Status status[CPT_NPROCS*NFOLDS];
    int count;

#endif /* EVEN_DECOMP */

    if (0 == cmtWorldRank) {
/* Read pos */
      size = sizeof (pos);
      if (size != readn (prot_sockfd, (char *) pos, size)) {
        fprintf (stderr, "serv_set_state: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      } 			/* End if readn */

/* Safeguards for pos */
      check_pos (pos);

/* Read stateid */
      size = sizeof (stateid);
      if (size != readn (prot_sockfd, (char *) &stateid, size)) {
        fprintf (stderr, "serv_set_state: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */

/* Check stateid */
      if ((0 > stateid) || (NumOfStates-1 < stateid)) {
        fprintf (stderr, "serv_set_state: bad stateid\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (stateid) */

    }				/* End if (0 == cmtWorldRank) */

/* Execute MPI_Bcast to communicate pos, stateid */
    MPI_Bcast (pos, 3, MPI_INT, 0, cmtCommCommand);
    MPI_Bcast ((int *) &stateid, 1, MPI_INT, 0, cmtCommCommand);
 
/* Define my_x, my_y, my_z, [y|z]_start, [y|z]_end */
    if (0 == pos[0]) { 		/* x free */
#ifdef EVEN_DECOMP
      my_x = CPT_DIMX;
#else
      my_x = CPT_DIMX[cmtWorldRank];
#endif /* EVEN_DECOMP */
    } else {			/* pos[0] != 0 */
      my_x = 1;			/* Assumes full extent of line/plane/cube */
    }				/* End if (pos[0]) */

    if (0 == pos[1]) { 		/* y free */
      my_y = DIMY;
    } else {			/* pos[1] != 0 */
      my_y = 1;			/* Assumes full extent of line/plane/cube */
    }				/* End if (pos[1]) */
    y_start = 0;
    y_end = y_start+my_y;	/* Both cases */

    if (0 == pos[2]) { 		/* z free */
      my_z = DIMZ;
    } else {			/* pos[2] != 0 */
      my_z = 1;			/* Assumes full extent of line/plane/cube */
    }				/* End if (pos[2]) */
    z_start = 0;
    z_end = z_start+my_z;	/* Both cases */

#ifdef EVEN_DECOMP
/* A subtle point: if (1 == my_x) only one PE has work to do. The one 
   whose cmtWorldRank satisfies:
   cmtWorldRank == (((pos[0]-1)/CPT_F_X)%CPT_NPROCS)
   (taking into account that we start from x=1).
   The division gives us the number of strip, the modulo operation
   maps it to a fold. */

    work = 1;

    if (1 == my_x) {
      work = (cmtWorldRank == (((pos[0]-1)/CPT_F_X)%CPT_NPROCS));
#ifdef DEVELOP
      if (work) fprintf (stdout, "%i: work!\n", cmtWorldRank);
#endif /* DEVELOP */

    }				/* End if my_x */

#else /* EVEN_DECOMP */
/* A subtle point: if (1 == my_x) only one PE has work to do. 
   Calculate this PE and find the x_disp at the same time */

  work = 1;		/* Initialisation */
  x_disp = Radius;	/* The others done below */

  if (1 == my_x) {
    int top = 1;
    int found = 0;
    strip = 0;
    while (strip < NFOLDS) {

      proc = 0;
      while (proc < CPT_NPROCS) {
        top += CPT_F_X[proc][strip];
        if (pos[0] < top) {
          found = 1;
        }			/* End if (pos[0] < top) */
        if (found) break;
        proc++;

      }				/* End while (proc) */
      if (found) {
        x_disp += pos[0]-(top-CPT_F_X[proc][strip]); 
        /* The distance from the current strip start */
        break;
      }				/* End if (found) */
      x_disp += CPT_S_X[cmtWorldRank][strip];
      strip++;	/* This *must* be the last command of the loop! */

    }				/* End while (strip) */

    if (proc != cmtWorldRank) {
      work = 0;
    }				/* End if (proc != cmtWorldRank) */

#ifdef DEBUG_SERV_SET_STATE
    for (z=0; z < CPT_NPROCS; z++) {
      MPI_Barrier (cmtCommCommand);
      if (z == cmtWorldRank) {
        fprintf (stdout, "rank: %i ", cmtWorldRank);
        for (y=0; y < 3; y++) {
          fprintf (stdout, "pos[%i]: %i\t", y, pos[y]);
        }
        fprintf (stdout, "\n");
        fprintf (stdout, "proc: %i, strip: %i, x_disp: %i ", proc, strip, x_disp);
        fprintf (stdout, "work: %i\n", work);
      } 			/* End if (z == work_rank) */
    }				/* End for */

#endif /* DEBUG_SERV_SET_STATE */

  }				/* End if my_x */

  if ((0 != cmtWorldRank) && !work) {
    goto exit;
  }				/* End if ((0 != cmtWorldRank) && !work) */

#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP
/* Allocate memory for per-PE status data. my_x is the number of elements
   required */
  tmp_data = (unsigned char *) malloc (my_x*cpt_state_size[stateid]);
  if (!tmp_data) {
    fprintf (stderr, "serv_set_state: failed to allocate per-PE scratch space\n");
    MPI_Abort (cmtCommCommand, -1);
  }

/* Calculate tmp_data2_size; all PEs must know so as to receive in Scatter */
  tmp_data2_size = 1;
  if (0 == pos[0]) {
    tmp_data2_size = DIMX;
  }				/* End if (0 == pos[0]) */

/* Allocate memory for temporary storage of received state data */
  if (cmtWorldRank == 0) {
    tmp_data2 = (u_char *) malloc (tmp_data2_size*cpt_state_size[stateid]);
    if (!tmp_data2) {
      fprintf (stderr, "serv_set_state: failed to allocate scratch space\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data2) */

    tmp_data3 = (u_char *) malloc (tmp_data2_size*cpt_state_size[stateid]);
    
    if (!tmp_data3) {
      fprintf (stderr, "serv_set_state: malloc failed, exiting\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data3) */

  }				/* End if (0 == cmtWorldRank) */

#else /* EVEN_DECOMP */

/* Allocate memory for per-PE status data. my_x is the number of elements
   required for each line */
  if (work) {
    tmp_data = (unsigned char *) malloc (my_x*my_y*my_z*cpt_state_size[stateid]);

    if (!tmp_data) {
      fprintf (stderr, "serv_set_state: failed to allocate scratch space\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data) */
  }				/* End if (work) */

/* Allocate memory for temporary storage of received state data.
   Need enough space either to hold DIMX, or just 1 element from the
   working process */
  if (0 == cmtWorldRank) {
/* Calculate tmp_data2_size */
    tmp_data2_size = 1;

    if (1 != my_x) {
      tmp_data2_size = DIMX;
    }				/* End if (1 != my_x) */

    tmp_data2 = (u_char *) malloc (tmp_data2_size*my_y*my_z*cpt_state_size[stateid]);

    if (!tmp_data2) {
      fprintf (stderr, "serv_set_state: failed to allocate scratch space\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!tmp_data2) */
  }				/* End if (0 == cmtWorldRank) */

#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP 
/* Moving down:
   The initial statements were 
    cp1 = CA_REF( ca1, Radius, Radius, Radius );
    cp2 = CA_REF( ca2, Radius, Radius, Radius );
   to handle halo displacement. This statement takes us to the first "real 
   datum". 

   z-axis displacement:
   We need to move DIMY*CPT_X to go to the end of a plane, and then need 
   to skip 2*Radius*CPT_X to go to the first "real datum" of new plane.
   This sums up to (DIMY+2*Radius)*CPT_X, which is equal to CPT_Y*CPT_X.
   - How many planes do we move down?
     if (0 == pos[2]) we must not move at all, else pos[2]-1.

   y-axis displacement:
  Skiping CPT_X elements takes us a y-line down. Similarly with above, 
   we need either naught or (pos[1]-1)*CPT_X elements skipped.

   x-axis displacement:
   We move down (pos[0]-1)%CPT_DIMX positions if (pos[0] <> 0)

*/

/* Calculate displacements */
  z_disp = y_disp = x_disp = Radius;
#else
/* x_disp has already been taken care of */
  z_disp = y_disp = Radius;
#endif /* EVEN_DECOMP */
  if (0 != pos[2]) {
    z_disp += pos[2]-1;
  }				/* End if (pos[2]) */

  if (0 != pos[1]) {
    y_disp += pos[1]-1;
  }				/* End if (pos[1]) */

#ifdef EVEN_DECOMP
  if (0 != pos[0]) {
/* (pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X moves us to the right strip
   (pos[0]-1)%CPT_F_X moves us to the right column in the strip. */
    x_disp += (((pos[0]-1)%CPT_F_X)+((pos[0]-1)/(DIMX/NFOLDS)*CPT_S_X));
  }				/* End if (pos[0]) */
#else
/* x_disp already done */
#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP
/* readn counts in chars (ie bytes) */
    if (0 == cmtWorldRank) {
      size = tmp_data2_size*cpt_state_size[stateid];
    }			/* End if (0 == cmtWorldRank) */

    for (z = z_start; z < z_end; z++) {

      for (y = y_start; y < y_end; y++) {

        cp1 = CA_REF (ca1, z_disp+z, y_disp+y, x_disp);
        cp2 = CA_REF (ca2, z_disp+z, y_disp+y, x_disp);

	if (0 == cmtWorldRank) {

          if (size != readn (prot_sockfd, (char *) tmp_data2, size)) {
            fprintf (stderr, "serv_set_state: readn error!\n");
            MPI_Abort (cmtCommCommand, -1);
          }			/* End if readn */

          get_scatter_ptr (&scatter_ptr, tmp_data2, tmp_data3, 
                           tmp_data2_size, cpt_state_size[stateid]);

	}			/* End if (0 == cmtWorldRank) */

	MPI_Scatter (scatter_ptr, my_x, cpt_state_mpidt[stateid],
		     tmp_data, my_x, cpt_state_mpidt[stateid],
		     0, cmtCommCommand);

        if (work) {
          set_x_line (cp1, stateid, tmp_data, my_x);
          set_x_line (cp2, stateid, tmp_data, my_x);
        }			/* End if (work) */

      }				/* End for(y) */

    }				/* End for(z) */
#else
/* Read data from socket */
  if (0 == cmtWorldRank) {
    size = tmp_data2_size*my_y*my_z*cpt_state_size[stateid];
    if (size != readn (prot_sockfd, (char *) tmp_data2, size)) {
      fprintf (stderr, "serv_set_state: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }			/* End if readn */
  }			/* End if (0 == cmtWorldRank) */

/* Distribute data. If (1 != my_x) multiple point to point operations.
   Remember that in this case the whole x-dimension is assumed.
   If (1 == my_x) (which is expected in this case) easier */

  if (1 != my_x) {

/* For each process role (send-receive) need 2 types of vectors, a 
   small and a large one. The large can be found in CPT_F_X[last][0],
   since it is decided to distribute extraneous cells to the processes
   from last to first and place them in strips from first to last.
   Similarly, the smallest strip will be in proc 0, designated as
   first_small_strip_ind[0]. See calc_x_sizes() for more. */

/* Check above assumption */
    if (CPT_F_X[last][0] < CPT_F_X[0][first_small_strip_ind[0]]) {
      fprintf (stderr, "%s\n", 
               "serv_set_state: assumption about large and small strip failed");
      fprintf (stderr, "large: %i, small: %i\n", CPT_F_X[last][0],
               CPT_F_X[0][first_small_strip_ind[0]]);
      MPI_Abort (cmtCommCommand, -1);
    }					/* End if (large < small) */

/* Send types. */
    MPI_Type_vector (my_y*my_z, CPT_F_X[last][0], tmp_data2_size,
                     cpt_state_mpidt[stateid], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[0]) {
      MPI_Type_vector (my_y*my_z, CPT_F_X[0][first_small_strip_ind[0]],
                       tmp_data2_size, cpt_state_mpidt[stateid], 
                       &send_strip_vec_small);
      MPI_Type_commit (&send_strip_vec_small);
    }			/* End if (first_small_strip_ind[0]) */

/* Receive types. */
    MPI_Type_vector (my_y*my_z, CPT_F_X[cmtWorldRank][0], my_x,
                     cpt_state_mpidt[stateid], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);
    if (NFOLDS != first_small_strip_ind[cmtWorldRank]) {
      MPI_Type_vector (my_y*my_z, 
                     CPT_F_X[cmtWorldRank][first_small_strip_ind[cmtWorldRank]],
                     my_x, cpt_state_mpidt[stateid], &recv_strip_vec_small);
      MPI_Type_commit (&recv_strip_vec_small);
    }			/* End if (first_small_strip_ind[cmtWorldRank]) */

/* Send data */
    if (0 == cmtWorldRank) {
/* First run strip then run processor, so as to traverse tmp_data2 linearly.
   tmp_data2 contains the data in physical order. Going down the x-axis
   one meets first strip 0 of process 1 and then strip 1 of process 0 */

      i = 0;		/* Index to position in tmp_data2[] */
      tag = 0;		/* Tag for comm and request[] index */

      for (strip = 0; strip < NFOLDS; strip++) {

        for (proc = 0; proc < CPT_NPROCS; proc++) {
          if (strip < first_small_strip_ind[proc]) {
#ifdef DEBUG_SERV_SET_STATE
            fprintf (stdout, "Sent LARGE proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_SERV_SET_STATE */
            MPI_Isend (&tmp_data2[i], 1, send_strip_vec_large, proc, 
                       tag, cmtCommCommand, &request[tag]);
          } else {
#ifdef DEBUG_SERV_SET_STATE
            fprintf (stdout, "Sent SMALL proc %i, strip %i, tag %i\n", proc, strip, tag);
#endif /* DEBUG_SERV_SET_STATE */
            MPI_Isend (&tmp_data2[i], 1, send_strip_vec_small, proc, 
                       tag, cmtCommCommand, &request[tag]);
          }				/* End if strip */
          i += CPT_F_X[proc][strip]*cpt_state_size[stateid];
          tag++;

        }				/* End for (proc) */
      }					/* End for (strip) */
    }				/* End if (0 == cmtWorldRank) */

/* Receive data */
    i = 0;		/* Index to position in tmp_data[] */
    for (strip = 0; strip < NFOLDS; strip++) {
      tag = strip*CPT_NPROCS + cmtWorldRank;
      if (strip < first_small_strip_ind[cmtWorldRank]) {
#ifdef DEBUG_SERV_SET_STATE
        fprintf (stdout, "proc %i, Recv LARGE strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_SERV_SET_STATE */
        MPI_Recv (&tmp_data[i], 1, recv_strip_vec_large, 0, 
                  tag, cmtCommCommand, &status[cmtWorldRank]);
      } else {
#ifdef DEBUG_SERV_SET_STATE
          fprintf (stdout, "proc %i, Recv SMALL strip %i, tag %i\n", cmtWorldRank, strip, tag);
#endif /* DEBUG_SERV_SET_STATE */
        MPI_Recv (&tmp_data[i], 1, recv_strip_vec_small, 0, 
                  tag, cmtCommCommand, &status[cmtWorldRank]);
      }				/* End if strip */
      i += CPT_F_X[cmtWorldRank][strip]*cpt_state_size[stateid];

    }				/* End for (strip) */

    if (0 == cmtWorldRank) {
      if (MPI_SUCCESS != MPI_Waitall (CPT_NPROCS*NFOLDS, request, status)) {
        fprintf (stderr, "serv_set_state: MPI_Waitall failed\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (MPI_SUCCESS != MPI_Waitall) */
    }				/* End if (0 == cmtWorldRank) */

  }				/* End if (1 != my_x) */

  else { /* (1 == my_x) */

/* Send types. */
    MPI_Type_vector (my_y*my_z, 1 , tmp_data2_size,
                     cpt_state_mpidt[stateid], &send_strip_vec_large);
    MPI_Type_commit (&send_strip_vec_large);

/* Receive types. */
    MPI_Type_vector (my_y*my_z, 1, my_x,
                     cpt_state_mpidt[stateid], &recv_strip_vec_large);
    MPI_Type_commit (&recv_strip_vec_large);

/* Send data */
    if (0 == cmtWorldRank) {

#ifdef DEBUG_SERV_SET_STATE
      fprintf (stdout, "Send LARGE proc %i, strip %i, tag 0\n", proc, strip);
#endif /* DEBUG_SERV_SET_STATE */

      MPI_Isend (tmp_data2, 1, send_strip_vec_large, proc, 
                 0, cmtCommCommand, &request[0]);
    }				/* End if (0 == cmtWorldRank) */

/* Recv data */
    if (proc == cmtWorldRank) {
#ifdef DEBUG_SERV_SET_STATE
      fprintf (stdout, "proc %i, Recv LARGE strip %i, tag 0\n", cmtWorldRank, strip);
#endif /* DEBUG_SERV_SET_STATE */
      MPI_Recv (tmp_data, 1, recv_strip_vec_large, 0, 
                0, cmtCommCommand, &status[cmtWorldRank]);
    }				/* End if (proc == cmtWorldRank) */

    if (0 == cmtWorldRank) {
      if (MPI_SUCCESS != MPI_Waitall (1, request, status)) {
        fprintf (stderr, "serv_set_state: MPI_Waitall failed\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if (MPI_SUCCESS != MPI_Waitall) */
    }				/* End if (0 == cmtWorldRank) */

  }				/* End else */

  if (work) {
    ptr1 = tmp_data;
    advance = my_x*cpt_state_size[stateid]; /* Calc taken out */

    for (z = z_start; z < z_end; z++) {

      for (y = y_start; y < y_end; y++) {

        cp1 = CA_REF (ca1, z_disp+z, y_disp+y, x_disp);
        cp2 = CA_REF (ca2, z_disp+z, y_disp+y, x_disp);

          set_x_line (cp1, stateid, ptr1, my_x);
          set_x_line (cp2, stateid, ptr1, my_x);
          ptr1 += advance;

      }				/* End for(y) */

    }				/* End for(z) */
#ifdef EVEN_DECOMP
#else
    free (tmp_data);
#endif /* EVEN_DECOMP */

  }				/* End if (work) */
#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP
  free (tmp_data);
#else
#endif /* EVEN_DECOMP */

#ifdef EVEN_DECOMP
#else
  MPI_Type_free (&send_strip_vec_large);
  if ((NFOLDS != first_small_strip_ind[cmtWorldRank]) && (1 != my_x)) {
    MPI_Type_free (&send_strip_vec_small);
  }			/* End if (first_small_strip_ind[cmtWorldRank]) */
  MPI_Type_free (&recv_strip_vec_large);
  if ((NFOLDS != first_small_strip_ind[0]) && (1 != my_x)) {
    MPI_Type_free (&recv_strip_vec_small);
  }			/* End if (first_small_strip_ind[0]) */
#endif /* EVEN_DECOMP */

  if (0 == cmtWorldRank) {
    free (tmp_data2);
#ifdef EVEN_DECOMP 
    free (tmp_data3);
#else
#endif /* EVEN_DECOMP */

  }				/* End if (0 == cmtWorldRank) */

exit:
  return (SET_STATE);

} /* End serv_set_state */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_set_param
 * 
 * Reads global CA parameters. Adapted from cmt_read_global()
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads in, checks and sets the parameters. Loops over the 
 *	no_of_params to set the parameters to be changed. MPI_Bcasts
 * 	the new version of cpt_globpar[].
 *
 * Returns: (req_code)
 *	(SET_PARAM)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_set_param (void) {
  register size_t size;
  register int i;
  int no_of_params;
  int param_id;

  if (0 == cmtWorldRank) {
#ifdef DEVELOP
    for (i = 0; i < NumOfGlobPar; i++) {
      fprintf (stdout, "cpt_globpar[%i]: %f\n", i, cpt_globpar[i]);
    }				/* End if for */
#endif /* DEVELOP */

/* Read no_of_params */
    size = sizeof (no_of_params);
    if (size != readn (prot_sockfd, (char *) &no_of_params, size)) {
      fprintf (stderr, "serv_set_param: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */

/* Check for bad no_of_params */
    if ((no_of_params > NumOfGlobPar) || (no_of_params < 0)) {
      fprintf (stderr, "serv_set_param: illegal no_of_params: %i\n", no_of_params);
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if no_of_params */

    for (i = 0; i < no_of_params; i++) {
/* Read param_id */
      size = sizeof (param_id);
      if (size != readn (prot_sockfd, (char *) &param_id, size)) {
        fprintf (stderr, "serv_set_param: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */

/* Check for bad param_id */
      if ((param_id > NumOfGlobPar-1) || (param_id < 0)) {
        fprintf (stderr, "serv_set_param: illegal param_id: %i\n", param_id);
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if param_id */

/* Read cpt_globpar[param_id] */
      size = sizeof (cpt_globpar[0]);
      if (size != readn (prot_sockfd, (char *) &cpt_globpar[param_id], size)) {
        fprintf (stderr, "serv_set_param: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
    }				/* End for */
  }				/* End if (0 == cmtWorldRank) */

/* Broadcast cpt_globpar */
  MPI_Bcast (cpt_globpar, sizeof(cpt_globpar)/sizeof(float), MPI_FLOAT,
	       0, cmtCommCommand);

#ifdef DEVELOP
  if (0 == cmtWorldRank) {
    for (i = 0; i < NumOfGlobPar; i++) {
      fprintf (stdout, "cpt_globpar[%i]: %f\n", i, cpt_globpar[i]);
    }				/* End if for */

  }				/* End if (0 == cmtWorldRank) */
#endif /* DEVELOP */

  return (SET_PARAM);

} /* End serv_set_param */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_get_param
 * 
 * Reads global CA parameters. Adapted from cmt_read_global()
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Only process 0 works in this function. It reads the param_id
 *	checks it and writes the parameter value back to the socket.
 *	If the param_id check fails, it writes IGNORED.
 *
 * Returns: (req_code)
 *	(GET_PARAM)   	on success.
 *	(IGNORED)   	if bad paramid
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_get_param (void) {
  register size_t size;
  int param_id;
  const float ign = IGNORED;

  if (0 == cmtWorldRank) {

/* Read param_id */
    size = sizeof (param_id);
    if (size != readn (prot_sockfd, (char *) &param_id, size)) {
      fprintf (stderr, "serv_get_param: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */

/* Check for bad param_id */
    if ((param_id > NumOfGlobPar-1) || (param_id < 0)) {
      fprintf (stderr, "serv_get_param: illegal param_id: %i\n", param_id);
/* Write IGNORED */
      size = sizeof (cpt_globpar[0]);
      if (size != writen (prot_sockfd, (char *) &ign, size)) {
        fprintf (stderr, "serv_get_param: writen error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if writen */
      return (IGNORED);
    }				/* End if param_id */

/* Write cpt_globpar[param_id] */
    size = sizeof (cpt_globpar[0]);
    if (size != writen (prot_sockfd, (char *) &cpt_globpar[param_id], size)) {
      fprintf (stderr, "serv_get_param: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */
  }				/* End if (0 == cmtWorldRank) */

  return (GET_PARAM);

} /* End serv_get_param */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_terminate
 * 
 * Clears the planes off the lists 
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Calls corresponding functions for all_planes and vis_list.
 *
 * Returns: (req_code)
 *	(TERMINATE)
 *
 *----------------------------------------------------------------------*/

static req_code serv_terminate (void) {

  clear_plane_list (&all_planes); 	/* Deletes planes */ 
  clear_list (&vis_list);		/* Deletes cells only */

  return (TERMINATE);

} /* End serv_terminate */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_add_plane
 * 
 * Adds a plane to the visualisation list.
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads the plane details from the socket then adds it to the list
 *	of planes and the visualisation list. It also writes the ID 
 *	appointed to it to the GUI
 *
 * Returns: (req_code)
 *	(ADD_PLANE)   	on success.
 *	(IGNORED)	if plane already exists
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_add_plane (void) {
  register size_t size = sizeof (int);
  plane *pl_ptr;
  int ret;
  cell *c_ptr;
  int ID_same = -1;
  req_code req;

  pl_ptr = (plane *) calloc (1, sizeof (plane));
  if (!pl_ptr) {
    fprintf (stderr, "serv_add_plane: calloc error, exiting!\n");
    MPI_Abort (cmtCommCommand, -1);
  }                           /* End if (!pl_ptr) */

  if (0 == cmtWorldRank) {
    if (0 > read_plane (prot_sockfd, pl_ptr)) {
      fprintf (stderr, "serv_add_plane: read_plane error, exiting!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if read_plane */

    check_plane (pl_ptr);
  }				/* End if (0 == cmtWorldRank) */

/* Broadcast plane */
  bcast_plane (pl_ptr, cmtCommCommand);

#ifdef DEBUG
    fprintf (stdout, "serv_add_plane: pl_ptr->pt.pos[0]: %i\t", pl_ptr->pt.pos[0]);
    fprintf (stdout, "all_planes.no_planes: %i\n", all_planes.no_planes);
#endif

/* If the ID set by check_plane () is OK, insert the plane to all_planes 
   and fix class_ptr. Also initialise c_ptr */
  if (IGNORED != pl_ptr->ID) {
    ret = add_plane (&all_planes, pl_ptr, &ID_same, 1);
    if ((0 > ret) && (IGNORED != ret)) {
      fprintf (stderr, "serv_add_plane: add_plane error, exiting!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if add_plane failed */

/* Initialise cell */
    c_ptr = (cell *) malloc (sizeof (cell));
    if (!c_ptr) {
      fprintf (stderr, "serv_add_plane: malloc error, exiting!\n");
      MPI_Abort (cmtCommCommand, -1);
    } 				/* End if (!c_ptr) */

    init_cell (c_ptr, pl_ptr, cpt_generation);

  } else {			/* The plane is illegal, c_ptr = 0 */
    c_ptr = (cell *) NULL;
  }				/* End if (IGNORED != pl_ptr->ID) */

/* Write the ID (even if ignored) to the GUI */
  if (0 == cmtWorldRank) {
    if (size != writen (prot_sockfd, (char *) &(pl_ptr->ID), size)) {
      fprintf (stderr, "serv_add_plane: writen error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if writen */
/* Write ID_same to the GUI if plane already exists */
    if (-1 != ID_same) {
      if (size != writen (prot_sockfd, (char *) &ID_same, size)) {
        fprintf (stderr, "serv_add_plane: writen error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if writen */
    }				/* End if (-1 != ID_same) */

#ifdef DEVELOP
    fprintf (stdout, "ADD_PLANE: ID: %i\n", pl_ptr->ID);
#endif /* DEVELOP */
  }				/* End if (0 == cmtWorldRank) */

  if (0 > pl_ptr->ID) {
    if (c_ptr) {
      pl_ptr->ID = ID_same;	/* Nasty hack for immediate visualisation */
      tx_vis_pack (c_ptr, (char) 1); /* Force visualisation */
      free (c_ptr);	
    }				/* End if (c_ptr) */

    disc_plane (&pl_ptr);

    return (IGNORED);
  }

/* Insert pl_ptr to the visualisation list */
  insert (&vis_list, c_ptr);

#ifdef DEBUG
  if (0 == cmtWorldRank) {
    traverse (vis_list, '%');
  }
#endif

  if (0 == cmtWorldRank) {
/* Send ack first, to avoid consume_vis_pack() in immediate visualisation */
    if (ADD_PLANE != send_ack (ADD_PLANE)) {
      fprintf (stderr, "serv_add_plane: send_ack() failed\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (ADD_PLANE != send_ack) */

/* Read VIS_PACK */
    if (size != readn (prot_sockfd, (char *) &req, size)) {
      fprintf (stderr, "serv_add_plane: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */

/* This *must* be VIS_PACK, else the protocol is broken. Hack to avoid
   Bcast of what in fact is an acknowledgement. */
    if (VIS_PACK != req) {
      fprintf (stderr, "serv_add_plane: VIS_PACK expected, %i received\n",
                        req);
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!VIS_PACK) */

  }				/* End if (0 == cmtWorldRank) */

/* Immediate visualisation */
  tx_vis_pack (c_ptr, (char) 1); /* Force visualisation */

  return (ADD_PLANE);

} /* End serv_add_plane */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_del_plane
 * 
 * Deletes a plane from the plane and visualisation lists
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads the plane ID from the socket, then deletes it from the 
 *	2 lists.
 *
 * Returns: (req_code)
 *	(DEL_PLANE)   	on success.
 *	(IGNORED)	if plane does not exist
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_del_plane (void) {
  int ID;
  req_code ret;
  register size_t size = sizeof (ID);

  if (0 == cmtWorldRank) {
/* Read ID */
    if (size != readn (prot_sockfd, (char *) &ID, size)) {
      fprintf (stderr, "serv_del_plane: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */
  }				/* End if (0 == cmtWorldRank) */

/* Broadcast ID */
  MPI_Bcast ((int *) &ID, 1, MPI_INT, 0, cmtCommCommand);

  if (0 > (ret = del_ID (&vis_list, ID))) {
    fprintf (stderr, "serv_del_plane: warning, ID %i not in vis_list\n", ID);
    return (ret);
  }				/* End if (0 > del_ID) */

  if (0 > (ret = rem_plane (&all_planes, ID))) {
    fprintf (stderr, "serv_del_plane: warning, ID %i not in all_planes\n", ID);
    return (ret);
  }				/* End if (0 > rem_plane) */

  return (ret);

} /* End serv_del_plane */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_proj_read
 * 
 * Reads a given project file to update the state of the Engine
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads the name of the file from the socket, then calls cmt_read_all
 *
 * Returns: (req_code)
 *	(READ_PROJECT) 	on success.
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_proj_read (void) {
  FILE *fout;
  static char in_basename[PATH_MAX] = "";
  int length;
  req_code ret;
  char *fname;
  char *c;
  register size_t size;

  if (0 == cmtWorldRank) {

/* Read filename length as read by strlen () on the other side */
    size = sizeof (int);
    if (size != readn (prot_sockfd, (char *) &length, size)) {
      fprintf (stderr, "serv_proj_read: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */
#ifdef DEBUG
    fprintf (stdout, "serv_proj_read: Got length: %i\n", length);
#endif

/* malloc fname to fit terminating '\0' */
    if (0== (fname = (char *) malloc ((length+1)*sizeof (char)))) {
      fprintf (stderr, "serv_proj_read: malloc error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if malloc */

/* Read filename */
    if (length != readn (prot_sockfd, fname, length)) {
      fprintf (stderr, "serv_proj_read: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */

/*  Terminate fname */
    fname [length] = (char) 0;
#ifdef DEBUG
    fprintf (stdout, "serv_proj_read fname: %s\n", fname);
#endif
    strncpy (in_basename, fname, length+1);

    c = strrchr (in_basename, '.');
    if (c != NULL)
      *c = (char) 0;

  } 				/* End if (0 == cmtWorldRank) */

  if (0 > (ret = (req_code) cmt_read_all (in_basename))) {
    fprintf (stderr, "serv_proj_read: cmt_read_all failed\n");
    return (ret);
  }

  return (READ_PROJECT);
} /* End serv_proj_read */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_proj_save
 * 
 * Saves all the CA Engine data to project and substate files
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads filename root from socket then constructs appropriate
 *	.cpj and .cmt files to store the data.
 *
 * Returns: (req_code)
 *	(SAVE_PROJECT) 	on success.
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_proj_save (void) {
  static char out_name[PATH_MAX];
  static char filename[PATH_MAX];
  char *dirname = NULL, *basename = NULL;
  int length;
  int ret, i;
  char *fname;
  char *c;
  register size_t size;

  if (0 == cmtWorldRank) {
    size = sizeof (int);

/* Read filename length as read by strlen () on the other side */
    if (size != readn (prot_sockfd, (char *) &length, size)) {
      fprintf (stderr, "serv_proj_save: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */
#ifdef DEBUG
    fprintf (stdout, "serv_proj_save: Got length: %i\n", length);
#endif

/* malloc fname to fit terminating '\0' */
    if (0== (fname = (char *) malloc ((length+1)*sizeof (char)))) {
      fprintf (stderr, "serv_proj_save: malloc error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if malloc */

/* Read filename */
    if (length != readn (prot_sockfd, fname, length)) {
      fprintf (stderr, "serv_proj_save: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */
/*  Terminate fname */
    fname [length] = (char) 0;
#ifdef DEBUG
    fprintf (stdout, "serv_proj_save fname: %s\n", fname);
#endif

    strncpy (out_name, fname, sizeof(out_name));
  
    c = strrchr (out_name, '/');

    if (c != NULL) {
      *c = (char) 0;
      dirname = out_name;
      basename = c+1;
    }

    else {
      dirname = NULL;
      basename = out_name;
    }

    c = strrchr (basename, '.');
    if (c != NULL)
      *c = (char) 0;

    if (dirname != NULL)
      sprintf (filename, "%s/%s.cpj", dirname, basename);
    else
      sprintf (filename, "%s.cpj", basename);

    cmt_write_global (filename);
    } 				/* End if (0 == cmtWorldRank) */

    for (i = 0; i < NumOfStates; i++) {

      if (cmtWorldRank == 0) {

        if (dirname != NULL)
	  sprintf (filename, "%s/%s%c%c%c.cmt", dirname, basename,
		  '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
	else
	  sprintf (filename, "%s%c%c%c.cmt", basename,
		   '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
      }					/* End if (cmtWorldRank == 0) */

        cmt_write (filename, i);
    }					/* End for (i) */

  return (SAVE_PROJECT);

} /* End serv_proj_save */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_periodic_save
 * 
 * Assigns the save_step and filename root for periodic full CA Engine
 * state saves
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Reads the name of the file and the save_step from the socket, then 
 *	broadcasts save_step to all the other cells.
 *
 * Returns: (req_code)
 *	(PERIODIC_SAVE)	on success.
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_periodic_save (void) {
  static char out_name[PATH_MAX];
  int length;
  int ret, cfs;
  char *c;
  register const size_t size = sizeof (int);

  if (0 == cmtWorldRank) {

/* Read filename length as read by strlen () on the other side */
    if (size != readn (prot_sockfd, (char *) &length, size)) {
      fprintf (stderr, "serv_periodic_save: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */
#ifdef DEBUG
    fprintf (stdout, "serv_periodic_save: Got length: %i\n", length);
#endif

/* Read filename */
    if (length != readn (prot_sockfd, out_name, length)) {
      fprintf (stderr, "serv_periodic_save: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }                         /* End if readn */
/*  Terminate fname */
    out_name [length] = (char) 0;

/* Read save_step */
    if (size != readn (prot_sockfd, (char *) &save_step, size)) {
      fprintf (stderr, "serv_periodic_save: readn error!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if readn */
#ifdef DEBUG
    fprintf (stdout, "serv_periodic_save: Got save_step: %i\n", save_step);
#endif

    c = strrchr (out_name, '/');

    if (c != NULL) {
      *c = (char) 0;
      out_dirname = out_name;
      out_basename = c+1;
    }

    else {
      out_dirname = NULL;
      out_basename = out_name;
    }

    c = strrchr (out_basename, '.');
    if (c != NULL) {
      *c = (char) 0;
    }

    if (OVER_W == (cfs = check_fs (out_dirname, out_basename))) {
      fprintf (stderr, "serv_periodic_save: Warning: periodic save files will be overwritten!\n");
    } else if (0 > cfs) {
      fprintf (stderr, "serv_periodic_save: Error: check_fs failed!\n");
    }				/* End if (OVER_W) */

/* Writen cfs */
    if (size != writen (prot_sockfd, (char *) &cfs, size)) {
      fprintf (stderr, "serv_periodic_save: writen failed!\n");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End bad writen */
  } 				/* End if (0 == cmtWorldRank) */

/* All the cells execute MPI_Bcast to get save_step */
  MPI_Bcast ((int *) &save_step, 1, MPI_INT, 0, cmtCommCommand);

  return (PERIODIC_SAVE);

} /* End serv_periodic_save */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: serv_set_minmax
 * 
 * Sets the automatic minimum and maximum for a substate
 * 
 * Parameters:
 *	void
 * 
 * Implementation:
 *	Process 0 reads the substate and l_minmax from the socket and Bcasts
 *	them to the other processes. The auto_map flag and minmax[] values
 *	are set after testing the substate and l_minmax read. The system
 *	returns to automap mode if l_minmax[] are both equal to 0.
 *
 * Returns: (req_code)
 *	(SET_MINMAX)	on success.
 *	(IGNORED)	if illegal data.
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

static req_code serv_set_minmax (void) {
    double l_minmax[2];
    size_t size;
    int substate;

    if (0 == cmtWorldRank) {
/* Read substate */
      size = sizeof (int);
      if (size != readn (prot_sockfd, (char *) &substate, size)) {
        fprintf (stderr, "serv_set_minmax: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */

/* Read l_minmax */
      size = 2*sizeof (double);
      if (size != readn (prot_sockfd, (char *) l_minmax, size)) {
        fprintf (stderr, "serv_set_minmax: readn error!\n");
        MPI_Abort (cmtCommCommand, -1);
      }				/* End if readn */
#ifdef DEVELOP
      fprintf (stdout, "serv_set_minmax: Got l_minmax: %i-%i\n", 
               l_minmax[0], l_minmax[1]);
#endif
    }				/* End if (0 == cmtWorldRank) */

/* All cells execute MPI_Bcast to get substate, l_minmax */
    MPI_Bcast (&substate, 1, MPI_INT, 0, cmtCommCommand);
    MPI_Bcast (l_minmax, 2, MPI_DOUBLE, 0, cmtCommCommand);

/* Check substate, l_minmax values */
    if ((substate < 0) || (substate > NumOfStates-1)) {
      fprintf (stdout, "serv_set_minmax: bad substate definition, ignored\n");
      return (IGNORED);
    }				/* End if bad substate */

    if (l_minmax[0] > l_minmax[1])  {
      fprintf (stdout, "serv_set_minmax: bad minmax definition, ignored\n");
      return (IGNORED);
    }				/* End if bad minmax */

/* If they are both equal to zero, return to automap mode */
    if ((0 == l_minmax[0]) && (0 == l_minmax[1])) {
      auto_map[substate] = 1;
      minmax[substate][0] = DBL_MAX;
      minmax[substate][1] = -DBL_MAX;
#ifdef DEVELOP
      fprintf (stdout, "serv_set_fold: manual folds unset\n");
#endif
    } else {			/* End if auto_map */
/* Set auto_map, minmax[substate][] */
      auto_map[substate] = 0;
      minmax[substate][0] = l_minmax[0];
      minmax[substate][1] = l_minmax[1];
    }				/* End if manual mapping */

    return (SET_MINMAX);

} /* End serv_set_minmax */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: send_ack
 * 
 * Sends acknowledgement to GUI. 
 * 
 * Parameters:
 *	I   ack	Int value to be sent to GUI
 * 
 * Implementation:
 * 	Call of writen function checked for errors.
 *
 * Returns: (req_code)
 *	(ack)   on success.
 *	(-ve) on failure.
 *
 *----------------------------------------------------------------------*/

static req_code send_ack (req_code ack) {
  int size = sizeof (ack);

  if (size != writen (prot_sockfd, (char *) &ack, size)) {
    fprintf (stderr, "send_ack: writen failed!\n");
    MPI_Abort (cmtCommCommand, -1);
  }				/* End bad writen */
  return ack;
}

/*----------------------------------------------------------------------*/
/* 
 * PUBLIC: cpt_save
 * 
 * Saves all the CA Engine data to project and substate files. Only 
 * available for steering
 * 
 * Parameters:
 *	I   fname	Char *name of file to save data
 * 
 * Implementation:
 *	Just like serv_proj_save(), only without socket interaction
 *
 * Returns: (int)
 *	(SAVE_PROJECT) 	on success.
 *	(-ve) 		on failure.
 *
 *----------------------------------------------------------------------*/

int cpt_save (char *fname) {
  char *dirname = NULL, *basename = NULL;
  char *c;
  char fname2[PATH_MAX];
  char filename[PATH_MAX];
  int i;

  if (!in_steering) {
    fprintf (stderr, "cpt_save available only inside the steering section\n");
    return (0);
  }				/* End if (!in_steering) */

  if (0 == cmtWorldRank) {
  
    strncpy(fname2, fname, PATH_MAX);
    c = strrchr (fname2, '/');

    if (c != NULL) {
      *c = (char) 0;
      dirname = fname2;
      basename = c+1;
    }

    else {
      dirname = NULL;
      basename = fname2;
    }

    c = strrchr (basename, '.');
    if (c != NULL) {
      *c = (char) 0;
    }

    if (dirname != NULL)
      sprintf (filename, "%s/%s.cpj", dirname, basename);
    else
      sprintf (filename, "%s.cpj", basename);

    cmt_write_global (filename);
  } 				/* End if (0 == cmtWorldRank) */

  for (i = 0; i < NumOfStates; i++) {

    if (cmtWorldRank == 0) {

      if (dirname != NULL)
        sprintf (filename, "%s/%s%c%c%c.cmt", dirname, basename,
                 '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
      else
        sprintf (filename, "%s%c%c%c.cmt", basename,
                 '0'+(i/100), '0'+((i%100)/10), '0'+((i%100)%10));
      }					/* End if (cmtWorldRank == 0) */

      cmt_write (filename, i);
  }					/* End for (i) */

  return (SAVE_PROJECT);

} /* End cpt_save */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: check_plane
 *
 * Checks plane members for CA Engine spatial, substate and vis_step
 * consistency, as well as plane definition.
 *
 * Parameters:
 *      IO  pl_ptr	Plane pointer to be checked
 *
 * Implementation:
 *	Checks and corrects plane for CA Engine spatial consistency.
 *	Then checks if it is really a plane, and if substate and vis_step
 * 	are valid.
 *	Sets ID to IGNORED if plane improper, else to -1.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void check_plane (plane *pl_ptr) {
  int *pos = pl_ptr->pt.pos; 	/* Alias */
  int i;			/* Just an index */
  int is_plane = 0;		/* Flag, not binary though */

/* Check point for CA Engine spatial consistency */
  check_pos (pos);

/* Check point for plane definition spatial consistency */
/* If the CA is a line, then pos is a "plane"  (ie acceptable) */
  if ((1 == DIMY) && (1 == DIMZ)) {
    is_plane = 1;
  } 			/* End if ((1 == DIMY) && (1 == DIMZ)) */

  else {

    for (i = 0; i<3; i++) {
      if (pos[i] > 0) {
        is_plane++;
      }			/* End if (pos[i]) */

    } 			/* End for */

  } 			/* End else */

  if (1 != is_plane) {
    fprintf (stderr, "check_plane: plane incorrectly defined, IGNORED\n");
    pl_ptr->ID = IGNORED;
    return;
  }			/* End if (is_plane) */

/* Check substate consistency */
  if ((pl_ptr->substate < 0) || (pl_ptr->substate > NumOfStates)) {
    fprintf (stderr, "check_plane: incorrect substate, plane IGNORED\n");
    pl_ptr->ID = IGNORED;
    return;
  }			/* End if (substate) */

/* Check vis_step consistency */
  if (pl_ptr->vis_step < 1) {
    fprintf (stderr, "check_plane: incorrect vis_step, plane IGNORED\n");
    pl_ptr->ID = IGNORED;
    return;
  }			/* End if (substate) */

  pl_ptr->ID = -1;

} /* End check_plane */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: bcast_plane
 *
 * Broadcast the details of a plane in process 0 to all the 
 * processes in the communicator. 
 *
 * Parameters:
 *	IO  pl_ptr  Plane pointer to the plane to be broadcast
 *	I   comm    MPI_Comm to which data are broadcast
 *
 * Implementation:
 *	MPI_Bcast ID, pos, substate, vis_step.
 *	If (IGNORED == ID) then the rest of the data are not broadcast.
 *	No point in Bcasting class_ptr data since they are set by
 *	add_plane()
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void bcast_plane (plane *pl_ptr, MPI_Comm comm) {

  MPI_Bcast (&(pl_ptr->ID), 1, MPI_INT, 0, comm);
  if (IGNORED == pl_ptr->ID) {
    return;
  }

  MPI_Bcast (pl_ptr->pt.pos, 3, MPI_INT, 0, comm);
  MPI_Bcast (&(pl_ptr->substate), 1, MPI_INT, 0, comm);
  MPI_Bcast (&(pl_ptr->vis_step), 1, MPI_INT, 0, comm);

} /* End bcast_plane */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: check_pos
 *
 * Corrects pos[] for possibly unacceptable values and corrects it
 * Same as GUI_check_pos but with checks against DIM? instead
 * of xyzdims[]
 *
 * Parameters:
 *      IO  pos		Int pointer 
 *
 * Implementation:
 *	If pos[i] is illegal set it to 0. If the dimension is 1, set
 *	the coefficient to 1 to avoid interpreting (0,0,0) as a cube
 *	if it's only a plane.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void check_pos (int *pos) {

  if (pos[0] > DIMX || pos[0] < 0) {
    fprintf (stdout, "x-axis request greater than DIMX ignored\n");
    pos[0] = 0;
  }				/* End if pos[0] */

  if (pos[1] > DIMY || pos[1] < 0) {
    fprintf (stdout, "y-axis request greater than DIMY ignored\n");
    pos[1] = 0;
  }				/* End if pos[1] */
    
  if (pos[2] > DIMZ || pos[2] < 0) {
    fprintf (stdout, "z-axis request greater than DIMZ ignored\n");
    pos[2] = 0;
  }				/* End if pos[2] */

/* If a DIM is 1, then set pos to 1 to enable plane checking */
  if (1 == DIMX) {
    pos[0] = 1;
  }				/* End if (DIMX) */

  if (1 == DIMY) {
    pos[1] = 1;
  }				/* End if (DIMY) */

  if (1 == DIMZ) {
    pos[2] = 1;
  }				/* End if (DIMZ) */

} /* End check_pos */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: get_val_size
 *
 * Determines number of elements of array storing substate values given
 * the pos identifier of the coordinates
 * Same as GUI_get_val_size but with checks against DIM? instead
 * of xyzdims[]
 *
 * Parameters:
 *      I  pos		Int pointer 
 *
 * Implementation:
 *	Trivial
 *
 * Returns: (int)
 * 	(ve) 	on success
 *	(-ve) 	on failure
 *
 *----------------------------------------------------------------------*/

static int get_val_size (const int *pos) {
  int val_size = 1;

    if (0 == pos[0]) {
      val_size *= DIMX;
    }				/* End if (0 == pos[0]) */

    if (0 == pos[1]) {
      val_size *= DIMY;
    }				/* End if (0 == pos[1]) */

    if (0 == pos[2]) {
      val_size *= DIMZ;
    }				/* End if (0 == pos[2]) */

    return val_size;
} /* End get_val_size */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: get_x_line
 * Collects an x-line of given number of data into a previously 
 * initialised buffer. Assumes the source of data is appropriately initialised.
 *
 * Parameters:
 *	I  cp		CptCell ptr to initial position of data in CA Engine
 *	I  substate	Int substate identifier
 *	O  tmp_data	U_char ptr to line data
 *      I  my_x		Int Number of elements
 *
 * Implementation:
 *	If the data are located in more than one fold, it calls 
 *	cpt_get_state with fold-size length argument
 *	and coalesces the results in tmp_data skipping haloes.
 *	N.B.: It does not alter the position (address) of cp.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void get_x_line (CptCell *cp, int substate, u_char *tmp_data, int my_x) {
  int j;	/* DEBUG */
  int f;	/* DEBUG */

  if (1 == my_x) {
    cpt_get_state (cp, substate, tmp_data, my_x);
#ifdef DEBUG
    if (0 != tmp_data[0]) {
      fprintf (stdout, "tmp_data: %i\n", tmp_data[0]);
    }
#endif /* DEBUG */
  }				/* End if (1 == my_x) */

  else {

#ifdef EVEN_DECOMP
    for (f = 0; f < NFOLDS; f++) {
      cpt_get_state (cp, substate, tmp_data, CPT_F_X);
#ifdef DEBUG
      for (j = 0; j < CPT_F_X; j++) {
        if (0 != tmp_data[j]) {
          fprintf (stdout, "x: %i\n", (cmtWorldRank*CPT_F_X)+(f*CPT_NPROCS*CPT_F_X)+j+1);
        }			/* End if (0 != tmp_data[j]) */
      }				/* End for (j) */
#endif /* DEBUG */

      cp += CPT_S_X;
      tmp_data += CPT_F_X*cpt_state_size[substate];
    }				/* End for (f) */

#ifdef DEBUG

/* Return to original position so as to be reused */
    tmp_data -= CPT_DIMX*cpt_state_size[substate];

    for (j=0; j<CPT_DIMX; j++) {
      if ((0!=tmp_data[j])) {
        f = j/CPT_F_X;
        fprintf (stdout, "x: %i\n", (cmtWorldRank*CPT_F_X)+(f*CPT_NPROCS*CPT_F_X)+(j%CPT_F_X)+1);
      }
    }				/* End for (j) */
#endif /* DEBUG */

#else /* EVEN_DECOMP */

    for (f = 0; f < NFOLDS; f++) {
      cpt_get_state (cp, substate, tmp_data, CPT_F_X[cmtWorldRank][f]);
      cp += CPT_S_X[cmtWorldRank][f];
      tmp_data += CPT_F_X[cmtWorldRank][f]*cpt_state_size[substate];
    }				/* End for (f) */

#endif /* EVEN_DECOMP */
  }				/* End else */

} /* End get_x_line */

#ifdef EVEN_DECOMP
/*----------------------------------------------------------------------*/
/*
 * PRIVATE: get_write_ptr
 *
 * Returns the appropriate u_char pointer to write data to a socket
 * after converting folds internal representation to normal line
 *
 * Parameters:
 *	I   tmp_data2	U_char ptr containing initial data
 *	I   tmp_data3   U_char ptr for fold2line use
 *	I   tmp_data3_size Int number of elements in tmp_data3
 *	I   x		Int x-extent of entity (folded)
 *	I   y		Int y-extent of entity
 *	I   z		Int z-extent of entity
 *	I   el_size	Int size of tmp_data/tmp_data3 elements
 *	I   work	Int flag denoting if active cell
 *	I   work_size	Int size of communicator
 *
 * Implementation:
 *	If the data are located in more than one fold, it calls fold2line 
 *	to turn them to line representation else it copies tmp_data2 to
 *	tmp_data3. If process 0 does not work, it skips the memory space 
 * 	in tmp_data2 allocated for process 0.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void get_write_ptr (u_char *tmp_data2, 
                           u_char *tmp_data3, int tmp_data3_size, 
                           int x, int y, int z,
                           int el_size, int work, int work_size) {
  register int i;

/* Only process 0 executes this */
  if (0 != cmtWorldRank) {
    return;
  }				/* End if (0 != cmtWorldRank) */

/* Skip 0 proc data */
  if (!work) {
    tmp_data2 += el_size;
  }				/* End if (!work) */

  while (z) {

    for (i = 0; i < y; i++) {
  
      if (DIMX == tmp_data3_size) { 	/* All cells working */
        fold2line (tmp_data2, tmp_data3, el_size);
      } else {			/* Assumes tmp_data3_size == 1 */
        memcpy (tmp_data3, tmp_data2, el_size);
      }				/* End if (DIMX != tmp_data3_size) */

      tmp_data2 += x*work_size*el_size;
      tmp_data3 += tmp_data3_size*el_size;

    }  				/* End for (i) */

    z--;
  }  				/* End while (z) */

} /* End get_write_ptr */

#endif /* EVEN_DECOMP */

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: set_x_line
 * Sets the substate data of a given-length line. Assumes the source of 
 * data is appropriately initialised.
 *
 * Parameters:
 *	I  cp		CptCell ptr to initial position of data in CA Engine
 *	I  substate	Int substate identifier
 *	O  tmp_data	U_char ptr to line data
 *      I  my_x		Int number of elements
 *
 * Implementation:
 *	Discerns between one element and full line case. In the former case 
 *	it just calls cpt_set_state. Otherwise, if the data are located in 
 *	more than one fold, it calls cpt_set_state NFOLDS times with fold-size
 *	length argument and writes the results to cp skipping haloes.
 *	N.B.: It does not alter the position (address) of cp.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void set_x_line (CptCell *cp, int substate, u_char *tmp_data, int my_x) {
  int j = 0;	/* DEBUG */
  int f;

  if (1 == my_x) {
    cpt_set_state (cp, substate, tmp_data, my_x);
#ifdef DEBUG
    if (0 != tmp_data[0]) {
      fprintf (stdout, "tmp_data: %i\n", tmp_data[0]);
    }
#endif /* DEBUG */
  }				/* End if (1 == my_x) */

  else {

#ifdef EVEN_DECOMP
    for (f = 0; f < NFOLDS; f++) {
      cpt_set_state (cp, substate, tmp_data, CPT_F_X);
#ifdef DEBUG
      for (j = 0; j < CPT_F_X; j++) {
        if (0 != tmp_data[j]) {
          fprintf (stdout, "x: %i\n", (cmtWorldRank*CPT_F_X)+(f*CPT_NPROCS*CPT_F_X)+j+1);
        }			/* End if (0 != tmp_data[j]) */
      }				/* End for (j) */
#endif /* DEBUG */

      cp += CPT_S_X;
      tmp_data += CPT_F_X*cpt_state_size[substate];
    }				/* End for (f) */

#ifdef DEBUG

/* Return to original position so as to be reused */
    tmp_data -= CPT_DIMX*cpt_state_size[substate];

    for (j=0; j<CPT_DIMX; j++) {
      if ((0!=tmp_data[j])) {
        f = j/CPT_F_X;
        fprintf (stdout, "x: %i\n", (cmtWorldRank*CPT_F_X)+(f*CPT_NPROCS*CPT_F_X)+(j%CPT_F_X)+1);
      }
    }				/* End for (j) */
#endif /* DEBUG */

#else /* EVEN_DECOMP */

    for (f = 0; f < NFOLDS; f++) {
      cpt_set_state (cp, substate, tmp_data, CPT_F_X[cmtWorldRank][f]);
      cp += CPT_S_X[cmtWorldRank][f];
      j += CPT_S_X[cmtWorldRank][f];
      if (j > CPT_X) fprintf (stdout, "j: %i, CPT_X: %i\n",j, CPT_X);
      tmp_data += CPT_F_X[cmtWorldRank][f]*cpt_state_size[substate];
    }				/* End for (f) */

#endif /* EVEN_DECOMP */

  }				/* End else */

} /* End set_x_line */

#ifdef EVEN_DECOMP 

/*----------------------------------------------------------------------*/
/*
 * PRIVATE: get_scatter_ptr
 *
 * Returns the appropriate u_char pointer to MPI_Scatter data
 * after converting normal line to folds internal representation.
 *
 * Parameters:
 *	O   scatter_ptr	U_char ptr ptr to point the fold data
 *	I   tmp_data2	U_char ptr containing initial data
 *	I   tmp_data3   U_char ptr for line2fold use
 *	I   tmp_data2_size Int number of elements in tmp_data2
 *	I   el_size	Int size of tmp_data2/tmp_data3 elements
 *
 * Implementation:
 *	If the data are located in more than one fold, it calls line2fold 
 * 	to turn them to fold representation else it returns the address of 
 * 	tmp_data2. 
 *	N.B.: tmp_data3 could be avoided as an argument if a huge array
 *	      is statically allocated in the body of the function
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void get_scatter_ptr (u_char **scatter_ptr, u_char *tmp_data2, 
                           u_char *tmp_data3, int tmp_data2_size, 
                           int el_size) {
  int rank, j, f;  /* DEBUG */

#ifdef DEBUG

  for (j = 0; j < DIMX; j++) {
    if (0 != tmp_data2[j]) {
      f = (j/CPT_F_X)%NFOLDS;
      rank = j/CPT_DIMX;
      fprintf (stdout, "x: %i\n", (rank*CPT_F_X)+(f*CPT_NPROCS*CPT_F_X)+(j%CPT_F_X)+1);
    }
  }

#endif

  if (DIMX == tmp_data2_size) { 	/* All cells working */
    line2fold (tmp_data2, tmp_data3, el_size);
    *scatter_ptr = tmp_data3;

#ifdef DEBUG
    for (j = 0; j < DIMX; j++) {
      if (0 != tmp_data3[j]) {
        fprintf (stdout, "x: %i\n", j+1);
      }
    }
#endif

  }				/* End if (DIMX == tmp_data2_size) */

  else {			/* Assumes tmp_data2_size == 1 */
    *scatter_ptr = tmp_data2;
  }				/* End if (DIMX != tmp_data2_size) */

} /* End get_scatter_ptr */

#endif /* EVEN_DECOMP */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: print_all_stats
 * 
 * Collects and prints the stats from all the processes at root process.
 * 
 * Parameters:
 *	(void)
 * 
 * Implementation:
 *	Creates a derived MPI_Datatype for the stats type (and needs a 
 *	datatype for the timer type) to Gather the statistics instance
 *	at the root process. Requires careful handling of the TOTAL timer
 *	because periodic saves mean that this timer is not stopped when 
 * 	printing the data.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void print_all_stats (void) {
#define T_DIM 8+1 /* Number of elements in type timer */
#define S_DIM 12+1 /* Number of elements in type stats */

  stats all_stats[CPT_NPROCS]; 		/* Receiver at root process */
  static MPI_Datatype *timer_mpi_t = NULL;	/* Datatype for timer */
  MPI_Aint t_displ[T_DIM];
  int t_blocklengths[T_DIM];
  MPI_Datatype t_types[T_DIM];

  static MPI_Datatype *stats_mpi_t = NULL;	/* Datatype for stats */
  MPI_Aint s_displ[S_DIM];
  int s_blocklengths[S_DIM];
  MPI_Datatype s_types[S_DIM];

  MPI_Aint start_address, address;
  timer dummy_t[2];
  int i;

/* Create timer_mpi_t */
  if (NULL == timer_mpi_t) {
    if (!(timer_mpi_t = (MPI_Datatype *) malloc (sizeof (MPI_Datatype)))) {
      fprintf (stderr, "print_all_stats: malloc failed, exiting!\n");
      perror ("");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!malloc) */

    for (i = 0; i < 7; i++) {
      t_blocklengths[i] = 1;
    }
    t_blocklengths[7] = TITLE_LENGTH;
    t_blocklengths[T_DIM-1] = 1;

    for (i = 0; i < 5; i++) {
      t_types[i] = MPI_DOUBLE;
    }
    t_types[5] = MPI_UNSIGNED_CHAR;
    t_types[6] = MPI_UNSIGNED_LONG;
    t_types[7] = MPI_CHAR;
    t_types[T_DIM-1] = MPI_UB;

    t_displ[0] = 0;
    MPI_Address (&(dummy_t[0].start), &start_address);
    MPI_Address (&(dummy_t[0].stop), &address);
    t_displ[1] = address-start_address;
    MPI_Address (&(dummy_t[0].sum), &address);
    t_displ[2] = address-start_address;
    MPI_Address (&(dummy_t[0].best), &address);
    t_displ[3] = address-start_address;
    MPI_Address (&(dummy_t[0].worst), &address);
    t_displ[4] = address-start_address;
    MPI_Address (&(dummy_t[0].started), &address);
    t_displ[5] = address-start_address;
    MPI_Address (&(dummy_t[0].called), &address);
    t_displ[6] = address-start_address;
    MPI_Address (dummy_t[0].title, &address);
    t_displ[7] = address-start_address;
    MPI_Address (&(dummy_t[1].start), &address);
    t_displ[T_DIM-1] = address-start_address;

    MPI_Type_struct (T_DIM, t_blocklengths, t_displ, t_types, timer_mpi_t);
    MPI_Type_commit (timer_mpi_t);
  }				/* End if (NULL == timer_mpi_t) */


/* Create stats_mpi_t */
  if (NULL == stats_mpi_t) {
    if (!(stats_mpi_t = (MPI_Datatype *) malloc (sizeof (MPI_Datatype)))) {
      fprintf (stderr, "print_all_stats: malloc failed, exiting!\n");
      perror ("");
      MPI_Abort (cmtCommCommand, -1);
    }				/* End if (!malloc) */

    for (i = 0; i < S_DIM; i++) {
      s_blocklengths[i] = 1;
    }

    for (i = 0; i < 6; i++) {
      s_types[i] = *timer_mpi_t;
    }
    s_types[6] = MPI_INT;
    s_types[7] = MPI_UNSIGNED_LONG;
    s_types[8] = MPI_INT;
    s_types[9] = MPI_UNSIGNED;
    s_types[10] = MPI_UNSIGNED_CHAR;
    s_types[11] = MPI_LONG;	/* Dangerous? this is a FILE * I am not using */
    s_types[S_DIM-1] = MPI_UB;

    s_displ[0] = 0;
    MPI_Address (&(all_stats[0].func), &start_address);
    MPI_Address (&(all_stats[0].vis), &address);
    s_displ[1] = address-start_address;
    MPI_Address (&(all_stats[0].prj), &address);
    s_displ[2] = address-start_address;
    MPI_Address (&(all_stats[0].bound), &address);
    s_displ[3] = address-start_address;
    MPI_Address (&(all_stats[0].steer), &address);
    s_displ[4] = address-start_address;
    MPI_Address (&(all_stats[0].total), &address);
    s_displ[5] = address-start_address;
    MPI_Address (&(all_stats[0].rank), &address);
    s_displ[6] = address-start_address;
    MPI_Address (&(all_stats[0].gens), &address);
    s_displ[7] = address-start_address;
    MPI_Address (&(all_stats[0].start_gen), &address);
    s_displ[8] = address-start_address;
    MPI_Address (&(all_stats[0].period), &address);
    s_displ[9] = address-start_address;
    MPI_Address (&(all_stats[0].work), &address);
    s_displ[10] = address-start_address;
    MPI_Address (&(all_stats[0].outfile), &address);
    s_displ[11] = address-start_address;
    MPI_Address (&(all_stats[1].func), &address);
    s_displ[S_DIM-1] = address-start_address;

    MPI_Type_struct (S_DIM, s_blocklengths, s_displ, s_types, stats_mpi_t);
    MPI_Type_commit (stats_mpi_t);
  }			/* End if (NULL == stats_mpi_t) */

/* Stop timer TOTAL before gathering stats */
  stop_one_timer_sts (&statistics, TOTAL, cpt_generation);

/* Gather */
  MPI_Gather (&statistics, 1, *stats_mpi_t,
              all_stats, 1, *stats_mpi_t, 0, MPI_COMM_WORLD);

/* Set outfile member same as the one in root process. This is no limitation
   given that the filename information is passed as a command-line arg */
  if (0 == cmtWorldRank) {
    for (i = 0; i < CPT_NPROCS; i++) {
      all_stats[i].outfile = statistics.outfile;
      print_sts (&all_stats[i]);
    }
  }

/* Reset the stats */
  reset_stats (&statistics);
/* Reset timer TOTAL, *after* reseting the stats */
  start_one_timer_sts (&statistics, TOTAL, cpt_generation);

} /* End print_all_stats */

/*----------------------------------------------------------------------
 Randomisation Functions
 ----------------------------------------------------------------------*/
/*----------------------------------------------------------------------*/
/*
 * PUBLIC: cpt_randomize
 *
 * Creates a new seed for the (pseudo)random number generator.
 *
 * Parameters:
 *	void
 *
 * Implementation:
 *	Philosophy is to call srand() with the current time as a seed.
 *	Contains wrapper to ensure it is not called with the same seed
 *	so as to avoid all cells starting with the same value.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

void cpt_randomize (void) {
  static unsigned int then;
  static short called = 0;
  unsigned int now = (unsigned int) time((time_t *) 0);

  if (!called) {
    srand (now);
    called = 1;
    then = now;
  }

  if (now != then) {
    srand (now);
    then = now;
  }

} /* End cpt_randomize */

/*----------------------------------------------------------------------*/
/*
 * PUBLIC: cpt_srandom
 *
 * Creates a new seed for the (pseudo)random number generator, only that 
 * it takes a user defined argument.
 *
 * Parameters:
 *	I s	Unsigned int randomisation seed
 *
 * Implementation:
 *	Same as cpt_randomize.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

void cpt_srandom (unsigned int s) {
  static unsigned int prev;
  static short called = 0;

  if (!called) {
    srand (s);
    called = 1;
    prev = s;
  }

  if (s != prev) {
    srand (s);
    prev = s;
  }

} /* End cpt_srandom */

/*----------------------------------------------------------------------
 Timer library
 ----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: init_tmr
 * 
 * Initialises a timer struct. Assumes it is malloced.
 * 
 * Parameters:
 *	O   tp		Timer ptr to be initialised
 *	I   title	Const char ptr to be the title member
 * 
 * Implementation:
 *	Trivial
 *	If title is NULL, no title initialisation takes place.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void init_tmr (timer *tp, const char *title) {
  u_int size;

  tp->started = 0;
  tp->sum = (float) 0;
  tp->called = (unsigned long) 0;
  tp->best = DBL_MAX;
  tp->worst = (double) 0;
  if (NULL != title) {
    size  = ((TITLE_LENGTH-1 < strlen (title)) ? TITLE_LENGTH-1 : strlen (title));
    strncpy (tp->title, title, size);
    if ((TITLE_LENGTH-1) < strlen (title)) {
      fprintf (stderr, "init_tmr: title truncated to %s\n", tp->title);
    }			/* End if ((TITLE_LENGTH-1) < strlen (title)) */
  }			/* End if (NULL != tp->title) */

} /* End init_tmr */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: start_tmr
 * 
 * Starts a timer.
 * 
 * Parameters:
 *	O   tp		Timer ptr to be started
 * 
 * Implementation:
 *	Checks if already started, in which case it does nothing more. 
 *	Else it sets the started flag, increases the called member and 
 *	records the start time.
 *
 * Returns: (int)
 *	(ve) 	on success
 *	(0) 	if already started
 *
 *----------------------------------------------------------------------*/

static int start_tmr (timer *tp) {

  if (tp->started) {
#ifdef DEBUG 
    fprintf (stderr, "start_tmr: Timer \"%s\" already started!\n", tp->title);
#endif /* DEBUG */
    return (0);
  }

  tp->started = 1;
  tp->called++;
  
  tp->start = MPI_Wtime();

  return (1);

} /* End start_tmr */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: stop_tmr
 * 
 * Stops a timer.
 * 
 * Parameters:
 *	O   tp		Timer ptr to be stopped
 * 
 * Implementation:
 *	Checks if already stopped, in which case it does nothing more. 
 *	Else it clears the started flag, records the stop time and adds
 *	the time of this session to the sum.
 *
 * Returns: (int)
 *	(ve) 	on success
 *	(0) 	if already stopped
 *
 *----------------------------------------------------------------------*/

static int stop_tmr (timer *tp) {
  double time;

  if (!(tp->started)) {
#ifdef DEBUG
    fprintf (stderr, "stop_tmr: Timer \"%s\" not started!\n", tp->title);
#endif /* DEBUG */
    return (0);
  }

  tp->started = 0;
  
  tp->stop = MPI_Wtime();

  (tp->sum) += (time = (tp->stop)-(tp->start));

  if (time > tp->worst) {
    tp->worst = time;
  }

  if (time < tp->best) {
    tp->best = time;
  }

  return (1);

} /* End stop_tmr */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: print_tmr
 * 
 * Prints the data for a timer.
 * 
 * Parameters:
 *	I   tp		Timer ptr to be printed
 *	O   f		FILE ptr where the data are written
 * 
 * Implementation:
 *	Checks if timer started, in which case it prints a warning.
 *	It prints the title followed by two tabs and the called member.
 *	It then prints one or two tabs according to the length of called 
 *	and prints the sum. All the output is in one line. 
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void print_tmr (const timer *tp, FILE *f) {

  if (tp->started) {
    fprintf (f, "print_tmr: %s timer still running!\n", tp->title);
  }			/* End if (tp->started) */

  fprintf (f, "%s:\t%i", tp->title, tp->called);
  if (10000000 > tp->called) {
    fprintf (f, "\t");
  }			/* End if (small tp->called) */
  fprintf (f, "\t%f", tp->sum);

  if (tp->called) {
    fprintf (f, "\t%f\t%f", tp->best, tp->worst);
  }

  fprintf (f,"\n");


} /* End print_tmr */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: init_sts
 * 
 * Initialises a stats structure
 * 
 * Parameters:
 *	O   stp		Stats ptr to be initialised
 *	I   rank	Int value of rank member to be set (process rank)
 *	I   work	U_char flag for work
 *	I   fname  	Char ptr for output of stats
 * 
 * Implementation:
 *	Sets work and exits if its cleared. Else it initialises the timers
 *	and set the rank and gens.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void init_sts (stats *stp, int rank, u_char work, 
	   	      u_int period, const char *fname) {
  FILE *f;

  stp->work = work;

  if (!(stp->work)) {
    return;
  } 			/* End if (!(stp->work)) */

  init_tmr (&(stp->func), "Update Function\t");
  init_tmr (&(stp->vis), "Visualisation\t");
  init_tmr (&(stp->prj), "Periodic Save\t");
  init_tmr (&(stp->bound), "Boundary Comm\t");
  init_tmr (&(stp->steer), "Steering\t");
  init_tmr (&(stp->total), "Total Execution Time");
  stp->rank = rank;

/* Open file */
  if (0 == cmtWorldRank) {
    if ((NULL == fname) || (!strcmp (fname, "-"))) {
      f = stdout;
    } else {
      if (!(f = fopen (fname, "w"))) {
        fprintf (stderr, "init_sts: cannot fopen %s\n", fname);
        perror ("");
        fprintf (stderr, "Statistics will be output to stdout\n");
        f = stdout;
      }				/* End if (!fopen) */
    }				/* End else */

    stp->outfile = f;
  } else {
    stp->outfile = NULL; 	/* For other processes */
  }				/* End if (0) */

  stp->gens = (unsigned long) 0;
  stp->period = period;

} /* End init_sts */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: reset_stats
 * 
 * Resets a stats structure
 * 
 * Parameters:
 *	O   stp		Stats ptr to be reset
 * 
 * Implementation:
 *	Exits if work is cleared. Else it resets the timer and gens members.
 *	Total is not reset as it is not stopped.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void reset_stats (stats *stp) {

  if (!(stp->work)) {
    return;
  } 			/* End if (!(stp->work)) */

  init_tmr (&(stp->func), (char *) NULL);
  init_tmr (&(stp->vis), (char *) NULL);
  init_tmr (&(stp->prj), (char *) NULL);
  init_tmr (&(stp->bound), (char *) NULL);
  init_tmr (&(stp->steer), (char *) NULL);
  init_tmr (&(stp->total), (char *) NULL);

  stp->gens = (unsigned long) 0;

} /* End reset_stats */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: start_one_timer_sts
 * 
 * Starts one of the timers in the stats struct
 * 
 * Parameters:
 *	O   stp		Stats ptr to be altered
 *	I   tmr		Tmr_code for the timer to start
 *	I   gen		Int generation for start_gen member
 * 
 * Implementation:
 *	Returns if stp is not working. If tmr is not TOTAL, it calls the
 *	corresponding start_tmr call and returns what that returns. If
 *	it is TOTAL it also checks and sets gen to start_gen.
 *
 * Returns: (int)
 *	(ve) 	on success.
 *	(0) 	if ignored.
 *	(-ve) 	on failure.
 *
 *----------------------------------------------------------------------*/

static int start_one_timer_sts (stats *stp, tmr_code tmr, int gen) {

  if (!(stp->work)) {
    return (0);
  } 			/* End if (!(stp->work)) */

  switch (tmr) {
    case FUNC: {
      return start_tmr (&(stp->func));
    }			/* End case FUNC */

    case VIS: {
      return start_tmr (&(stp->vis));
    }			/* End case VIS */

    case PRJ: {
      return start_tmr (&(stp->prj));
    }			/* End case PRJ */

    case BOUND: {
      return start_tmr (&(stp->bound));
    }			/* End case BOUND */

    case STEER: {
      return start_tmr (&(stp->steer));
    }			/* End case STEER */

    case TOTAL: {
      int ret;

      if (0 > gen) {
        fprintf (stderr, "start_one_timer_sts: erratic gen spec: %i\n",
                 gen);
        return (-1);
      }			/* End if (0 > gen) */

      ret = start_tmr (&(stp->total));

      if (ret > 0) {
        stp->start_gen = gen;
      }			/* End if (ret > 0) */
 
      return (ret);
    }			/* End case TOTAL */

    default: {
      fprintf (stderr, "start_one_timer_sts: unknown tmr_code %i\n", tmr);
      return (-1);
    }			/* End default */

  }			/* End switch */

} /* End start_one_timer_sts */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: stop_one_timer_sts
 * 
 * Stops one of the timers in the stats struct
 * 
 * Parameters:
 *	O   stp		Stats ptr to be altered
 *	I   tmr		Tmr_code for the timer to stop
 *	I   gen		Int generation
 * 
 * Implementation:
 *	Returns if stp is not working. If tmr is not TOTAL, it calls the
 *	corresponding stop_tmr call and returns what that returns. If
 *	it is TOTAL it also checks and adds start_gen-gen to gens.
 *
 * Returns: (int)
 *	(ve) 	on success.
 *	(0) 	if ignored.
 *	(-ve) 	on failure.
 *
 *----------------------------------------------------------------------*/

static int stop_one_timer_sts (stats *stp, tmr_code tmr, int gen) {

  if (!(stp->work)) {
    return (0);
  } 			/* End if (!(stp->work)) */

  switch (tmr) {
    case FUNC: {
      return stop_tmr (&(stp->func));
    }			/* End case FUNC */

    case VIS: {
      return stop_tmr (&(stp->vis));
    }			/* End case VIS */

    case PRJ: {
      return stop_tmr (&(stp->prj));
    }			/* End case PRJ */

    case BOUND: {
      return stop_tmr (&(stp->bound));
    }			/* End case BOUND */

    case STEER: {
      return stop_tmr (&(stp->steer));
    }			/* End case STEER */

    case TOTAL: {
      int ret;
      int gens;

      ret = stop_tmr (&(stp->total));
      if (ret > 0) {
        if (0 > (gens = (gen-stp->start_gen))) {
          fprintf (stderr, "stop_one_timer_sts: erratic gens spec: %i\n",
                   gens);
          return (-1);
        }		/* End if (0 > gens) */

        stp->gens += gens;
      }			/* End if (ret > 0) */

      return (ret);

    }			/* End case TOTAL */

    default: {
      fprintf (stderr, "stop_one_timer_sts: unknown tmr_code %i\n", tmr);
      return (-1);
    }			/* End default */

  }			/* End switch */

} /* End stop_one_timer_sts */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: print_sts
 * 
 * Prints the data for a stats structure.
 * 
 * Parameters:
 *	I   stp		Const stats ptr to be printed
 * 
 * Implementation:
 *	Checks if stats not working, in which case it returns.
 *	It prints a ~ separator, then prints the rank and gens members
 *	in one line. It then prints the timers except for total, which
 *	it prints after printing the sum of times of the other timers.
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void print_sts (const stats *stp) {

  if (!(stp->work)) {
    return;
  } 			/* End if (!(stp->work)) */

  fprintf (stp->outfile, "\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n");
  fprintf (stp->outfile, "Process: %i\t\t\t\t\t\tGenerations: %i\n\n", stp->rank, stp->gens);
  fprintf (stp->outfile, "\t\t\t Calls\t\t  Time\t\t  Best\t\t  Worst\n");
  fprintf (stp->outfile, "--------------------------------------------------------------------------------\n");
  print_tmr (&(stp->func), stp->outfile);
  print_tmr (&(stp->steer), stp->outfile);
  print_tmr (&(stp->bound), stp->outfile);
  print_tmr (&(stp->vis), stp->outfile);
  print_tmr (&(stp->prj), stp->outfile);

  fprintf (stp->outfile, "\nSum\t\t:\t\t\t%f\n", 
           stp->func.sum+stp->steer.sum+stp->vis.sum+stp->prj.sum+stp->bound.sum);

  print_tmr (&(stp->total), stp->outfile);

} /* End print_sts */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: close_file_sts
 * 
 * Closes the output file for the stats structure
 * 
 * Parameters:
 *	I   stp		Const stats ptr to close its file
 * 
 * Implementation:
 *	Only closes the file if it is not stdout
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void close_file_sts (const stats *stp) {

  if ((0 == cmtWorldRank) && stp->work && (stdout != stp->outfile) && fclose (stp->outfile)) {
    fprintf (stderr, "close_file_sts: fclose failed!\n");
    perror ("");
  }

} /* End close_file_sts */

/*----------------------------------------------------------------------
 state_dt library
 ----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: init_state_dt
 * 
 * Initialises a state_dt struct. Assumes it is malloced.
 * 
 * Parameters:
 *	O   st_dt_ptr	State_dt ptr to be initialised
 *  	I   data	MPI_Datatype of data in struct
 *
 * Implementation:
 *	Trivial
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void init_state_dt (state_dt *st_dt_ptr, MPI_Datatype data) {

/* Initialise members */
  st_dt_ptr->states = 0;
  st_dt_ptr->data = data;
  
} /* End init_state_dt */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: add_state
 * 
 * Add a state to state_dt
 * 
 * Parameters:
 *	O   st_dt_ptr	State_dt ptr to be augmented
 *	I   stateid	Int id to be inserted
 * 
 * Implementation:
 *
 * Returns: (int)
 *	(ve) 	on success
 *	(-ve)	on failure
 *
 *----------------------------------------------------------------------*/

static int add_state (state_dt *st_dt_ptr, int stateid) {

/* Check stateid */
  if ((0 > stateid) || (NumOfStates-1 < stateid) 
       || (cpt_state_mpidt[stateid] != st_dt_ptr->data)) {
    fprintf (stderr, "add_state: illegal stateid definition: %i\n",
             stateid);
    return (-1);
  }				/* End if (bad stateid) */

/* Check states */
  if (NumOfStates == st_dt_ptr->states) {
    fprintf (stderr, "add_state: too many states in array\n");
    return (-1);
  }				/* End if (too many states) */

/* Add state to state_ind */
  st_dt_ptr->state_ind[st_dt_ptr->states] = stateid;
/* Increment states */
  st_dt_ptr->states++;

  return (1);

} /* End add_state */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: init_state_dt_list
 * 
 * Initialises a state_dt_list struct. Assumes it is malloced.
 * 
 * Parameters:
 *	O   st_dt_l_ptr	State_dt_list ptr to be initialised
 *
 * Implementation:
 *	Sets many to 0, then calls add_state_dt() for all the substates
 *
 * Returns: (void)
 *
 *----------------------------------------------------------------------*/

static void init_state_dt_list (state_dt_list *st_dt_l_ptr) {
  int i;

/* Initialise member */
  st_dt_l_ptr->many = 0;
  
/* Add all states to dt_list */
  for (i = 0; i < NumOfStates; i++) {
    add_state_dt (dt_list, i);
  }				/* End for */

} /* End init_state_dt_list */

/*----------------------------------------------------------------------*/
/* 
 * PRIVATE: add_state_dt
 * 
 * Add a state to state_dt_list
 * 
 * Parameters:
 *	O   st_dt_l_ptr	State_dt_list ptr to be augmented
 *	I   stateid	Int id of substate to be inserted
 * 
 * Implementation:
 *
 * Returns: (int)
 *	(ve) 	on success
 *	(-ve)	on failure
 *
 *----------------------------------------------------------------------*/

static int add_state_dt (state_dt_list *st_dt_l_ptr, int stateid) {
  int i;
  int ret;

/* Check stateid */
  if ((0 > stateid) || (NumOfStates-1 < stateid)) {
    fprintf (stderr, "add_state_dt: illegal stateid definition: %i\n",
             stateid);
    return (-1);
  }				/* End if (bad stateid) */

/* Check many */
  if (NumOfStates == st_dt_l_ptr->many) {
    fprintf (stderr, "add_state_dt: too many states in array\n");
    return (-1);
  }				/* End if (too many states) */

/* Add state */
  for (i = 0; i < st_dt_l_ptr->many; i++) {
    if (st_dt_l_ptr->statetypes[i].data == cpt_state_mpidt[stateid]) {
      break; 	/* Found it! */
    }				/* End if (data == mpidt[stateid]) */
  }				/* End for (i) */

  if (i == st_dt_l_ptr->many) {	/* New state_dt */
    init_state_dt (&(st_dt_l_ptr->statetypes[i]), cpt_state_mpidt[stateid]);
/* Increment many */
    st_dt_l_ptr->many++;

  }				/* End if (new state_dt) */

  if (0 > (ret = add_state (&(st_dt_l_ptr->statetypes[i]), stateid))) {
    fprintf (stderr, "add_state_dt: add_state failure\n");
  }				/* End if (0 > add_state) */ 

  return (ret);

} /* End add_state_dt */
